#!/bin/bash

#
#  Please do not follow the coding example in this script.  The example.sh script has been designed
#  for the purpose of being cloned/modified for future script implementations.
#


unset CDPATH  # CDPATH causes every "cd" command to print the name of the target
	      # directory.

# comment these two lines out if you don't want an execution log generated.
# it slows the execution down slightly.

exec 2> /tmp/log
set -x

########################################################
# The first thing the script does is set up the empty
# state of the state variables.
########################################################

# required state variables

STATE=start         
CMDKEYS=abqhHcCdDn
TITLE="GIT BRANCH BROWSER:  HEAD"         
HELP=""             
SEPCHAR=""



# application specific state variables

TOP_DIR=""           # pathname just above .git
CUR_DIR=""           # underneath TOP_DIR somewhere (ends in /)
END_VERSION="HEAD"   # HEAD or end of current range
BEG_VERSION=""       # "" or "STARTVERSION" + ".."
BOTTOM="false"       # boolean, means BEG_VERSION is first commit
DIR_COMMIT=""        # the SHA1 of the commit to current dir of interest
DIR_ITEM_NAME=""     # file name on current directory row.
DIR_ITEM_BLOB=""     # blob (sha1) of item on current directory row
BRANCH="master"      # name of current branch


# notes:
#
#  "$BEG_VERSION$END_VERSION" defines the range of valid
#  versions for logging.  BEG_VERSION may be either an
#  empty string or "SHA1...".  Where SHA1 refers to a commit
#  not a blob.

printStateVariables()
{
   #
   #  This function should be called before any other
   #  output from the script.  It should presumbably
   #  be called AFTER the new STATE values are set based on
   #  an interpretation of the current state and user's
   #  inputs.
   #
   echo "=%==%--SEPCHAR=$SEPCHAR"
   echo "=%==%--STATE=$STATE"
   echo "=%==%--TITLE=$TITLE"
   echo "=%==%--CMDKEYS=$CMDKEYS"
   echo "=%==%--TOP_DIR=$TOP_DIR"
   echo "=%==%--CUR_DIR=$CUR_DIR"
   echo "=%==%--END_VERSION=$END_VERSION"
   echo "=%==%--BEG_VERSION=$BEG_VERSION"
   echo "=%==%--BOTTOM=$BOTTOM"
   echo "=%==%--DIR_COMMIT=$DIR_COMMIT"
   echo "=%==%--DIR_ITEM_NAME=$DIR_ITEM_NAME"
   echo "=%==%--DIR_ITEM_BLOB=$DIR_ITEM_BLOB"
   echo "=%==%--BRANCH=$BRANCH"

   # do NOT put HELP in here
   # do NOT put the names of POPUP_FIELD_* variables here
   # do NOT put POPUP_ABORTED in here
}

################################################################
# detect actual current state from command line.  command line options
# defined like this populate the new state:
#
#  --STATE_VARIABLE_NAME="value"
#
# The actual command line is long and looks basically like this:
#
#  --STATE=name1 --SVN2=value2 --SVN3=value3 ...   row col columnContents rowContents
#
# where SVN2 and SVN3 are state variable names
#
################################################################

while [ $# != 0 ]
do
   # parse options

   case "$1" in
     --*=*)

	  ;;     # process options beginning with --name=
     *)
	  break  # quit when the option does not begin with --
   esac

   name="${1/%=*/}"  # remove =value from a --name=value pair
   value="${1/#*=/}" # remove --name from the --name=value pair

   name=${name:2}    # remove leading --

   eval "$name=\"$value\""

   shift
done

###############################################################
# Initialize essential variables not set by the command line
# processing.
###############################################################

if [ "$SEPCHAR" = "" ]
then
   SEPCHAR=`echo -n -e '\001'`
fi


if [ "$TOP_DIR" = "" -a "$HELP" = "" ]
then
   #
   # we don't get here if help is being processed
   #

   STATE=start  # this situation only makes sense if we are just
		# starting, so lets do a little cleanup to
		# make sure...

   TOP_DIR=`git-topdir`

   if [     "$TOP_DIR" = "" \
	-o  ! -d "$TOP_DIR" \
      ]
   then
       echo "|Error"
       echo ""
       echo "TOP_DIR could not be established from the"
       echo "current directory.  Is this a valid git controlled"
       echo "directory?"
       exit 0
   fi

fi

cd "$TOP_DIR"



################################################################
# BEGIN HELP
################################################################

if [ "$HELP" != "" ]
then

#--------------------------------------------------------------------

dirHelp()
{
sed -e 's/.*/help:&/1' - <<EOH

dir   -- Presumes a directory has been displayed
         and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit

	    b     -- display the branches list and enter the "branch"
		     state.

	    enter -- Show the file or directory on current
		     line.  The most recently committed
		     version that was checked in during
		     the current time range will be displayed.
		     Switch to the "file" state if it is a file,
		     else stay in the directory state for the
		     new directory.

	    h     -- Show the history of the file or directory
		     on the current line within the current
		     time range.  Switch to state "history"

	    H     -- Show the history of the current
		     directory within the current time range.
		     Switch to state "history".
	    
	    d     -- If the current line defines a directory,
		     display it.  Otherwise, perform a diff
		     between the version of the file displayed
		     and its immediate predecessor.  Enter the "diff"
		     state.

	    D     -- If the current line defines a directory,
		     display it.  Otherwise, perform a diff between
		     the actual file in the current checked out view
		     and the checked in version under the cursor.  Enter
		     the "diff" state.

	    c     -- Copy the file under the cursor to a user prompted
		     location.  If the file is a directory, refuse.

	    a     -- If the current line defines a directory, display it.
		     Otherwise, print the results of 'git annotate' on
		     the selected version of the file.  Enter the 
		     "annotate" state.

EOH
}

#--------------------------------------------------------------------

fileHelp()
{
sed -e 's/.*/help:&/1' - <<EOH
file   -- Presumes that the user is looking at a version of a file
          and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit

	    h
	    enter -- Show the history of this file within the
		     currently selected range and enther the "history"
		     state.
	    
	    d     -- Diff this file version with its predecessor.
		     Enter the diff mode.

	    D     -- Diff this file version with the checked out copy.
		     Enter the diff mode.

	    c     -- Copy the file to a user prompted
		     location.  

	    H     -- Show the history of the current
		     directory within the current time range.
		     Switch to state "history".

	    a     -- Print the results of 'git annotate' on
		     the selected version of the file.  Enter the 
		     "annotate" state.
EOH
}

#--------------------------------------------------------------------

diffHelp()
{
sed -e 's/.*/help:&/1' - <<EOH
diff   -- Presumes that the user is looking at a diff log
          and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit
	    
	    c     -- Copy the diff to a user prompted location.
EOH
}

#--------------------------------------------------------------------

historyHelp()
{
sed -e 's/.*/help:&/1' - <<EOH
history - Presumes that the user is looking at a git log of a file
          and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit
	    
	    c     -- Copy the history to a user prompted location.

	    enter -- Within the git log, if the cursor is sitting on
		     a line that looks like this:

			 :10644 10644 1234... 5678...     directory/file

		     Then show a diff of the file, before and after
		     the commit and enter the diff state.


		     If the cursor is sitting on a line that looks
		     like this:

			 commit 1234567890....

		     Then show top level repository directory as it
		     was during the time of commit 1234...  This requires
		     that you renavigate to the current file location to
		     see the version of this file at that time, but it
		     also lets you see ALL the files at that revision
		     level.



	    n     -- Show the next 100 version page of the history.
		     At most 100 versions will be displayed in a single
		     history display.  See display history page.
EOH
}

#--------------------------------------------------------------------

branchHelp()
{
sed -e 's/.*/help:&/1' - <<EOH
branch -- Presumes that a list of the branches is showing and interprets
          and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit
	    
	    c     -- Copy the branch list to a user prompted location.

	    enter -- select the branch under the current line and
		     make it the active branch.  The display the top
		     level git directory and enter the "dir" state.
EOH
}

#--------------------------------------------------------------------

annotateHelp()
{
sed -e 's/.*/help:&/1' - <<EOH
annotate- Presumes that the user is looking at a diff log
          and interprets user input as follows:

	    q 
	    Q
	    ESC   -- exit

	    enter -- show change info for the line under cursor
	    
	    c     -- Copy the annotation to a user prompted location.
EOH
}


sed -e 's/.*/help:&/1' - <<EOH

GIT.SH -- A Git branch browsing tool

Abstract

   This tool allows you browse the GIT repository contents --
   though not necessarily the contents of the directories you
   might be used to seeing.  It lets you browse old versions
   of files, perform diffs, look at other branches etc.

Description

   You can look at directories and files and branches and old
   versions of them using the git repository information.  You
   can also look at annotations, differences, and change logs.

   The browser indicates what version of the files and 
   directories are being displayed at the moment.  If you are 
   looking at a directory page, the title indicates the most
   recent version that you may see on any page.  The history
   commands can be used to select older versions for viewing.
   You can invoke the history command using the 'h' or 'H'
   key, depending on the situation.

   The page you are displaying determines what the command keys
   mean.  For example, you can almost always select on any
   highlighted line using the enter key.  The enter key 
   executes the default action for the page.   For example,
   if you are looking at a directory listing, the enter key
   will cause the browser to show you the contents of the file
   on the line where the cursor is found.  If you are looking
   at a directory, it will show you a directory listing of
   that directory.

   The other command keys that are active depend on the current 
   page that you are showing.

   You are showing state = $STATE.  Commands available in this
   state are as follows:

EOH

   # all help output must being 'help:'


   echo "help:---------------------------------------------------------------"

   case "$STATE" in
       dir)        dirHelp        ;;
       file)       fileHelp       ;;
       annotate)   annotateHelp   ;;
       history)    historyHelp    ;;
       diff)       diffHelp       ;;
   esac
   
   # execute the function whose name is computed
   # from the current state and the word Help.

   echo "help:---------------------------------------------------------------"



    exit 0
fi

################################################################
# END HELP
################################################################

noCanDoBigGuy()
{
  echo "|Sorry"
  echo ""
  echo "[_!]Cannot display binary files"
  echo ""
}


####################
####################
catBlob()
{
   #
   #  Display the current item name and its contents as a show-file
   #

   if [ "$DIR_ITEM_NAME" = "" -o "$DIR_ITEM_BLOB" = "" ]
   then
      echo "|Error"
      echo ""
      echo "Expected DIR_ITEM_NAME and DIR_ITEM_BLOB to be set"
      return
   fi

   case "$DIR_ITEM_NAME" in

       *.png|*.pdf|*.mpg|*.mpeg|*.jpg|*.jpeg|*.wmv|*.avi|*.bmp)
	   noCanDoBigGuy
           return
	   ;;

       *.wav|*.ogg)
	   noCanDoBigGuy
           return
	   ;;

       *.o|*.exe|*.obj|*.lib|*.a|*.so|*.dll)
	   noCanDoBigGuy
           return
	   ;;


       *)
	   ;; # use the ascii display method
   esac

   #
   # display ascii files:
   #
   #   1.  determine if this is a csv of tsv file.  If not, go ahead and output
   #       the normal title line for this file.
   #
   #   2.  if it is a special file, then don't put out the standard title because
   #       the first line of the file has the titles in it.
   #

   csvFile=0
   tsvFile=0

   case "$DIR_ITEM_NAME" in
       *.csv)
	   #
	   #  DO NOT OUTPUT THE HEADER LINE ON CSV FILES
	   #
	   csvFile=1
	   ;;
       *.tsv)
           #
	   #  DO NOT OUTPUT THE HEADER LINE ON TSV FILES
           #
	   tsvFile=1
           ;;
       *)
	   echo "${SEPCHAR}[x]_______________________________________________________"  # title has filename in it
	   ;;
   esac

   git cat-file -p "$DIR_ITEM_BLOB" 2>&1 |
   if [ "$csvFile$tsvFile" == "00" ]
   then
       # this file needs no special processing
       sed -e 's/.*/[x]&/1'
   else

       #
       #  Fix up the first line of the csv and tsv files so that
       #  it is a proper header for this kind of file -- the first
       #  character of the line defines the separator character.
       #

       if [ "$csvFile" = 1 ]
       then
	   sed -e '1s/.*/,&/1;   1!s/.*/[x]&/1; '
       else

	   tab=`  echo -n -e '\t' `

	   sed -e "1s/.*/${tab}&/1;  1!s/.*/[x]&/1; "
       fi
   fi 

   echo ""  # guarantee at least 1 line due to bugs in TableCommander


}


####################
####################
displayDirectory()
{
  #
  #  display the $CUR_DIR's latest version within the range
  #  $BEG_VERSION$END_VERSION.  Note that directories don't
  #  have versions, exactly.  Commits imply versions.  Every
  #  commit defines the contents of every directory so the most
  #  recent commit anywhere in the directory defines its version.
  #
  #  Adjust the variable, DIR_COMMIT, so that it refers to the
  #  specified commit to $CUR_DIR
  #

  DIR_COMMIT=""

  eval $(
	  #
	  #  Evaluate the printed output of the following mess
	  #  as if it were script commands of the form:
	  #
	  #    DIR_COMMIT=value1
	  #

	  git log -1 "$BEG_VERSION$END_VERSION" -- "$CUR_DIR" |
	  while read commit commitHash
	  do
	      echo "DIR_COMMIT=$commitHash"
	      break # we should only get one, but saftey is good
	  done
       )

   if [ "$DIR_COMMIT" = ""  ]
   then
      echo "|Error"
      echo ""
      echo "Unable to read version or tree info for dir $CUR_DIR"
      echo ""
      exit 0
   fi

   TITLE="GIT: branch=$BRANCH ${DIR_COMMIT:0:8}... $CUR_DIR"

   printStateVariables

   echo "|[x]Type|[x]Attr|[x]Size____|[x]File_Name________________________|[x]Blob id"

   (
       cd "$CUR_DIR"
       git ls-tree -l "$DIR_COMMIT" 
   )|
   sed -e '
	      #
	      #  The ls-tree output looks like this
	      #
	      #     1    2      3     4    5
	      #    attr type version size name
	      #
	      
	      #
	      #  Convert tabs into spaces to simplify regexes
	      #
	      y/\t/ /; 
	      
	      #
	      #  Directories in the listing have no file size, so
	      #  put use this info. 
	      #
	      / - /{  s/\(.\)$/\1\//1  };   
	      
	      
	      #
	      #  Move the fields around to a more pleasing format
	      #
	      #     type attr size name version
	      #
	      s/...\(...\)  *\([^ ]*\)  *\([^ ]*\)  *\([^ ]*\)  *\(.*\)/\2|\1|\4|\5|\3/g
	      
	      # convert type to human readable form
	      
	      s/^tree/D/;
	      s/^blob/ /;
	      
	      #
	      # convert the file permissions into a readable form
	      #
	      #   |644| gets translated into |rwxrwxrwx|
	      
	      /^\([^|]*\)|6/{  s/^\([^|]*\)|6/\1|rw-/1 }
	      /^\([^|]*\)|4/{  s/^\([^|]*\)|4/\1|r--/1 }
	      /^\([^|]*\)|5/{  s/^\([^|]*\)|5/\1|r-x/1 }
	      /^\([^|]*\)|7/{  s/^\([^|]*\)|7/\1|rwx/1 }
	      
	      /^\([^|]*\)|\(...\)6/{  s/^\([^|]*\)|\(...\)6/\1|\2rw-/1 }
	      /^\([^|]*\)|\(...\)4/{  s/^\([^|]*\)|\(...\)4/\1|\2r--/1 }
	      /^\([^|]*\)|\(...\)5/{  s/^\([^|]*\)|\(...\)5/\1|\2r-x/1 }
	      /^\([^|]*\)|\(...\)7/{  s/^\([^|]*\)|\(...\)7/\1|\2rwx/1 }
	      
	      /^\([^|]*\)|\(......\)6/{  s/^\([^|]*\)|\(......\)6/\1|\2rw-/1 }
	      /^\([^|]*\)|\(......\)4/{  s/^\([^|]*\)|\(......\)4/\1|\2r--/1 }
	      /^\([^|]*\)|\(......\)5/{  s/^\([^|]*\)|\(......\)5/\1|\2r-x/1 }
	      /^\([^|]*\)|\(......\)7/{  s/^\([^|]*\)|\(......\)7/\1|\2rwx/1 }
	      
	      #
	      #  convert the directory attributes into blanks to eliminate
	      #  visual confusions
	      #
	      
	      s/^D|000|-|/D|  |  |/1;
	      
	      
	      #
	      #  embolden directories and enable commands
	      #
	      /\//{ s/.*/[x!]&/1;};
	      
	      #
	      #  enable commands on non directories
	      #
	      /\//!{ s/.*/[x]&/1; };

	  '
}

####################
####################
predecessorDiff()
(
    # note that the current directory is restored automatically
    # because this function is defined using () not {}

    cd "$CUR_DIR"

    #
    #  Execute git log to find the most recent two commits to the file
    #  of interest given the directory we are interested in.
    #
    #  Execute a git log, strip out all but the commits, read the two
    #  commit lines, turn them into executable shell statements, and
    #  then evaluate them.
    #

    eval $(
	      git log -2 "$BEG_VERSIONS$DIR_COMMIT" -- "$DIR_ITEM_NAME" |
	      grep '^commit' |
	      (
		 EARLIER=""
		 LATER=""
	  
		 read commit later
		 read commit earlier
	  
		 echo EARLIER=$earlier
		 echo LATER=$later
	  
	      )
	 )

    if [ "$EARLIER" = "" ]
    then
       echo "|Note regarding $DIR_ITEM_NAME"
       echo ""
       echo "There is no earlier version to compare against"
       echo ""
       return
    fi

    if [ "$LATER" = "" ]
    then
       echo "|Note regarding $DIR_ITEM_NAME"
       echo ""
       echo "Can't find the version info for this file"
       echo ""
       return
    fi

    echo "${SEPCHAR}[x]__________________________________________________________________________________"

    echo "[x]"

    git log -1 "$LATER" -- "$DIR_ITEM_NAME" |
    sed -e 's/.*/[x]&/1'

    echo ""
    echo "[x_!]__________________________________________________________________________________"
    echo "[x!]"

    git diff "$EARLIER".."$LATER" -- "$DIR_ITEM_NAME" |
    sed -e '
		1,4s/.*/[x!]&/1
		4a\
[x!]__________________________________________________________________________________
	     /^\[/! s/.*/[x]&/1; # on non-highlighted lines, make them executable
	   ' 

    echo "" # guarantee at least 1 line
)

####################
####################
currentDiff()
(
    # diff the current version of the specified file against the
    # historical version represented by DIR_ITEM_BLOB


    # note that the current directory is restored automatically
    # because this function is defined using () not {}

    cd "$CUR_DIR"

    echo "${SEPCHAR}[x]__________________________________________________________________________________________"

    echo "[x]"

    tmpfile=/tmp/currentDiff.$$

    git cat-file -p "$DIR_ITEM_BLOB" >$tmpfile

    diff $tmpfile "$DIR_ITEM_NAME" |
    sed -e 's/.*/[x]&/1'

    rm $tmpfile

    echo "" # guarantee at least 1 line
)


####################
####################
historyColumnHeaders()
{
    echo "${SEPCHAR}[x]_______________________________________________________"  # title has filename in it
}

####################
####################
historyFilter()
{
    sed -n -e '
	       /^commit/{
			  s/.*/[x_!]&/1
			}

	       /^:/{
		      s/.*/[x!]&/1
		   }
	       p;
	   '
}

####################
####################
fileAnnotate()
(
    # note that the current directory is restored automatically
    # because this function is defined using () not {}

    cd "$CUR_DIR"
    #
    #  Execute git log to find the most recent two commits to the file
    #  of interest given the directory we are interested in.
    #
    #  Execute a git log, strip out all but the commits, read the two
    #  commit lines, turn them into executable shell statements, and
    #  then evaluate them.
    #

    eval $(
	      git log -1 "$BEG_VERSIONS$DIR_COMMIT" -- "$DIR_ITEM_NAME" |
	      grep '^commit' |
	      (
		 LATER=""
	  
		 read commit later
	  
		 echo LATER=$later
	  
	      )
	 )

    if [ "$LATER" = "" ]
    then
       echo "|Note regarding $DIR_ITEM_NAME"
       echo ""
       echo "Can't find the version info for this file"
       echo ""
       return
    fi


    historyColumnHeaders

    git annotate "$LATER" -- "$DIR_ITEM_NAME" |
    sed -e '
	     s/(\(...........\).*)/\1  |/1
	     s/.*/[x]&/1
	   '
    

    echo "" # guarantee at least 1 line
)

####################
####################
handleAnnotationCopy()
{
    # we get here both before and after the popup that the
    # c command triggers and we have to interpret what case
    # we are in
    
    if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
    then
        #
        # This is the first invocation of the script in this
        # configuration (requesting the popup)
        #
        # Send the description of the popup to the script table viewer
        #
	echo "=%==%--POPUP=Save Annotation to;Destination;" |
             sed -e 's/!/!!/g'   \
                 -e 's/ /!20/g' \
                 -e 's/;/!3b/g' 
        exit 0
    else
        #
        #  This is a reinvocation of the script after a popup has
        #  run, now lets decide if it was aborted or was successful.
        #
    
        if [ "$POPUP_ABORTED" != "" ]
        then
            STATE=errorHandler
            printStateVariables
            echo "|User abort"
            echo ""
	    echo "The user aborted the save operation"
            exit
    
        else

            tmpfile=/tmp/copyblob.$$

	    if ! fileAnnotate | sed -e 's/^[^]]\+]//1' >"$POPUP_FIELD_1" 2>$tmpfile
            then
                STATE=errorHandler
                printStateVariables
		echo "|Save error"
                echo ""
		echo "Error saving the file"
                echo ""
                cat $tmpfile
                rm $tmpfile
                exit
            fi

            rm $tmpfile


	    STATE=annotationCopyComplete  # state for the next invocation
            printStateVariables
	    echo "|Annotation saved"
            echo ""
            echo "Data saved to $POPUP_FIELD_1"
            exit
        fi
    fi
}

####################
####################
handleAnnotateCommand()
{
   if [ "$1" = "c" ]
   then
       handleAnnotationCopy "$@"
   else
       STATE=history
       TITLE="Change to file $CUR_DIR$DIR_ITEM_NAME(${6:0:8})"
       printStateVariables
    
       echo "${SEPCHAR}[x]___________________________________________________________________"
       echo ""
       echo ""
       git log -1 --raw "${6:0:8}" |
       historyFilter
   fi


	      
}

####################
####################
itemHistory()
(
    # note that the current directory is restored automatically
    # because this function is defined using () not {}

    cd "$CUR_DIR"

    item="$1"


    #
    #  display the history of $DIR_ITEM_NAME with in the current range
    #
    
    echo "${SEPCHAR}[x]_____________________________________________________"
    
    git log -100 --raw "$BEG_VERSION$END_VERSION" -- "$item" 2>&1 |
    historyFilter

    echo ""

)

####################
####################
handleFileCopy()
{
    # we get here both before and after the popup that the
    # c command triggers and we have to interpret what case
    # we are in
    
    if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
    then
        #
        # This is the first invocation of the script in this
        # configuration (requesting the popup)
        #
        # Send the description of the popup to the script table viewer
        #
        echo "=%==%--POPUP=Copy File To;Desitination;" |
             sed -e 's/!/!!/g'   \
                 -e 's/ /!20/g' \
                 -e 's/;/!3b/g' 
        exit 0
    else
        #
        #  This is a reinvocation of the script after a popup has
        #  run, now lets decide if it was aborted or was successful.
        #
    
        if [ "$POPUP_ABORTED" != "" ]
        then
            STATE=errorHandler
            printStateVariables
            echo "|User abort"
            echo ""
            echo "The user aborted the copy operation"
            exit
    
        else

            tmpfile=/tmp/copyblob.$$

            if ! git cat-file -p "$DIR_ITEM_BLOB" >"$POPUP_FIELD_1" 2>$tmpfile
            then
                STATE=errorHandler
                printStateVariables
                echo "|Copy error"
                echo ""
                echo "Error copying the file"
                echo ""
                cat $tmpfile
                rm $tmpfile
                exit
            fi

            rm $tmpfile


            STATE=fileCopyComplete  # state for the next invocation
            printStateVariables
            echo "|Copy Complete"
            echo ""
            echo "Data saved to $POPUP_FIELD_1"
            exit
        fi
    fi
    
}

####################
####################
listBranches()
{
   echo "|[x]Active|Name|Commit|Comment"
   git branch -v --no-abbrev |
   sed -e '
	    # branch output looks like this
	    #
	    # * master commitID Comment
	    #   other  otherID  OtherComment

	    s/^\(.\) \+\([^ ]\+\) \+\(..........\)[^ ]\+ \+\(.*\)/[x]\1|\2|\3|\4/1

	  '
}


####################
####################
handleDirCommand()
{
    #
    #  Args are as follows:
    #   1   2 3 4  5  6   7     8       9    10        11
    #  '^J' 1 1 26 5 'x' 'x' rw-r--r-- 92 .gitignore 8da62b4a4fa4
    #
    #  Args have the following meaning:
    #
    #   1:  cmd key user pressed
    #   2:  row 
    #   3:  col
    #   4:  rows
    #   5:  cols
    #   6:  content of row,col
    #

    cmd="$1"
    type="$6"

    shift 9  # discard 1-9

    DIR_ITEM_NAME="$1"  # .gitignore in the above example
    DIR_ITEM_BLOB="$2"  # 8da62b4a4fa4 in the above


    case "$cmd" in

	a)
	    if [ "$type" != "D" ]
	    then
		STATE=annotate
		TITLE="Annotating $DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
		printStateVariables
		fileAnnotate  # uses DIR_ITEM_NAME and DIR_COMMIT
	    else
		echo "|Error"
		echo "Can't annotate directories yet"
	    fi
	    ;;

	d)
	    if [ "$type" != "D" ]
	    then
		STATE=diff
		TITLE="Diffing checked in copy of $DIR_ITEM_NAME(${DIR_COMMIT:0:8}) against predecessor"
		printStateVariables
		predecessorDiff  # uses DIR_ITEM_NAME and DIR_COMMIT
	    else
		echo "|Error"
		echo "Can't diff directories yet"
	    fi
	    ;;

	D)
	    if [ "$type" != "D" ]
	    then
		STATE=diff
		TITLE="Diffing $DIR_ITEM_NAME(${DIR_COMMIT:0:8}) against current copy"
		printStateVariables
		currentDiff  # uses DIR_ITEM_NAME and DIR_COMMIT
	    else
		echo "|Error"
		echo "Can't diff directories yet"
	    fi
	    ;;


	h)
	    STATE=history
	    TITLE="History of $DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
	    printStateVariables
	    itemHistory  "$DIR_ITEM_NAME" 
	    ;;

	H)
	    STATE=history
	    TITLE="History of $CUR_DIR(${DIR_COMMIT:0:8})"
	    printStateVariables
	    itemHistory  "." 
	    ;;

	c)
	    handleFileCopy "$@"
	    ;;

	b)
	    STATE=branch
	    TITLE="Branches in this repository"
	    printStateVariables
	    listBranches "$@"
	    ;;


	*)
	    #
	    #  just display the current file
	    #
            if [ "$type" != "D" ]
            then
		STATE=file
		TITLE="$DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
		printStateVariables
		catBlob  # uses DIR_ITEM_NAME and DIR_ITEM_BLOB as parms
	    else
		CUR_DIR="$CUR_DIR$DIR_ITEM_NAME" 
		displayDirectory
	    fi
	    ;;
    esac


   
}


####################
####################
handleFileCommand()
{
    case "$1" in

	a)
	    STATE=annotate
	    TITLE="Annotating $DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
	    printStateVariables
	    fileAnnotate  # uses DIR_ITEM_NAME and DIR_COMMIT
	    ;;

	h)
	    STATE=history
	    TITLE="History of $DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
	    printStateVariables
	    itemHistory  "$DIR_ITEM_NAME" 
	    ;;

	d)
	    STATE=diff
	    TITLE="Diffing checked in copy of $DIR_ITEM_NAME(${DIR_COMMIT:0:8}) against predecessor"
	    printStateVariables
	    predecessorDiff  # uses DIR_ITEM_NAME and DIR_COMMIT
	    ;;

	D)
	    STATE=diff
	    TITLE="Diffing $DIR_ITEM_NAME(${DIR_COMMIT:0:8}) against current copy"
	    printStateVariables
	    currentDiff  # uses DIR_ITEM_NAME and DIR_COMMIT
	    ;;

	c)
	    handleFileCopy "$@"
	    ;;

        *)
	    echo "|File command error, unbound key $1, ARGS="
	    for f in "$@"
	    do
	       echo "'$f'"
	    done
            ;;
    esac
}

gitDiffBlob()
{
   tmpFileA=/tmp/gitDiffBlob.$$.old
   tmpFileB=/tmp/gitDiffBlob.$$.new

   git cat-file -p $1 >$tmpFileA
   git cat-file -p $2 >$tmpFileB

   diff -tEp $tmpFileA $tmpFileB 2>&1

   rm $tmpFileA $tmpFileB


}


####################
####################
handleHistoryDiff()
{
   # given that argument 6 looks like this:  ':100644 100644 23b0581... 8da62b4... M .gitignore'
   # convert it into this:  'l="23b0581" ; g="8da62b4" ; file=".gitignore"'
   # then evaluate that string as a script command, which assigns the variables l, g, and file
   # to the appropriate values.


   eval $( echo "$6" | 
	   sed -e 's/...............\([^.]\+\)... \([^.]\+\)... ..\(.*\)/l="\1" ; g="\2" ; file="\3\"/1'
	 )           

   TITLE="Commit to file '$file' resulted in the following change"
   STATE=diff
   printStateVariables
   echo "${SEPCHAR}[x]__________________________________________________________________________________________"
   
   gitDiffBlob $l $g |
   sed -e 's/.*/[x!]&/1'

   echo "*EOF"
}

####################
####################
handleHistoryLimitRange()
{
    END_VERSION=$(  echo "$6" | sed -e 's/^commit //1' )
    TITLE="GIT: BRANCH=$BRANCH $BEG_VERSION$END_VERSION"
    STATE=dir
    CUR_DIR="./"
    displayDirectory

}


####################
####################
handleHistoryCommand()
{
    case "$1" in

	^J)
	    case "$6" in
		:*)
		    handleHistoryDiff "$@"          #   ":0000.* <filename>"
		    ;;

		 commit*)
		     handleHistoryLimitRange "$@"   #   "commit <versionid>"
		     ;;
	    esac
	    ;;

        *)
	    echo "|History command error, unbound key $1, ARGS="
	    for f in "$@"
	    do
	       echo "'$f'"
	    done
            ;;
    esac
}

####################
####################
handleBranchCommand()
{
    #
    #  Branch selection works like this:  If you are selecting the
    #  current branch, enter the dir state and show what we otherwise
    #  woul.
    # 
    #  If you are different branch, set the begin/end version range to the
    #  new branch and keep going.
    #


    CUR_DIR="./"
    cd "$TOP_DIR"

    if [ "$6" != "*" ]
    then
	BEGIN_VERSION="`git merge-base $9 HEAD`.."
	END_VERSION="$8"
    fi

    STATE=dir
    BRANCH="$8"
    displayDirectory

}

####################
####################
handleDiffCommand()
{

    STATE=file
    TITLE="$CUR_DIR$DIR_ITEM_NAME(${DIR_COMMIT:0:8})"
    printStateVariables
    catBlob


}

border="###################################################################"
border="###################################################################"
border="###################################################################"

ARGS=$#
let arg=1
for f in "$@"
do
     argument="arg[$arg]=$f"
     let arg=$arg+1
done


case "$STATE" in
    start)
	STATE=dir
	CUR_DIR="./"
	displayDirectory
	;;

    dir)
	#
	#  a directory is visible and the user has pressed a key
	#
	handleDirCommand "$@"
	;;

    annotate)
        #
	#  a file annotation is visible and the user has pressed a key
        #
	handleAnnotateCommand "$@"
	;;

    file)
        #
	#  a file's content is visible and the user has pressed a key
        #
	handleFileCommand "$@"
	;;

    history)
        #
	#  a git log is visible and the user has pressed a key
        #
	handleHistoryCommand "$@"
	;;

    diff)
        #
	#  a diff is being shown tothe user
        #
	handleDiffCommand "$@"
	;;

    branch)
	handleBranchCommand "$@"
	;;


    *)
	echo "|Error:  Invalid state=$STATE, ARGS are"

	for f in "$@"
	do
	   echo "'$f'"
	done

	;;
esac

exit 0

