#!/bin/bash

unset CDPATH
# comment these two lines out if you don't want an execution log generated.
# it slows the execution down slightly.

exec 2> /tmp/log
set -x


###############################################################
# ScriptTableViewer scripts are state machines which pass
# data back and forth to the table viewer.  In the initial
# invocation, the state is set to start.  The script prints
# intial information and transitions to some other state based
# on the needs of the script.
#
# In subsequent invocations, the state is passed back to the
# script, so that the role of the script in subsequent 
# invocations is to interpret the current state, the keys that
# the user pressed, and the data in the row and column that the
# user was in when he pressed a command key.
#
################################################################

########################################################
# The first thing the script does is set up the empty
# state of the state variables.
########################################################

# required state variables

STATE=listDir    # list .zip files in current directory
CMDKEYS=aAxX
TITLE="ZIP file browser"
HELP=""             
SEPCHAR=""
ARCHIVE=""   # archive file name
TABCHAR=""   # \t
TOP_DIR=""

# application specific state variables


printStateVariables()
{
   #
   #  This function should be called before any other
   #  output from the script.  It should presumbably
   #  be called AFTER the new STATE values are set based on
   #  an interpretation of the current state and user's
   #  inputs.
   #
   echo "=%==%--SEPCHAR=$SEPCHAR"
   echo "=%==%--TABCHAR=$TABCHAR"
   echo "=%==%--STATE=$STATE"
   echo "=%==%--TITLE=$TITLE"
   echo "=%==%--CMDKEYS=$CMDKEYS"
   echo "=%==%--ARCHIVE=$ARCHIVE"
   echo "=%==%--TOP_DIR=$TOP_DIR"

   # do NOT put HELP in here
   # do NOT put the names of POPUP_FIELD_* variables here
   # do NOT put POPUP_ABORTED in here
}

################################################################
# detect actual current state from command line.  command line options
# defined like this populate the new state:
#
#  --STATE_VARIABLE_NAME="value"
#
# The actual command line is long and looks basically like this:
#
#  --STATE=name1 --SVN2=value2 --SVN3=value3 ...   row col columnContents rowContents
#
# where SVN2 and SVN3 are state variable names
#
################################################################

while [ $# != 0 ]
do
   # parse options

   case "$1" in
     --*=*)

	  ;;     # process options beginning with --name=
     *)
	  break  # quit when the option does not begin with --
   esac

   name="${1/%=*/}"  # remove =value from a --name=value pair
   value="${1/#*=/}" # remove --name from the --name=value pair

   name=${name:2}    # remove leading --

   eval "$name=\"$value\""

   shift
done

if [ "$TOP_DIR" = "" ]
then
   TOP_DIR=`pwd`
fi

cd "$TOP_DIR"

################################################################
# BEGIN HELP
################################################################

if [ "$HELP" != "" ]
then
sed -e 's/.*/help:&/1' - <<EOH

ZIP file browser

EOH
    exit 0
fi


################################################################
# END HELP
################################################################

## startup initializations

if [ "$SEPCHAR" = "" ]
then
   SEPCHAR=` echo -n -e '\001' `
   TABCHAR=` echo -n -e '\t'   `
fi


##################################
##################################
handleDirCommands()
{
   if [ "$1" = "" ]
   then
       #
       # Initialization of the script
       #
       TITLE="`pwd`"
       printStateVariables
       
       echo "${SEPCHAR}[x]ZIP Files"
       
       for f in *.zip
       do
	  echo "[x]$f"
       done

   else

       case "$1" in

	   "^J")
               ARCHIVE="$6"
	       TITLE="$ARCHIVE"
	       STATE=zipContents
	       printStateVariables

	       echo "${SEPCHAR}[x]Length${SEPCHAR}Date${SEPCHAR}Time${SEPCHAR}Filename"

	       unzip -l "$6" |
	       sed -n -e "
		       4,\${
			    # lines 1-3 of the output are a header,
			    # just leave them alone.
			    # the second to the last line has \" ----\" and its a footer
			    # so leave it a lone.  The last line is a size summary so
			    # leave it alone.  All other lines are files, so make the
			    # active links

			    /^ -/!{
				     $!{
					  s/^  *//1
					  s/  */${SEPCHAR}/g
					  s/.*/[x!]&/1
					  p;
				       }
				  }

			  }
		      "


	       ;;

	   *)
	       echo "|Error"
	       echo "Unknown command:  $1"
	       echo ""
	       echo "Arguments are:"
	       for f in "$@"
	       do
		   echo "$f"
	       done
	       ;;

       esac
   fi

}
##################################
##################################
handleExtractCommand()
{
    # we get here both before and after the popup that the
    # x command triggers and we have to interpret what case
    # we are in

   cmd="$1"

   shift 9
   file="$1"


    if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
    then
	#
	# This is the first invocation of the script in this
	# configuration (requesting the popup)
	#
	# Send the description of the popup to the script table viewer
	#

	echo "=%==%--POPUP=Extract $file;Output Filename;$file" |
	     sed -e 's/!/!/g'   \
		 -e 's/ /!20/g' \
		 -e 's/;/!3b/g' 
	exit 0
    else
	#
	#  This is a reinvocation of the script after a popup has
	#  run, now lets decide if it was aborted or was successful.
	#

	if [ "$POPUP_ABORTED" != "" ]
	then
	    STATE=final # no commands avaiable

	    printStateVariables
	    
            echo "|Error Message"
	    echo ""
	    echo "Extraction aborted by user"
	    exit
    
	else
	    STATE=final  # invalid date

	    TITLE="Extracting from $ARCHIVE"
	    
            printStateVariables
	    
            echo "|Copying"
	    echo "From $1"
	    echo "To   $POPUP_FIELD_1"

	    O="$POPUP_FIELD_1"  # output

	    case "$1" in

		*/)
		    #
		    #  unzipping to directory
		    #

                    set +x

		    ( 

		      # note:  $1 has a trailing / on it!


		      if [ ! -r "$O" ]
		      then
			  mkdir "$O"
		      fi

		      if [ -d "$O" ]
		      then
		      
			  if [ ! -w "$O" ]
			  then
			      echo ""
			      echo "[_!]ERROR:  cannot write to $O"
			      echo ""
			  else
			      unzip  -d "$O" "$ARCHIVE" "$1*"

      
			      if [ $? != 0 ]
			      then
				  echo ""
				  echo "[!_] ERROR: cannot extract, see above" 
			      fi

			  fi

		      else
			  echo ""
			  echo "[_!]ERROR:  $O is not a writable directory"
			  echo ""
		      
		      fi

		    ) 2>&1

		    ;;


		*)

		    #
		    #  unzipping a file
		    #

                    set +x
		    (
			unzip -vc "$ARCHIVE" "$1" | sed -e '1,2d' >"$O" 

			if [ $? != 0 ]
			then
			    echo ""
			    echo "[!_] ERROR: cannot extract, see above" 
			fi
		    ) 2>&1
		    ;;

	    esac



	    

	    exit
	fi
    fi
}

##################################
##################################
handleZipCommands()
{
   cmd="$1"

   shift 9
   file="$1"

   case "$cmd" in
       ^J)

	   #
	   #  The enter key means view the file
	   #

	   case "$file" in
	       *.csv)
                   TITLE="$file contents"
                   printStateVariables
		   unzip -c "$ARCHIVE" "$file" | sed -e '1,2d; 3s#.*#,&#1; '
		   ;;

               *.tsv)
                   TITLE="$file contents"
                   printStateVariables
		   unzip -c "$ARCHIVE" "$file" | sed -e "1,2d; 3s#.*#${TABCHAR}&#1; "
		   ;;

	       *)

                   TITLE="$file contents"
                   printStateVariables
                   echo "${SEPCHAR}Contents"
                   
                   unzip -c "$ARCHIVE" "$file" | sed -e '1,2d'
		   ;;
	   esac

	   ;;

       x)
	   handleExtractCommand "$@"
	   ;;


       *)
	   TITLE="$file:  command error"
	   echo "|Message"
	   echo ""
	   echo "Unexpected command $cmd, arguments are"
	   for f in "$@"
	   do
	       echo "$f"
	   done
        
   esac


}


# 
#  Command line arguments
#
#  If state is genroot, they should be nothing if we get here
#  In all other states, you should see something like this:
#
#    ^J row col maxrows maxcols currentCol currentRowContents
#


case "$STATE" in
    listDir)
	    handleDirCommands "$@"
	    ;;

    zipContents)
	    handleZipCommands "$@"
	    ;;

    *)
	cmd="$1"

	if [ "$cmd" = "x" -o "$cmd" = "X" ]
	then

	    # we get here both before and after the popup that the
	    # x command triggers and we have to interpret what case
	    # we are in

	    if [ "$POPUP_ABORTED" = "" -a "$POPUP_FIELD_1" = "" ]
	    then
		#
		# This is the first invocation of the script in this
		# configuration (requesting the popup)
		#
		# Send the description of the popup to the script table viewer
		#
		echo "=%==%--POPUP=Some Popup;Field 1;value1;Some other Field;Some other value." |
		     sed -e 's/!/!/g'   \
			 -e 's/ /!20/g' \
			 -e 's/;/!3b/g' 
		exit 0
	    else
		#
		#  This is a reinvocation of the script after a popup has
		#  run, now lets decide if it was aborted or was successful.
		#

		if [ "$POPUP_ABORTED" != "" ]
		then
		  
		    STATE=final  # state for the next invocation
		    printStateVariables
		    echo "|State = $STATE"
		    echo "YOU ABORTED!"
		    exit
            
		else
		    STATE=final  # state for the next invocation
		    printStateVariables
		    echo "|State = $STATE"
		    echo "You entered ($POPUP_FIELD_1) and ($POPUP_FIELD_2)"
		    exit
		fi
	    fi
	fi


	STATE=final  # state for the next invocation
	printStateVariables
	echo "|State = $STATE"
	echo "final state!"
	;;
esac

exit 0

