#ifndef MATH_ASCII_CHART_H
#define MATH_ASCII_CHART_H
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file cxxtls/ascii_chart.h
///  Defines ascii_chart a simple textual function plotting tool.

#include <utility>
#include <vector>
#include <iostream>
#include <string>

namespace cxxtls
{

struct ascii_chart 
  ///
  /// An ascii_chart is a block of characters which contains a picture of a 
  /// set of x,y pairs.  The idea is that you use the plot() methods to fill
  /// the chart with spaces and stars representing data set.  
  ///
  /// You can write the chart to an ostream or manually print it in some other way.
  ///
  /// The un-adorned data of the chart can be accessed
  /// using the public variable data_.  See member print_helper() to understand
  /// how to print the chart yourself.
  ///
  /// The method plot() allows you to convert an array of x,y pairs into
  /// a chart.  You can then print the rows of text yourself -- they are
  /// null terminated.  You must add needed annotations.  The plot method
  /// gives you the min and max range of the input data for this purpose.
  ///
  /// The variable title_ holds a string which will be displayed centered above
  /// the chart when using operator<<.  Annotations of the max and min of the
  /// x and y range will be produced by operator<<.
  ///
  /// Example:  @sa chart_test.c
  ///
{
  typedef double X_t; ///<  actual implementation representation of X's
  typedef double Y_t; ///<  representation of Y's

  typedef std::pair<X_t,Y_t>   point_t; ///<  actual representation of data points.
  typedef std::vector<point_t> data_t;  ///< actual representation of input data.


  int width_;      ///<  number of columns in each row in data_
  int height_;     ///<  number of rows found at data_;

  bool dirty_;     ///<  ie not clean.  set by plot, cleared by clean

  char **data_;    ///<  a variable length array of null terminated strings
		   ///  data_ contains height+1 rows of data.  The last
		   ///  pointer stored there is null so you can check for it
		   ///  when iterating if you like.  data_[0] refers to the
		   ///  highest Y value.  data_[height-1] is the lowest y value.

  char blank_;     ///<  char representing empty 
  char star_;      ///<  char representing data here

  std::pair<X_t,X_t> xrange_;  ///< A range of X values
  std::pair<Y_t,Y_t> yrange_;  ///< A range of Y values

  std::string title_;          ///< The title of this chart

  ascii_chart(int width, int height, char blank=' ', char star='*');
 ~ascii_chart();

  void clear();  ///<   set back to all blanks.


  void plot(data_t const       &data,
	    int                 xola=0  // ignore this parm
	   );
   ///<
   ///  This function creates an ascii plot of the data stored in the
   ///  input vector.  That is, it fills an 'ascii_chart' with spaces and
   ///  stars that represent the input data. 
   ///  
   ///  The data is scaled to fit into the ascii_chart's width and height.
   ///  No annotations are added.  Put them in yourself when you print the
   ///  chart.
   ///
   ///  @param data the data to plot
   ///  @param xola ignore this paramer -- its existence is internal
   ///  to the class' implementation details.
   ///
   
  template<class X, class Y>
  void plot(std::vector< std::pair<X,Y> > const &data)
   ///  Template version of the plot method to reduce the need for user's
   ///  copying their data into the standard form.
   ///
   ///  @param data the data to be plotted.
  {

    // This function copies the data into the standard plot required
    // form, calls plot, then translates the plot return data to the
    // requested form.

    data_t tmp(data.size());

    typedef std::vector< std::pair<X,Y> > input_t;

    data_t::iterator cur = tmp.begin();

    for(typename input_t::const_iterator i = data.begin(); i != data.end(); ++i, ++cur)
    {
      cur->first  = i->first;
      cur->second = i->second;
    }

    plot(tmp,1);  // call the non-template version

  }

  template<class Container>
  void plot(Container const &c) 
    ///  Given some general container, copy it to a vector then plot it.
    ///
    ///  @param c the container of x/y pairs to plot
  {
    data_t tmp(c.size());

    typename Container::const_iterator first = c.begin(), 
				       last  = c.end();

    data_t::iterator output = tmp.begin();

    while(first != last)
    {
      output->first = first->first;
      output->second= first->second;

      ++first;
      ++output;
    }

    plot(tmp);

  }


  void print_helper(std::ostream& stream) const;

  friend std::ostream& operator<< (std::ostream&s, ascii_chart const &a)
    {
      a.print_helper(s);
      return s;
    }


private:
  ascii_chart& operator=(ascii_chart const &);  //< not implemented
  ascii_chart(ascii_chart const &);             //< not implemented

  void range(data_t const &data);               ///<  compute ranges
};



} // namespace cxxtls
#endif
