#ifndef CXXTLS_ASCII_PLOT_H_INCLUDED_P
#define CXXTLS_ASCII_PLOT_H_INCLUDED_P
//
// Copyright 2012, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <cxxtls/foreach.h>
#include <math.h>

#include <vector>
#include <string>

namespace cxxtls
{

    //---------------------------------------------------------------------------------------------
    //---------------------------------------------------------------------------------------------
    //---------------------------------------------------------------------------------------------

    class AsciiPlotter
        ///  An object that holds a "plot" of zero to many sets of points.
        ///  The "plot" is a 2 dimensional array of characters which
        ///  can either be a space character (to indicate that the point is not
        ///  used) or some other character, aka "color", that indicates
        ///  that there is an active data point in that cell.
        ///
    {
        double minX_;       ///< adjusted left hand side of the displayable space
        double minY_;       ///< adjust bottom side of the displayable space
        double maxX_;       ///< user supplied right side of displayable space
        double maxY_;       ///< upser supplied top of the displayable space

        // note that width_ and height_ are 1 less than the
        // actual storage -- this allows the plot function to include
        // the very last data point.

        size_t   width_;    ///< adjusted char cell count (width) for display
        size_t   height_;   ///< adjusted char cell count (height) for display

        size_t   maxCol_;   ///< computed maximum addressed cell (width)
        size_t   maxRow_;   ///< computed maximum addressed cell (height)

        double rangeX_;     ///< range used for scaling calculations (at some point)
        double rangeY_;     ///< range used for scaling calculations (at some point)

        double maxRange_;   ///< the larger of rangeX_ and rangeY_

        bool scaling_;  ///< In plot, either scale to fit width_/height_ or truncate

        std::vector<std::string>  universe_;   ///< actual plotted points

    public:


         static double roundUpToIntegerMultiple(double value, double modulus)
             ///<  Round value up to an integer multiple of modulus.
             ///
             ///   The modulus must be positive. 
             ///
             ///   Examples
             ///
             ///   @li 1.7 rounds up to 2.0.
             ///   @li -1.7 rounds up to -1.0
             ///
         {
             if(value < 0)
             {
                value = 0 - value;

                value -= fmod(value,modulus);

                value = 0 - value;
             }
             else
             {
                double modval = fmod(value,modulus);

                if(modval)
                   value += modulus - modval;
             }

             return value;

         }

         static double roundDownToIntegerMultiple(double value, double modulus)
             ///<  Round value down to an integer multiple of modulus.
             ///
             ///   The modulus must be positive. 
             ///
             ///   Examples
             ///
             ///   @li 1.7 rounds down to 1.0
             ///   @li -1.7 rounds down to -2.0
             ///
         {
             if(value < 0)
             {
                value = 0 - value;

                double modval = fmod(value,modulus);

                if(modval)
                     value += modulus - modval;

                value = 0 - value;
             }
             else
             {
                value -= fmod(value,modulus);
             }

             return value;

         }

        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------

        void setUniverseSize()
            /// Populate the universe_ variable, which is a 2 dimensional array of characters,
            /// with blanks in the appropriate size -- which is one more row and one more column
            /// than is specified in the width_ and height_ variables, because of the need to
            /// plot the last point, but scale based on the (end - beginning) for floating
            /// point range reasons.
        {
           universe_.clear();

           size_t width = width_+1;
           size_t height= height_ +1;

           while(height--)
           {
               universe_.push_back(std::string());

               universe_.back().resize(width, ' ');

           }
        }

        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------

        void roundMinMax()
        {
           minX_ = roundDownToIntegerMultiple(minX_, 10);
           minY_ = roundDownToIntegerMultiple(minY_, 10);
           maxX_ = roundUpToIntegerMultiple(maxX_, 10);
           maxY_ = roundUpToIntegerMultiple(maxY_, 10);
        }

        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------

        AsciiPlotter( double minX, double minY, double maxX, double maxY, size_t width, size_t height)
            ///  Construct an AsciiPlotter that covers a given rectangular region of 2D space
            ///  and uses a particular width * height number of character cells to draw the plot.
            ///
            ///  This constructor is meant to give you exact control but is harder to use than
            ///  the other constructor(s).
            ///
        :  minX_(minX)
        ,  minY_(minY)
        ,  maxX_(maxX)
        ,  maxY_(maxY)
        ,  width_(width)
        ,  height_(height)
        ,  maxCol_(0)
        ,  maxRow_(0)
        ,  scaling_(false)
        {

           // force the minX_ and minY_ to be an even multiple of 10
           // to make the plotting easier because right now we don't 
           // scale during the plotting.

           // the "maxes" are ignored for the moment.


           roundMinMax();


           if(minX_ < minX)
           {
              width_ += minX - minX_;
              width = width_;
           }

           if(minY_ < minY )
           {
             height_ += minY - minY_;
             height = height_;
           }

           rangeX_ = maxX_ - minX_;
           rangeY_ = maxY_ - minY_;

           maxRange_ = rangeX_;

           if(rangeY_ > rangeX_)
              maxRange_ = rangeY_;


           setUniverseSize(); // one bigger than each of width_ and height_ (on purpose!)
        };


        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------


        void computeRanges()
        {
            rangeX_ = maxX_ - minX_;
            rangeY_ = maxY_ - minY_;
        
            maxRange_  = rangeX_ > rangeY_ ? rangeX_ : rangeY_;
           
        }


        AsciiPlotter(double minX, double minY, double maxX, double maxY, size_t pointCount)
            ///  An easy to use constructor that automates the scaling, and so forth.  It
            ///  requires that you precompute the rectangular region that holds all the
            ///  points of interest and then, based on the number of points to be plotted,
            ///  scales the plot to maximimize visibility combined with aspect ratio.
        {
            minX_ = minX;
            minY_ = minY;
            maxX_ = maxX;
            maxY_ = maxY;
        
            maxCol_ = 0;
            maxRow_ = 0;

            bool firstMajorPass = true;

            setParameters:
        
            computeRanges();
        
            double pointsPerUnit = pointCount / maxRange_;
        
            scaling_=false;
            
            size_t scaleTo=100;
        
            if(maxRange_ > 100)
            {
                scaling_ = true;   // we are scaling but lets go ahead and adjust the min/max to multiples of
                                   // of 10 to make the print out look prettier.  The error in accuracy is
                                   // 

                 roundMinMax();
                
                 if(firstMajorPass)
                 {
                    firstMajorPass = false;
                    goto setParameters;
                 }

            }
            else
            {
               if(maxRange_ <= 10)
               {
                   if(maxRange_ <= 1)
                   {
                      scaleTo = 10;
        
                      if(pointsPerUnit >= 3)
                      {
                         scaling_ = true; // scale up to a 10 wide graph to make room all those points..
                      }
                      else
                      {
                         // not scaling_, just plot the points in the 0-1 rectangle in the lhs of
                         // the 10 wide graph (which will be truncated in the output process anyway
                         // so you will really see just a 1 column result
                      }
                   }
                   else
                   {
                      // maxRange_ is 1+ to 10 
        
                      if(pointsPerUnit > 3)
                      {
                         scaling_ = true; // scale up to full width (100 pts)
                      }
                      else
                      {
                         // just plot the points on the left side of the rectangle unscaled

                         roundMinMax();

                         if(firstMajorPass)
                         {
                            firstMajorPass = false;
                            goto setParameters;
                         }


                      }
        
                   }
               }
               else
               {
                  // maxRange_ is between 10+ and 100
        
                  if(pointsPerUnit >= 3)
                     scaling_ = true;      // spread points across the whole 100 cells available
                  else
                  {
                      // leave scaling_ as false and scaleTo as 100 so the points will be on the left
 
                     roundMinMax();

                     if(firstMajorPass)
                     {
                        firstMajorPass = false;
                        goto setParameters;
                     }

                  }
               }
            }
        
        
            width_  = scaleTo;
            height_ = scaleTo;
        
            setUniverseSize();
        
        }




        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------

        template<class Container>
        void plot2D(Container const &points, char color)
        {
            typedef typename Container::value_type value_type;

            CXXTLS_FOREACH(value_type const &cur, points)
            {
               plot( cur.get(0), cur.get(1), color );
            }

        }

        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------


        void plot( double x, double y, char color)
            // place a color in a specific location in the plot
        {

           size_t col = x - minX_;
           
           size_t row = y - minY_;


           if(scaling_)
           {
               x -= minX_;
               y -= minY_;

               x *= width_;    // both are one less than the actual storage available
               y *= height_;

               x += 1 / width_;
               y += 1 / height_;

               x /= maxRange_;
               y /= maxRange_;

               col = x;
               row = y;

           }

           if(col <= width_ && row <= height_)    // universe_(height_) is a valid location!
           {
              universe_[row].replace(col, 1, 1, color);

              if(col > maxCol_)
                 maxCol_ = col;

              if(row > maxRow_)
                 maxRow_ = row;

           }

        }

        //----------------------------------------------------------------------------------------------
        //----------------------------------------------------------------------------------------------

        template<class Ostream>
        void plotHorizontalRuler(Ostream &s) const
        {
           s << "+";

           double colWidth = 1;

           if(scaling_)
           {
              colWidth = rangeX_ / width_;
           }


           for(size_t i = 0; i < maxCol_ +1; ++i)
           {
               if( i && ((i % 5) == 0) )
               {
                   if( maxRange_ >= 10 && (i % 10 == 0) ) 
                   {
                       if(scaling_)
                       {
                          s << '|';
                       }
                       else
                       {
                           s << char('0' + (size_t(i * colWidth +minX_)/10) % 10);
                       }


                   }
                   else
                   {
                       s << ':';
                   }
               }
               else
                  s << "-";
           }

           s << '+';  // do not \n in this function!
        }


        template<class Ostream>
        void output(Ostream &s ) const
        {

           if(scaling_)
           {

              s << "Cell[" << rangeX_ / width_ << "," << rangeY_ / height_ << "]\n";
           }


           s << "[" << minX_ << "," << maxY_ << "]\n";

           plotHorizontalRuler(s);

           s << '\n';


           for(size_t countRow = 0; countRow <= maxRow_; ++countRow)
           {
              size_t row = maxRow_ - countRow;

              char rowTag='?';

              if(row && ((row % 5) == 0) )
              {
                  double rowHeight = 1;

                  if(scaling_)
                  {
                      rowHeight = rangeY_ / height_;
                  }


                  if( maxRange_ >= 10 && (row % 10 == 0) ) 
                  {
                     if(scaling_)
                     {
                         rowTag = '-';
                     }
                     else
                     {
                         rowTag = char('0' + (size_t(minY_ +row * rowHeight)/10) % 10);
                     }
                  }
                  else
                  {
                     rowTag = ':';
                  }

              }
              else
              {
                  rowTag= '|';
              }

              s << rowTag;

              s.write(universe_[row].data(), maxCol_+1); //maxCol_ is not a count, it is the last


              s << rowTag ;

              if(countRow == 0)
              {
                  if(scaling_)
                  {
                      s << "[" << maxX_ << "," << maxY_ << "]";
                  }
                  else
                  {
                      s << "[" << minX_ + maxCol_ << "," << minY_ + maxRow_ << "]";
                  }

              }
              
              s << '\n';
           }

           plotHorizontalRuler(s);

           if(scaling_)
           {
               s << "[" << maxX_ << "," << minY_  << "]" << '\n';
           }
           else
           {
               s << "[" << minX_ + maxCol_ << "," << minY_  << "]" << '\n';
           }

           s << "[" << minX_ << "," << minY_ << "]\n";

           if(scaling_)
           {

              s << "Cell[" << rangeX_ / width_ << "," << rangeY_ / height_ << "]\n";
           }

        }


    };

    //-----------------------------------------------------------------------------------------------------
    //-----------------------------------------------------------------------------------------------------
    //-----------------------------------------------------------------------------------------------------

    template<class Ostream>
    Ostream &operator<<  (Ostream &s, AsciiPlotter const &plotter)
    {
       plotter.output(s);

       return s;
    }



}; // namespace cxxtls





#endif
