#ifndef _MSC_VER
namespace {
   int unused();
}

#else
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#include <cxxtls/CommandConsole.h>

#include <iostream>
#include <stdlib.h>
#include <iomanip>
#include <fstream>
#include <map>
#include <vector>



// windows code
#include <windows.h>
#include <tlhelp32.h>

#include <string>
#include <cxxtls/exception.h>
#include <cxxtls/file.h>
#include <cxxtls/foreach.h>

using namespace std;
using namespace cxxtls;


namespace cxxtls
{

namespace WindowsConsoleMethods
{

//-------------------------------------------------------------------------------------------------------
//    output routines to make using windws structures easier
//-------------------------------------------------------------------------------------------------------

template<class Stream>
Stream & operator<< (Stream &s, COORD const &c)
{
    return s << "{X(" << c.X << "),Y(" << c.Y << ")}";
}

inline bool operator != (COORD const &l, COORD const &r)
{
   return      l.X != r.X
          ||   l.Y != r.Y
          ;
}

template<class Stream>
Stream & operator<< (Stream &s, POINT const &c)
{
    return s << "{x(" << c.x << "),y(" << c.y << ")}";
}


template<class Stream>
Stream & operator<< (Stream &s, SMALL_RECT const &c)
{
    return s << "{L(" << c.Left << "),T(" << c.Top << "), R(" << c.Right << "),B(" << c.Bottom << ")}" ;
}

inline bool operator != (SMALL_RECT const &l, SMALL_RECT const &r)
{
    return      l.Top     != r.Top
           ||   l.Bottom  != r.Bottom
           ||   l.Left    != r.Left
           ||   l.Right   != r.Right
           ;
}


template<class Stream>
Stream & operator<< (Stream &s, CONSOLE_SCREEN_BUFFER_INFO const &c)
{
   return s << "dwSize=" << c.dwSize << ", srWindow=" << c.srWindow << ", curpos=" << c.dwCursorPosition;
}

//-------------------------------------------------------------------------------------------------------
//   Exceptions for dealing with console 
//-------------------------------------------------------------------------------------------------------

struct BadConsole
: public cxxtls::Exception
{
    BadConsole()
    :  cxxtls::Exception("BadConsole")
    {
    }
};

struct OutOfControlConsole
: public cxxtls::Exception
{
    OutOfControlConsole(char const *msg, char const *value)
    :  cxxtls::Exception("OfControlConsole", msg, value)
    {
    }
};

//-------------------------------------------------------------------------------------------------------
// The windows command console ipmlementation
//-------------------------------------------------------------------------------------------------------

struct TemporaryConsoleMode
{
    DWORD    console_mode_;
    HANDLE   input_handle_;

    TemporaryConsoleMode(DWORD newMode)
    {
        input_handle_ = GetStdHandle(STD_INPUT_HANDLE);

        GetConsoleMode(input_handle_, &console_mode_);

        SetConsoleMode(input_handle_, newMode);
    }

   ~TemporaryConsoleMode()
   {
        SetConsoleMode(input_handle_, console_mode_);
   }


};


class ConsoleIF
:  public CommandConsole
    /// The Microsoft Windows console implementation of the CommandConsole class.
{

    CONSOLE_SCREEN_BUFFER_INFO original_csbi_;
    CONSOLE_SCREEN_BUFFER_INFO working_csbi_;

    HANDLE handle_;  // output handle

    HANDLE input_;


public:

    int getMaxX() const
    {
       return working_csbi_.dwSize.X;
    }

    int getMaxY() const
    {
       return working_csbi_.dwSize.Y;
    }


    void computeWorkingCSBI( CONSOLE_SCREEN_BUFFER_INFO &working, CONSOLE_SCREEN_BUFFER_INFO &original)
        // if at all possible, normalize the working csbi's srWindow to the
        // bottom left of the total screen buffer size.
    {
        memcpy( &working, &original, sizeof(working));

        int visibleWidth  = working.srWindow.Right  - working.srWindow.Left + 1;
        int visibleHeight = working.srWindow.Bottom - working.srWindow.Top + 1;

        working.srWindow.Left = 0;
                                                              
        if(working.dwCursorPosition.X - working.srWindow.Left >= visibleWidth)
        {
            working.srWindow.Left = working.dwCursorPosition.X - visibleWidth / 2;

            if( working.srWindow.Left + visibleWidth >=  working.dwSize.X )
            {
                working.srWindow.Left = working.dwCursorPosition.X - visibleWidth;

            }
        }

        working.srWindow.Right = working.srWindow.Left + visibleWidth - 1;

        working.srWindow.Top = working.dwSize.Y - visibleHeight;

        if(working.dwCursorPosition.Y < working.srWindow.Top)
        {
            working.srWindow.Top = working.dwCursorPosition.Y - visibleHeight / 2;

            if( int(working.srWindow.Top) < 0)
            {
                working.srWindow.Top = 0;
            }
        }

        working.srWindow.Bottom = working.srWindow.Top + visibleHeight - 1;


    }


    ConsoleIF(std::string const &historyFile, std::vector<std::string> &historyLines)
    :  CommandConsole(historyFile, historyLines)
    {
        memset(&original_csbi_, 0, sizeof(original_csbi_));
        memset(&working_csbi_, 0, sizeof(working_csbi_));

        
        input_  = GetStdHandle(STD_INPUT_HANDLE);
        handle_ = GetStdHandle(STD_OUTPUT_HANDLE);

        if(!handle_)
           throw BadConsole();

        if(!GetConsoleScreenBufferInfo (handle_, &original_csbi_))
        {
           throw BadConsole();
        }

        computeWorkingCSBI( working_csbi_, original_csbi_);

        SetConsoleScreenBufferSize (handle_, working_csbi_.dwSize);
        SetConsoleWindowInfo       (handle_, true, &working_csbi_.srWindow);
        SetConsoleCursorPosition   (handle_, working_csbi_.dwCursorPosition);


    }

   ~ConsoleIF()
   {
       SetConsoleScreenBufferSize (handle_, original_csbi_.dwSize);
       SetConsoleWindowInfo       (handle_, true, &original_csbi_.srWindow);
       SetConsoleCursorPosition   (handle_, original_csbi_.dwCursorPosition);
       
   }

   virtual bool readline(std::string &input, int &number, std::string const &directory);

   void killChildren();  // kill child processes of the current process

public:



   bool kbhit()
   {
      INPUT_RECORD tmp;

      DWORD count = 0;

      PeekConsoleInput(input_,
                       &tmp,
                       1,
                       &count
                      );

      if(count == 0)
         Sleep(100);  // sleep 1/10 of second if no data -- this keeps us from burning up the cpu

      return count != 0;
      

   }

   int readKey()   // including up, down, BS, tab, home, end, ^A, ^E, page up, page down, etc.
   {
      INPUT_RECORD tmp;

      DWORD count = 0;

      TemporaryConsoleMode old_mode(0);  // set new mode to 0 (aka raw mode) temporarily

      memset(&tmp,0,sizeof(tmp));  // make sure all bits in tmp are properly 0

       for(;;)
       {
           {    
                CONSOLE_SCREEN_BUFFER_INFO new_csbi;

                memset(&new_csbi, 0, sizeof(new_csbi));

                if(!GetConsoleScreenBufferInfo (handle_, &new_csbi))
                {
                   throw BadConsole();
                }

                if(    new_csbi.dwSize   != working_csbi_.dwSize  )   // note:  do not include srWindow in this test or you get infinite resize loops...
                {
                    computeWorkingCSBI( working_csbi_, new_csbi);
        
                    SetConsoleScreenBufferSize (handle_, working_csbi_.dwSize);
                    SetConsoleWindowInfo       (handle_, true, &working_csbi_.srWindow);

                    COORD newpos = getCursorPosition();

                    if(!GetConsoleScreenBufferInfo (handle_, &new_csbi))
                    {
                       throw BadConsole();
                    }

                    return SCREEN_RESIZED;

                }


           }
           
           
           
           if(!kbhit())    // sleeps to avoid excess cpu utilization
              continue; 

           if(!ReadConsoleInput(input_, &tmp, 1, &count))
              return -1;  // end of file
           
           if(tmp.Event.KeyEvent.bKeyDown
              && tmp.Event.KeyEvent.wVirtualKeyCode != 0
             )    // eat the down stroke of the key
              break;

       }

       if(tmp.Event.KeyEvent.uChar.AsciiChar)
           return tmp.Event.KeyEvent.uChar.AsciiChar;

       return 0x100 + tmp.Event.KeyEvent.wVirtualKeyCode;  
       
   }


   COORD  getCursorPosition()
   {
       CONSOLE_SCREEN_BUFFER_INFO csbi;

       memset(&csbi, 0, sizeof(csbi));

       GetConsoleScreenBufferInfo(handle_, &csbi);

       COORD rv;

       rv.X = csbi.dwCursorPosition.X;
       rv.Y = csbi.dwCursorPosition.Y;

       return rv;
                    

   }

   void   setCursorPosition(COORD c)
   {
       working_csbi_.dwCursorPosition.X = c.X;
       working_csbi_.dwCursorPosition.Y = c.Y;

       SetConsoleCursorPosition(handle_, c);


   }

   inline void setCursorPosition(int row, int col)
   {
       COORD c;

       c.X = col;
       c.Y = row;

       setCursorPosition(c);
   }

   size_t print(char c)
       /// print the character c at the current cursor position in a forward going
       /// manner.  Move the cursor to the right.
       ///
       /// @return If the cursor is supposed to wrap to the next line on the screen, but doesn't
       /// because we are at the bottom of the scrolling region, return 0.
       // Missed scrolls only happen when you are the bottom line of the screen at the bottom of the scrolling region.
   {
       DWORD bytesWritten;

       CONSOLE_SCREEN_BUFFER_INFO csbi;

       memset(&csbi, 0, sizeof(csbi));

       DWORD rc;

       rc = GetConsoleScreenBufferInfo(handle_, &csbi);

       if(rc == 0)
          return 0;  // we must be having write errors

       rc = WriteConsole(handle_, &c, 1, &bytesWritten, 0);

       if(rc 
          && (csbi.dwCursorPosition.Y == (csbi.dwSize.Y-1))
          && (csbi.dwCursorPosition.X == (csbi.dwSize.X-1))
         )
       {
          // we got a successful write and we were on the last character of the scroll region

          CONSOLE_SCREEN_BUFFER_INFO newcsbi;
          
          memset(&newcsbi, 0, sizeof(newcsbi));

          GetConsoleScreenBufferInfo(handle_, &newcsbi);  // if the other one worked, why would this fail

          if(newcsbi.dwCursorPosition.Y == csbi.dwCursorPosition.Y)
             return 1;  // we should have scrolled but did not.



       }

       return 0;

   }

   size_t print(std::string const &s)
       // print the specified string and keep count of how many times we should have scrolled but did not.
       // Missed scrolls only happen when you are the bottom line of the screen at the bottom of the scrolling region.
   {
      size_t countOfMissedScrolls = 0;

      for(size_t i = 0; i < s.size(); ++i)
          countOfMissedScrolls += print(s[i]);  // call so slow that no optimization is valuable

      return countOfMissedScrolls;
   }

   template<class Iterator>
   size_t print(Iterator f, Iterator l)
       // Print the characters in the specified range and return the count of scrolls that did not occur.
       // Missed scrolls only happen when you are the bottom line of the screen at the bottom of the scrolling region.
   {
      size_t countOfMissedScrolls = 0;

       while(f != l)
          countOfMissedScrolls = print(*f++);

      return countOfMissedScrolls;
   }

   void clearToEndOfLine()
   {
     COORD curpos = getCursorPosition();

     int maxX = getMaxX();

     int bytesToWrite = maxX - curpos.X;

     DWORD Garbage = 0;

     FillConsoleOutputCharacterA(handle_, ' ', bytesToWrite, curpos, &Garbage);

     setCursorPosition(curpos);

   }


   void clearToEndOfScreen()
   {
     COORD curpos = getCursorPosition();

     COORD orginal = curpos;

     clearToEndOfLine();

     while(int(curpos.Y) < getMaxY()-1)
     {
         ++curpos.Y;
         curpos.X = 0;
         setCursorPosition(curpos);
         clearToEndOfLine();
     }

     SetConsoleWindowInfo(handle_, true, &working_csbi_.srWindow);

     setCursorPosition(orginal);


   }


}; // ConsoleIF

}  // namespace WindowsConsoleMethods


using namespace WindowsConsoleMethods;

bool 
WindowsConsoleMethods::
ConsoleIF::
readline(std::string &input, int &number, std::string const &directory)     
   {

       COORD startPos = getCursorPosition();

       COORD curPos;

       bool firstTimeThrough = true;

       size_t charOffset = 0;

       size_t historyLine = history_.size();

       // begin variables used to handle tabbing to get file names from a directory

                 string pattern;

                 list<string>  matchList;  // matching file names

                 string prefix;  // part of the command line before the matching region
                 string suffix;  // part of the command line after the matching region

                 size_t firstTryCharOffset=0;

                 char leaderCharacter=' ';  // the character before the first word of the tab region,
                                            // some character earlier than firstTryOffset character

       // end variables used to handle tabbing to get file names from a directory

   int pushedBackKey = 0;              // a key to interpret instead of reading one from the
                                       // console.  When 0, read from the console

   // get these values for safety

   curPos = getCursorPosition();  // edit cursor position, within line is at the end on the first time through
   charOffset = input.size();

   int tabKeyPushes = 0;


   top:                                   // goto here when you want to repain the whole line

       setCursorPosition(startPos);

       size_t missedScrolls = print(input);

       if(missedScrolls)
       {
           // handle the possibility that we are at the bottom of the screen and of the scrolling region
           // and that we have not actually scrolled to a new line like we should have -- because there are
           // no more new lines to scroll too.

           if(startPos.Y >= missedScrolls)
           {
              startPos.Y -= missedScrolls;

              setCursorPosition(startPos);

              missedScrolls = print(input);

           }
           else
              startPos.Y = 0;
       }

       if(firstTimeThrough)
       {
           curPos = getCursorPosition();  // edit cursor position, within line is at the end on the first time through
           charOffset = input.size();

           firstTimeThrough = false;
       }

       clearToEndOfLine();

       int key;

       setCursorPosition(curPos);

       bool done = false;





       while(!done)                      // loop here while not done editing
       {

           int key;

           // read a new key from the console or use an existing key that has been pushed-back.
           
           if(pushedBackKey)
           {
              key = pushedBackKey;

              pushedBackKey = 0;
           }
           else
               key = readKey();

           if (key != '\t')
           {
              tabKeyPushes = 0;         // reset count of times the tab key pushed itself back
              matchList.clear();        // all keys but the tab key erase the tabbing list
           }
           switch(key)
           {  
               case SCREEN_RESIZED:
               {
                  // user resized the window
                  firstTimeThrough = true;
                  goto top;
               }
               break;
               case 0x03:   // ^C
               {
                  print("^C\n");
                  killChildren();

                  input.erase(input.begin(), input.end());
    
                  charOffset = 0;
                  curPos = startPos;
                  goto top;


               }
               break;

               case ENDFILE:  
               {
               }
               return false;
    
               case HOM:
               case 0x01:  // ^A
               {
                   charOffset = 0;
                   curPos = startPos;
                   goto top;
               }
               break;
    
               case END:
               case 0x05:  // ^E
               {
                   firstTimeThrough = true;    // let top handling record charOffset and curpos
                   goto top;
               }
               break;
    
               case RT:  // right arrow
               {
                  if(input.size() && charOffset < input.size() -1 )
                  {
                     ++charOffset;
                     ++curPos.X;
    
                     if(curPos.X >= getMaxX())
                     {
                        ++curPos.Y;
                        curPos.X = 0;
                     }
    
                     setCursorPosition(curPos);
                  }
                  else
                  {
                     firstTimeThrough = true;
                     goto top;
                  }
               }
               break;

               case CTW: // control w jump to next word
               {
                  if(input.size() && charOffset < input.size() -1 )
                  {
                      // we are not at the end of the line -- scan through
                      // the current word and then search for the begining
                      // of the next word

                      while(charOffset < input.size() && isalnum(input[charOffset]) )
                      {
                         ++charOffset;
                         ++curPos.X;
                      
                         if(curPos.X >= getMaxX())
                         {
                            ++curPos.Y;
                            curPos.X = 0;
                         }
                      
                         setCursorPosition(curPos);
                      }

                      while(charOffset < input.size() && !isalnum(input[charOffset]) )
                      {
                         ++charOffset;
                         ++curPos.X;
                      
                         if(curPos.X >= getMaxX())
                         {
                            ++curPos.Y;
                            curPos.X = 0;
                         }
                      
                         setCursorPosition(curPos);
                      }

                  }
                  else
                  {
                     firstTimeThrough = true;
                     goto top; // repaint the whole line, leaving
                               // the cursor after the last character in the line
                  }
               }
               break;
    
               case LT:  // left arrow
               {
                  if(input.size() && charOffset )
                  {
                     --charOffset;
    
                     if(curPos.X)
                         --curPos.X;
                     else
                     {
                        curPos.X = getMaxX()-1;
                        --curPos.Y;
                     }
    
                     setCursorPosition(curPos);
                  }
               }
               break;

               case CTQ:  // control Q jump to prev word
               {
                  while(input.size() && charOffset && isalnum(input[charOffset]) )
                  {
                     --charOffset;
    
                     if(curPos.X)
                         --curPos.X;
                     else
                     {
                        curPos.X = getMaxX()-1;
                        --curPos.Y;
                     }
    
                     setCursorPosition(curPos);
                  }

                  while(input.size() && charOffset && !isalnum(input[charOffset]) )
                  {
                     --charOffset;
    
                     if(curPos.X)
                         --curPos.X;
                     else
                     {
                        curPos.X = getMaxX()-1;
                        --curPos.Y;
                     }
    
                     setCursorPosition(curPos);
                  }

                  while(input.size() && charOffset && isalnum(input[charOffset]) )
                  {
                     --charOffset;
    
                     if(curPos.X)
                         --curPos.X;
                     else
                     {
                        curPos.X = getMaxX()-1;
                        --curPos.Y;
                     }
    
                     setCursorPosition(curPos);
                  }

                  if(input.size() && charOffset < input.size() && !isalnum(input[charOffset]))
                  {
                     ++charOffset;
                     ++curPos.X;
    
                     if(curPos.X >= getMaxX())
                     {
                        ++curPos.Y;
                        curPos.X = 0;
                     }
    
                     setCursorPosition(curPos);
                  }



               }
               break;

    
               case BS:  // backspace
               {
                  if(input.size() && charOffset )
                  {
                     --charOffset;
    
                     if(curPos.X)
                         --curPos.X;
                     else
                     {
                        curPos.X = getMaxX()-1;
                        --curPos.Y;
                     }
    
                     setCursorPosition(curPos);
    
                     input.erase(input.begin() + charOffset);
    
                     print(input.begin() + charOffset, input.end());
    
                     clearToEndOfLine();
                     setCursorPosition(curPos);
    
                  }
               }
               break;
    
               case DEL:  // delete
               {
                  if(input.size() && charOffset < input.size() )
                  {
                     input.erase(input.begin() + charOffset);
    
                     print(input.begin() + charOffset, input.end());
    
                     clearToEndOfLine();
                     setCursorPosition(curPos);
    
                  }
               }
               break;

               case CTK: // control k
               {
                  if(input.size() && charOffset < input.size() )
                  {
                     input.erase(input.begin() + charOffset, input.end());
    
                     clearToEndOfLine();
                     setCursorPosition(curPos);
    
                  }
               }
               break;

    
               case '\t':  // tab
               {
                  // interpret the tab character as the command completion char
                  //
                  // In the process, assume that the user is typing a valid ms dos file name but
                  // using the brash language string rules.  That is, in ms dos, \ is not an 
                  // escape character bur rather is the directory path separator.  In brash,
                  // a bourne shell clone, \ is an escape separator.  So we have to do our best
                  // to bridge between the two worlds.  To do so, we are going to make assumptions
                  // about what the user is trying to do.  The primary assumption is that he
                  // would not be pressing the tab key unless he was trying to do tab-completion
                  // of file names.  To facilitate this, here are some rules and guidlines about
                  // the following code:
                  //
                  //   1)  The input and output string must have \\ whenever a filename has a \
                  //       in it
                  //
                  //   2)  But the find_matching function does not know how to deal with double \\
                  //       and does not know how to deal with "\ " (that is, a backslash before a
                  //       space (or any other character).
                  //   
                  //   3)  So the input data must be kept in the brash expected string syntax, but
                  //       local copies of the input string have to be filtered to remove the
                  //       escape-usage of the \ character.  So while the user might see this:
                  //
                  //         c:\\Program\ Files\Microsoft ...
                  //
                  //       The internal copy (particularly in the pattern variable, below) will look
                  //       like this:
                  //
                  //         c:\Program Files\Micrsoft
                  //
                  //   4)  But note that this code does not implement a generalized escape character
                  //       handling usage of the \ character.  It is only the "\ " (slash-space)
                  //       character combination that is supported.  I mean really, do you honestly
                  //       expect to be non-printing characters in file names?  What kind of 
                  //       disastrous effects would this cause for anyone reading a directory listing?
                  //       Its basically an insane idea, and I will leave getting that to work to 
                  //       someone else.  DEC-10 systems used to let you do it, but in my youth I used
                  //       feature mainly to mess with the people, not do anything useful...


                  bool matchListStartedEmpty = matchList.empty();  // tab pressed after some other kind of character
    
                  if (!matchListStartedEmpty)
                      charOffset = firstTryCharOffset;
                  else
                      firstTryCharOffset = charOffset;

                  string lastWord;
    
                  string::const_iterator cur = input.begin() + charOffset;
                  string::const_iterator start =cur;

                  size_t savedCharOffset = charOffset;
    
                  bool foundForwardSlash = false;

                  while(cur != input.begin())
                  {  // search backwards for the begining of the current filepath
                     // 
                     // search backwards until we find an un-escaped space or 
                     // one of the brash string leader characters:
                     //
                     //    "
                     //    '
                     //    `
                     //    ( -- as in $(...)
                  
                     // move backwards from the current cursor position (in the input line)
                     // to find the start of the current pathname fragment

                     --cur;
    
                     char c = *cur;

                     // if we get to space, handle the many confusions associated with its interpretation.
                     // In particular, if it is an escaped space, don't treat it like a real space, just
                     // skip over it.

                     if(c == ' ' && cur != input.begin())
                     {
                        // check for escape sequences in the name of the current file or directory path

                        string::const_iterator prev = cur;

                        --prev;

                        if(*prev == '\\')
                        {
                           // hopefully we have an escape sequence protecting the current character
                           // which is a space

                          

                           if(prev != input.begin())
                           {
                              --prev;

                              if(*prev == '\\')
                              {
                                 // *cur is a space, *(cur-1) is a \, and (*cur-2) is also a \\
                                 // so the \ is not escaping the  space, so stop parsing
                                 // backwards -- that is, do not skip over the space but
                                 // let the code below treat it as the "leading character"
                                 // of a tabable region
                                  foundForwardSlash = true; 
                              }
                              else
                              {
                                 // we do not have "\\ ", but only "\ ", so we do have an escaped
                                 // space.

                                 --charOffset;  // sync up to the space
                                 --cur;         // sync up to the \ before the space
                                 --charOffset; // 
                                 continue;     // searching backwards
                              }
                           }
                           else
                           {
                               // we did find "\ ", and \ is at the beginning, so
                               // prev points to begining, so the leading space is not a special case
                               cur = input.begin();
                               charOffset = 0;
                               continue;  // which drops us out of the loop
                           }

                        }

                     } // handle protected spaces in the file name

                     // now that we have determined we are not treating the currrent character as a 
                     // space, lets handle some other situations -- for example, do we have
                     // either / or \ in the name.  If we do, then we will need to convert them
                     // to the normalized form.  In the normal form, we are going to make our
                     // pathnames look like this:
                     //
                     //   c:\\Program\ Files\\Microsoft
                     //
                     // So if we find a / or \ character that is not used in exactly this way,
                     // then we will need to trigger a re-structuring of the input string to meet
                     // this proper definition.  When this occurs, we will repaint the string,
                     // correctly.  The variable, foundForwardSlash, should really be named "foundSlash"
                     // because both types trigger the restructuring -- assuming the structure is
                     // wrong.  So the following code checks for the surrounding characters when a \ 
                     // is found.  If we already have the right structure, no reason to trigger the
                     // restructuring.

                     if(c == '/')
                         foundForwardSlash = true;
                     else
                     if (c == '\\')
                     {
                         string::const_iterator next = cur; 
                         if (next != start)
                         {
                             ++next;
                             if (*next != '\\' && *next != ' ')
                             {
                                 string::const_iterator prev = cur;

                                 if (cur == input.begin())
                                     foundForwardSlash=true;
                                 else
                                 {
                                     --prev;
                                     if (*prev != '\\' && leaderCharacter != '\'')
                                         foundForwardSlash = true;
                                 }
                             }
                         }
                     }

    
                     if(isspace(c) || c == '\'' || c == '"' || c == '(' || c == '`' || c == '>' || c == '<')
                     {
                        leaderCharacter = c;
                        ++cur;
                        break;
                     }
                     else
                        --charOffset;
                  }  // search backwards for the begining of the current filepath


                  if(charOffset < input.size() && input[charOffset] == '~')
                  {
                      // handle ~ expansion

                      char const *home = ::getenv("USERPROFILE");

                      if(home)
                      {
                          input.erase(charOffset,1);

                          input.insert(charOffset, home);

                          charOffset = savedCharOffset + strlen(home) - 1;

                          firstTimeThrough = true;
                          pushedBackKey = '\t';
                          goto top;

                      }

                  }

    
                  #ifdef _MSC_VER
                  if(matchListStartedEmpty && foundForwardSlash)
                  {
                     // we need to redraw the output with the / characters converted into \\ characters

                     string tmp;

                     size_t count = 0;

                     string::const_iterator f = input.begin(),
                                            l = input.end();
                     
                     while(f != l)
                     {
                         char c = *f++;

                         size_t curCharOffset = f - input.begin();

                         if(curCharOffset < charOffset)
                         {
                            // we are in a region that is before the start of the tab area.
                            tmp += c;
                            continue;  
                         }
                         else
                         if(curCharOffset >= savedCharOffset)
                         {
                            // just copy the rest of the data to tmp and quit
                            // because we only want to change the characters
                            // in the tab region.

                             if (c == '/')
                             {
                                 if (leaderCharacter != '\'')
                                     tmp += "\\\\";  // converting / into \\\\ as we go in unprotected regions
                                 else
                                     tmp += '\\';     // in single quoted strings, just use the one
                             }
                             else
                                tmp += c; // we  have already consumed the current character from the stream

                            while(f != l)
                                tmp += *f++;

                            break;
                         }
                         

                         if(c == '/')
                         {
                             foundForwardSlash = false;      // we'll fix them all in this cycle

                             if (leaderCharacter == '\'')
                                 tmp += '\\';
                             else
                                 tmp += "\\\\";
                         }
                         else
                         if (c == '\\' && leaderCharacter != '\'')
                         {
                             if (f == l)
                                tmp += c;
                             else
                             {
                                 if (*f == '\\')
                                 {
                                    tmp += c;
                                    tmp += c;
                                    ++f;   // output the pair of slashes
                                 }
                                 else
                                 if (*f == ' ')
                                 {
                                     tmp += c;
                                     tmp += ' ';
                                     ++f;    // output the /-space
                                 }
                                 else
                                 {
                                      tmp += c;
                                      tmp += c;
                                      // do not increment f, disallow all other escape sequences
                                 }
                             }
                         }
                         else
                            tmp += c;
                     }

                     input = tmp;
                     firstTimeThrough = true;

                     // about to push a tab key back into the input for re-interpretation, but we  don't want to get into an
                     // infinite loop of re-pushing the same key over and over back into the stream.
                     //
                     // This only happens in a couple of cases, so lets just check for them here and not
                     // allow infinite loops.

                     char lastInputChar = input.empty() ? ' ' : input[input.size()-1];

                     if (lastInputChar == '/')
                         lastInputChar = '\\';

                     if (tabKeyPushes < 3)
                     {
                         ++tabKeyPushes;
                         pushedBackKey = '\t';
                         goto top;
                     }
                     else
                     {
                        tabKeyPushes = 0;
                     }
                        
                  }
                  #endif


                  while(cur != start)
                  {
                     #ifdef _MSC_VER
                         char c = *cur++;

                         if(c == '/')
                            c = '\\';      // on windows, normalize to \ despite the trouble this will cause.  File names passed to 
                                            // windows commands need to have \, so the user needs to learn to use quotes around the
                                            // the file names.  sorry, its MS's fault, not mine

                        lastWord += c;
                     #else
                         lastWord += *cur++;
                     #endif
                  }
    
                  if(lastWord.size())
                  {
                     bool matchListWasEmpty = matchList.empty();
    
                     // there is a last word
    
                     if(matchListWasEmpty)
                     {
                         // this is the first time that the tab key has been pushed in this sequence of
                         // tab key presses so lets read the directory and store the results in matchList
                         // and other variales mentioned above.
                         
                         FileName tmp(lastWord);

                         if(tmp.is_absolute_path())
                         {
                             pattern = lastWord + "*";
                         }
                         else
                         {
                             #ifdef _MSC_VER
                             pattern = directory + "\\"  +lastWord + "*";
                             #else
                             pattern = directory + "/"  +lastWord + "*";
                             #endif
                         }

                         FileName::sorted_names_t  matches; // matching file names


                         {  // filter duplicate directory separators in the pattern

                             char const *scan = pattern.data();
                             char const *end = scan + pattern.size();

                             string tmp;

                             while (scan != end)
                             {
                                 char c = *scan++;

                                 if (c == '\\')
                                 {
                                     if (scan != end && *scan != '\\')
                                     {
                                         if (*scan == ' ')
                                         {
                                             tmp += ' ';
                                             ++scan;
                                             continue;
                                         }
                                         tmp += c;
                                         continue;
                                     }
                                     else
                                     {
                                         // filtering duplicates
                                         tmp += '\\';

                                         while (scan != end && *scan == '\\')
                                             ++scan;  // eliminate duplicates
                                     }
                                 }
                                 else
                                 if (c == '/')
                                 {
                                     tmp += c;

                                     while (scan != end && *scan == '/')
                                         ++scan;
                                 }
                                 else
                                     tmp += c;
                             }

                             pattern = tmp;


                         }

                         
                         FileName::find_matching(pattern.c_str(), &matches, 0x10);   // case insensitive comparison
                         
                         if (matches.empty())
                         {
                            charOffset = savedCharOffset;
                            break;   // nothing to do
                         }
                         
                         // convert the matching "set" into a list for the purposes
                         // of tabbing through the list
                         
                         CXXTLS_FOREACH(string const &cur, matches)
                         {
                             if( cur.size() == 1 && cur[0] == '.')
                                continue;


                             if( cur.size() == 2 && cur[0] == '.' && cur[1] == '.')
                                continue;

                             matchList.push_back(cur);
                         }

                         prefix = input.substr(0,charOffset);
                         suffix = input.substr(charOffset+lastWord.size());

                     }

                     // at this point, the matchList variable is populated with the
                     // list of matches and the prefix and suffix variables are
                     // populated with the text before and after the matching
                     // region.
    
                     // now cycle through the list of matches.  Each time
                     // the tab key is pressed with no other intervening keys
                     // we will rotate the matchList contents so that the first
                     // on is always the next string to display to the user

                     string matchString = *matchList.begin();
    
                     matchList.erase( matchList.begin() );
                     matchList.push_back(matchString);
    

                     if(leaderCharacter != '\'' && leaderCharacter != '"')
                     {  // put escape characters in the names of files returned from the
                        // directory when the file has spaces or other annoying characters

                         string tmp;

                         CXXTLS_FOREACH(char c, matchString)
                         {
                            #ifdef _MSC_VER
                            if(c == ' ' || c == '(' || c == ')' )
                               tmp += '\\';
                            #else
                            if(c == ' ')
                               tmp += '\\';
                            #endif

                            tmp += c;
                         }

                         matchString = tmp;
                     }

    
                     // insert matchString into the output

                     size_t dirSep = lastWord.find_last_of("/\\");

                     if (lastWord.size() > dirSep && lastWord[dirSep + 1] == ' ')
                         dirSep = lastWord.size();

                     if(dirSep < lastWord.size())
                     {
                         lastWord = lastWord.substr(0, dirSep);

                         // due to the use of the tab completion characters as file names
                         // we really need to use the correct directory separator character

                         #ifdef _MSC_VER
                             static string dirSepChar("\\");
                         #else
                             static string dirSepChar("/");
                         #endif

                         if (matchListStartedEmpty)
                         {
                             if (input.size() > 1
                                 && input[input.size() - 1] == '\\'
                                 && input[input.size() - 2] != '\\'
                                 )
                             {
                                 lastWord += '\\';     // really need two trailing backslashes
                             }

                             input = prefix + lastWord + dirSepChar + matchString + suffix;

                             prefix = prefix + lastWord + dirSepChar;
                         }                             
                         else
                             input = prefix + matchString + suffix;
                     }
                     else
                     {
                         input = prefix + matchString + suffix;
                     }
    
                     // display everything up to the point that the
                     // user pressed tab -- minus the length of the tab
                     // word that is the trigger for the tab command

                     setCursorPosition(startPos);

                     print(prefix);

                     // now print the current matching string
                     
                     print(matchString.begin(), matchString.end());
    
                     // save the cursor position for editing purposes

                     curPos = getCursorPosition();

                     // now print the rest of the current command line
    
                     print(suffix.begin(), suffix.end());

                     clearToEndOfLine();
    
                     // now go back to the end of the first tab position
                     // so the user can either accept or press tab again

                     setCursorPosition(curPos);
    
                     charOffset = prefix.size()  + matchString.size();
    
    
    
                  }  // last word on the line before the tab key press was NOT empty
    
    
               }        // tab
               break;

               case UP: // up arrow
               {
                  if(historyLine)
                  {
                     --historyLine;
                     firstTimeThrough=true;
                     input = history_[historyLine];
                     goto top;
                  }
               }
               break;

               case DN:  // down arrow
               {
                  if(   !history_.empty() 
                     && (historyLine < history_.size() -1 ) 
                    )
                  {
                     ++historyLine;
                     firstTimeThrough=true;
                     input = history_[historyLine];
                     goto top;
                  }
                  else
                  {
                     input = "";
                     firstTimeThrough = true;
                     goto top;
                  }
               }
               break;

    
    
    
               case '\n':  // enter or newline
               case '\r':  // carriage return
                  done= true;
               break;
    
               default:
                   if(key < 0x80 && key >= ' ')
                   {
                        if(charOffset < input.size())
                           input.insert(input.begin() + charOffset, char(key));
                       else
                           input += char(key);
    
                       int missedScrolls = print(input.begin() + charOffset, input.end());

                       if(missedScrolls == 0)
                       {
                           // we did not scroll off the bottom of the window
                           
                           ++charOffset;
                           ++curPos.X;
                           
                           if(curPos.X >= getMaxX())
                           {
                              ++curPos.Y;
                              curPos.X = 0;
                           
                              if(curPos.Y >= getMaxY())
                              {
                                  if(startPos.Y == 0)
                                  {
                                        startPos.X = 0;
                                  }
                                  else
                                      --startPos.Y;
                           
                           
                                  firstTimeThrough = true;
                                  goto top;   // repaint the whole damn thing
                              }
                           
                           }
                           
                           setCursorPosition(curPos);
                       }
                       else
                       {
                           // we did scroll off the bottom of the console window scroll region.

                           // make sure we have room to paint the whole input line

                           ++charOffset; // to account for the character we just added
                           ++curPos.X;  

                           if(startPos.Y < (missedScrolls ))
                           {
                               // either small screen or gigantic input line

                               startPos.Y = 0;
                               curPos.Y   = 0;
                           }
                           else
                           {
                               startPos.Y -= missedScrolls ;
                               curPos.Y   -= missedScrolls ;
                           }

                           // since we can't easily get the text of the command prompt back,
                           // lets handle this rare situation by positioning the cursor not at the
                           // end of the command prompt, as normal, but at the beginning of the row.


                           DWORD x = startPos.X;

                           startPos.X = 0;

                           if(curPos.X >= x)
                              curPos.X -= x;
                           else
                           {
                               // should not get here, but who knows
                               //
                               // if we can't position the cursor correctly in the line, put it at the end
                               firstTimeThrough = true;
                           }

                    
                           goto top;

                       }
    
                   }
                   // ignore other characters
                   break;
    
           } // switch(key)
       } // while not done
       
#if 0
       //
       //  this code should not be here.  The command console object manipulates the history in complex
       //  ways using the ScriptLineStream object.   Particularly, by calling the saveHistory() function which actually
       //  puts the lines in the history_ object.
       //
       if(!input.empty())
           history_.push_back(input);
#endif
       number = ++lineNumber_;
       print("\n");
       return true;
   }


typedef std::map<DWORD, DWORD>   ParentChildMap;


static BOOL GetProcessList( ParentChildMap &processes)
{
  HANDLE hProcessSnap;
  HANDLE hProcess;
  PROCESSENTRY32 pe32;

  // Take a snapshot of all processes in the system.
  hProcessSnap = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
  if( hProcessSnap == INVALID_HANDLE_VALUE )
  {
    return( FALSE );
  }

  // Set the size of the structure before using it.
  pe32.dwSize = sizeof( PROCESSENTRY32 );

  // Retrieve information about the first process,
  // and exit if unsuccessful
  if( !Process32First( hProcessSnap, &pe32 ) )
  {
    CloseHandle( hProcessSnap );          // clean the snapshot object
    return( FALSE );
  }

  // Now walk the snapshot of processes, and
  // save the parent child relationship
  do
  {
    processes.insert( std::pair<DWORD,DWORD>(pe32.th32ProcessID, pe32.th32ParentProcessID));
  } while( Process32Next( hProcessSnap, &pe32 ) );

  CloseHandle( hProcessSnap );
  return( TRUE );
}

void
ConsoleIF::
killChildren()
{
    ParentChildMap tmp;

    DWORD me = GetCurrentProcessId();

    if(GetProcessList(tmp))
    {
        CXXTLS_FOREACH( ParentChildMap::value_type const &cur, tmp )
        {
            if(cur.second == me)
            {
               HANDLE victim = OpenProcess(PROCESS_TERMINATE,false,cur.first);
               
               if(victim)
                   TerminateProcess(victim, 1);

               CloseHandle(victim);

               // cerr << "Killing process id " << cur.first << endl;
            }
        }
    }
}


CommandConsole*
CommandConsole::
getDefaultConsole(std::string const &historyFile, std::vector<std::string> &historyLines)
{
    return new ConsoleIF(historyFile, historyLines);
}


void 
CommandConsole::
readHistoryFile()
{
   std::ifstream s(historyFileName_);
   
   std::string line;

   while(std::getline(s, line))
   {
      history_.push_back(line);
   }
  
}

void 
CommandConsole::
writeHistoryFile()
{
   std::ofstream s(historyFileName_);

   if(history_.size() > 1000)
   {
       // make sure that we do not save more than 1000 lines

       size_t delta = history_.size() - 1000;

       history_.erase( history_.begin(), history_.begin() + delta );
   }


   CXXTLS_FOREACH(std::string const &cur, history_)
   {
       s.write(cur.data(), cur.size());
       s.write("\n", 1);
   }
}


} // namespace cxxtls

#endif
