#ifndef stringexpression_parser_h_included_p
#define stringexpression_parser_h_included_p

//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
#include "stringexpression.h"
#include <ctype.h>

namespace cxxtls
{

namespace StringExpr
{
   // implement some function bodies that can NOT be defined at the top of the file du
   // the order of definitions required by the compiler.


    template<class Iterator1, class Iterator2>
    StringExpression *parseStrExpr(Iterator1& f, Iterator2 const &l, bool discoverRedirections);

    template<class Iterator1, class Iterator2>
    StringExpression *parseSubshellBody(Iterator1&f, Iterator2 const &l, char terminator, bool discoverRedirections);

    struct Node;

    template<class Iterator1, class Iterator2>
    StringExpr::Node *parseDollarExpression(Iterator1&f, Iterator2 const &l, bool discoverRedirections);

    //==============================================================


    //==============================================================

    struct SubExpressionNode
        ///  A node in a StringExpression that is wrapper around an entire
        ///  StringExpression whose value will be returned as the value of
        ///  this node when executeHelperd.
    : public Node
    {
        std::auto_ptr<StringExpression> value_;  

        bool splitting_;

        SubExpressionNode(StringExpression* n)
        :  value_(n)
        ,  splitting_(true) // only " changes this false
        {
        }

        virtual bool allowsNoResult() const 
        {
            if(value_.get() == 0)
               return false;

            if (value_->empty())
                return false;

            for(StringExpr::Expression::const_iterator f = value_->begin(), l = value_->end();
                f != l;
                ++f
               )
            {
                if(!(*f)->allowsNoResult())
                   return false;
            }

            return true;  // only return true if all of the sub-expressions return true
        }
        

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const;
            ///
            ///  This kind of expression node is just a wrapper around the
            ///  sub-expressions which comprise it.  A true value will be returned
            ///  to indicate that this kind of sub-expression is not subject
            ///  to glob expansion.
            ///
            ///  During the course of execution, any shell variable expansions
            ///  will be drawn from the context.  Command line arguments will
            ///  also be drawn from the context.  Any optional behavior
            ///  interpretations will be decided upon based on options in the
            ///  context. 
            ///
            ///  Upon completion, the result of this node's calculations will
            ///  be left in parameter rv.
            ///

        virtual std::string executeEncapsulated(ExecutionContext *context) const;

        virtual std::string stringify() const;

    };



    //==============================================================
    struct PidNode
        /// When executeHelperd, this type of expression node returns the process id
        /// of the current process as a string.
    : public Node
    {

        PidNode()
        {
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            ///  Store the current process id number as a string the return
            ///  value.  
            ///
            ///  return true to indicate this text should not be glob expanded.
            ///  numbers don't glob anyway.
        {
           char buffer[40];

           sprintf(buffer, "%d", getpid());

           output.push_back(buffer);

           return true;  
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const
            // safe to return un-encapsulated
        {
           char buffer[40];

           sprintf(buffer, "%d", getpid());

           return buffer; 
            
        }

        virtual std::string stringify() const { return "$$"; }  // debug print helper

    };

    //==============================================================
    struct ErrorCodeNode
        /// When executeHelperd, this type of expression node returns the process id
        /// of the current process as a string.
    : public Node
    {

        ErrorCodeNode()
        {
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            ///  Store the current process id number as a string the return
            ///  value.  
            ///
            ///  return true to indicate this text should not be glob expanded.
            ///  numbers don't glob anyway.
        {
           char buffer[40];

           sprintf(buffer, "%d", context->errorCode());

           output.push_back(buffer);

           return true;  
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const
            // safe to return un-encapsulated
        {
           char buffer[40];

           sprintf(buffer, "%d", context->errorCode());

           return buffer; 
            
        }


        virtual std::string stringify() const { return "$?"; }  // debug print helper

    };



    //==============================================================

    struct VariableExpansionNode
        ///
        /// this type of node, when executeHelperd, simply evaluates an
        /// environment variable lookup.
        ///
    : public Node
    {
        std::string name_;                 ///< ${name}
        
        std::auto_ptr<StringExpression> pattern_;     ///< Used in this syntax ${name%pattern}.  @see substring_, below.
        std::auto_ptr<StringExpression> replacement_; ///< Used in this syntax ${name/pattern/replacement}.  @see substring_, below

        bool first_;                // remove leading instance of pattern
        bool last_;                 // remove trailing instances of pattern
        bool all_;                  // either replace all instances, or
                                    // replace the longest match of the pattern,
                                    // depending on the interpretation of first_ and
                                    // last_.
        bool beginOnly_;            // replace string only if match is at beginning of string
        bool endOnly_;              // replace string only if match is at end of string

        int upperCase_;             // 0 means no, 1 means first letter, 2 means all
        int lowerCase_;             // 0 means no, 1 means first letter, 2 means all

        bool length_;               // true means that we are only interested in the length of the
                                    // varabile

        bool substring_;            // this variable overrides all the others, if true.
                                    // it meas that pattern is really a string that defines
                                    // the offset within the variable defined by name
                                    // and that the replacement is really a length of a substring
                                    // foudn there.

        // this constructor handles most cases

        VariableExpansionNode(std::string const &name, 
                              StringExpression *pattern=0, 
                              StringExpression *replacement=0, 
                              bool first=false, 
                              bool last=false, 
                              bool all=false,
                              bool beginOnly=false,
                              bool endOnly=false,
                              int upperCase=0,
                              int lowerCase=0,
                              bool length=false
                             )
        :  name_(name)
        ,  pattern_(pattern)
        ,  replacement_(replacement)
        ,  first_(first)
        ,  last_(last)
        ,  all_(all)
        ,  beginOnly_(beginOnly)
        ,  endOnly_(endOnly)
        ,  upperCase_(upperCase)
        ,  lowerCase_(lowerCase)
        ,  length_(length)
        ,  substring_(false)
        {
        }

        // this one is a special case for substrings

        VariableExpansionNode(std::string const &name, 
                              bool               substring,      // you MUST set this to true to use this constructor
                              StringExpression *pattern=0, 
                              StringExpression *replacement=0
                             )
        : name_(name)
        ,  pattern_(pattern)
        ,  replacement_(replacement)
        ,  first_(false)
        ,  last_(false)
        ,  all_(false)
        ,  beginOnly_(false)
        ,  endOnly_(false)
        ,  upperCase_(false)
        ,  lowerCase_(false)
        ,  length_(false)
        ,  substring_(substring)
        {
        }


        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const;
            ///  Evaluate the specified variable expansion.  Store the
            ///  evaluated string in rv.  Return false to indicate that the
            ///  glob mechanism can further expand the results.

        virtual std::string executeEncapsulated(ExecutionContext *context) const;

        virtual std::string stringify() const ;

        virtual bool allowsNoResult() const 
            // only if you are evaluating "$@" can you possibly get no result at all, not even an empty string.
        {
           return (
                       name_.size() == 1
                    && name_[0] == '@'
                   );
        }

    };


    //==============================================================

    struct SubShellNode
        ///  When executeHelperd, a SubShellNode invokes a command from the
        ///  default operating system command line script enterpreter.
        ///  The entire output of that command, including embedded 
        ///  newlines are returned as the value.
    : public Node
    {
        StringExpression *command_;   // `cmd`
        bool              splitResults_;
        bool              encodeResults_;  // encode when we have "$(echo hi)" as opposed to just $(echo hi)

        SubShellNode(StringExpression *c, bool splitResults, bool encodeResults)
        :  command_(c)
        ,  splitResults_(splitResults)
        ,  encodeResults_(encodeResults)
        {
        }

        virtual ~SubShellNode();

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const;
            ///  Evaluate the string expression that defines this command
            ///  then executeHelper that string as a system command.  Store the 
            ///  full stdout contents of that operation in rv with \n separators
            ///  between lines.  Return true to indicate
            ///  that the results should not be globbed.

        virtual std::string executeEncapsulated(ExecutionContext *context) const;

        virtual std::string stringify() const ;

    };


    //==============================================================

    struct GlobNode
        ///  This kind of expression node is used to evaluate file
        ///  name pattern matches.
        ///
    : public Node
    {
        std::string glob_;   // * or ???

        GlobNode(std::string const &s)
        :  glob_(s)
        {
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            /// Concatenate the glob text to the return string, output.
            ///
            /// return false to indicate this fragment should be globbed.
        {
           output.push_back(glob_);
           
           return false; 
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const
        {
            return glob_;  // glob nodes should never be quoted
        }

        virtual std::string stringify() const 
        {
           return glob_;
        }


    };


    //==============================================================

     inline bool isSeparator(char c)
         ///  return true if the specified character is a blank
         ///  or should be treated like one.
     {
        return    isspace(c) 
               || (c == '\n')
               || (c == '\r')
               ;
               
     }

    //==============================================================

    struct IsGlobable
    {
        bool operator()(char c) const
        {
            return (c == '*') || (c == '?');
        }
    };



    //==============================================================

     inline bool isSpecial(char c)
          ///  Return true if the specified character is not allowed
          ///  to appear in an unprotected string.   An unprotected
          ///  string looks like this:
          ///
          ///      stuff
          ///
          ///  As opposed to a protected string which looks like this:
          ///
          ///    "stuff"
          ///    'stuff'
     {
         return      c == ';' 
                 ||  c == '['
                 ||  c == '('
                 ||  c == ')'
                 ||  c == ']'
                 ||  c == '|'
                 ||  c == '&'
                 ||  c == '{'
                 ||  c == '}'
                 ;

     }

    //==============================================================

     inline bool isName(char c)
         ///  Return true if the specified character is valid as
         ///  an environment variable name.  That is, as defined
         ///  by expressions of the form:  $name.  Not with
         ///  regard to statements setting variables values.
     {
         return      (c >= 'a' && c <= 'z')
                 ||  (c >= 'A' && c <= 'Z')
                 ||  (c == '_')
                 ||  (c >= '0' && c <= '9')
                 ;

     }




        inline bool SubExpressionNode::executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const
        { 
            value_->executeHelper(context, output, false, splitting_);

            return !splitting_;
        }

        inline std::string  SubExpressionNode::stringify() const 
        { 
           return value_->stringify(); 

        }  // debug print helper

        inline SubShellNode::~SubShellNode()
        {
            delete command_;
        }

        inline std::string SubShellNode::stringify() const 
        {
           static std::string prefix("$( ");
           static std::string suffix(" )");

           return  prefix + command_->stringify() + suffix;
        }  // debug print helper



    //==============================================================

    template<class Iterator1, class Iterator2>
    inline
    void skipSeparators(Iterator1& f, Iterator2 const &l)
        ///  Given an iterator range, [f,l), move f past any
        ///  leading white space characters -- including
        ///  escaped white space.
    {
       while(f != l)
       {
          char c = *f;

          if(c == '\\')
          {
             ++f;

             if(isSeparator(*f))
                continue;
             else
             {
               --f;
               break;
             }
          }
          else
          if(!isSeparator(c))
             break;

          ++f;   // skip sepators

       }
    }

    //==============================================================

    template<class Iterator1, class Iterator2>
    void skipTo(Iterator1 &f, Iterator2 const &l, char terminator)
        /// Given an iterator range, [f,l), move f towards l.
        /// Stop when *f equals one of: terminator, ')', '}', or when f == l.  
        /// Handle nested (), nested {}, quotes:  '', "", ``, and 
        /// escaped characters.
        ///
        /// Leave f pointing to the character that terminated
        /// the parsing, or to l if no terminators were found.
    {

        char c;

        while(f != l)
        {
            c = *f;

            if( c == terminator)
               break;

            if( c == '}')
              break;   // we only get here if there is a mismatched {} pair
                       // so handle it as a normal termination.


            if(c == '\\')
            {
               ++f;  

               if(f != l)
                 ++f;

            }
            else
            if(c == '"')
            {
              ++f;
                skipTo(f,l,'"');

                if(f == l)
                   break;

                ++f;  // skip the trailing '"'

            }
            else
            if(c == '(')
            {
              ++f;
                skipTo(f,l,')');  // yeah, ')', is built in, but what else would I use for the term character?

                if(f == l)
                   break;

                ++f;  // skip the trailing ')'
            }
            else
            if(c == '{')
            {
              ++f;
                skipTo(f,l,'}');  

                if(f == l)
                   break;

                ++f;  // skip the trailing '}'
            }
            else
            if(c == '\'')
            {
              ++f;
                while(f != l && *f != '\'')
                   ++f;

                ++f;  // skip the trailing '\''
            }
            else
               ++f; // normal characters

        }

    }

    //==============================================================

    template<class Iterator1, class Iterator2>
    StringExpression *parseStrExpr(Iterator1& f, Iterator2 const &l, bool discoverRedirections)
        ///
        /// Consume the text in the range [f,l) and convert that
        /// text into a single string expression.  The parsing may
        /// stop if the string expression terminates early.
        /// leading spaces will be removed but no trailing spaces.
        /// The expression will also terminate if one of the special
        /// characters is encountered.  See function isSpecial().  
        ///
        /// If a syntax error occurs, this function returns null.
        ///
        /// Otherwise, it returns a StringExpression that you must delete.
        ///
        /// You should call the empty() method of the return value to see
        /// if the expression can be ignored.  You'll still have to delete
        /// it though.
        ///
    {
       // parseStrExpr a string of the following variations
       //
       //   [^"' \t\r\n()$]\+   non-special printable characters
       //
       //   "expandable string"
       //   'literal'
       //   $name
       //   ${expr}
       //   $(subshell)
       //   `subshell`
       //   [*?]
       //
       //  all over the above combined in any order but with no intervening
       //  spaces or other separator characters

       std::auto_ptr<StringExpression> rv(new StringExpression);

       // skip leading spaces before the string expression.  Once in an expression,
       // no spaces are allowed unless they are protected with \, ', or ".  Note that
       // if you see random "\ " (backslash spaces in the place where a separator would
       // normally be, then call that a separator to

       skipSeparators(f,l);  // skip separators before the token we are looking for.

       // now gather one or more string expression nodes

       while(f != l)
       {  // gather all nodes


         if(isSeparator(*f))
           break;                 // expression nodes cannot contain unprotected separators

          if(f == l)
             break;

          char c= *f;

          if(isSpecial(c))
          {
              if(!rv->empty())
                  break;  // cannot include special characters
                          // (unprotected) in normal string values.


              // special characters include:  (, |, ;, [, etc

              std::string tmp;

              tmp += c;

              ++f;

              bool fileNamePatternSetBegins = false;  // is this character [ and is there a non-whitespace and non [ following it.

              if(f != l)
              {
                 if(*f == c)
                 {
                    // we have ((, ||, &&, ]], etc

                    ++f;
                    tmp += c;
                 }
                 else
                 if (c == '[' && !isspace(*f))
                 {
                     fileNamePatternSetBegins = true;
                     --f;
                     goto parseGlobSubexpression;
                 }
              }

              if (!fileNamePatternSetBegins)
              {

                 std::auto_ptr<LiteralNode> node(new UnencodedLiteralNode(tmp));
                 rv->setSpecial();

                 rv->adoptNode(  node.release() );

                 break; // cannot include unprotected special characters in
                        // normal strings -- they always stand alone or in pairs.
              }

          }
          else
          if(c == '"')
          {
               ++f;

               //
               //  quoted strings look like this:
               //
               //    '"'{(constant|$expansion|'`'subshell'`')}'"'
               //
               //  We are going to just append each piece to the
               //  current string expression as we encounter them.
               //

               StringExpression* working;

               StringExpr::SubExpressionNode* nestedExpression = new StringExpr::SubExpressionNode( working = new StringExpression );

               nestedExpression->splitting_ = false;

               rv->adoptNode(nestedExpression);


               while(f != l)
               {
                  if(*f == '"')
                    {
                      ++f;
                     break;
                    }

                  c = *f;

                  if(c == '$')
                  {
                     // handle var expansions

                     ++f;
                     
                     if(f == l)
                     {
                        working->adoptNode( new StringExpr::LiteralNode("$") );
                     }
                     else
                     if(*f == '?')
                     {
                         ++f;
                         working->adoptNode( new StringExpr::ErrorCodeNode() );
                     }
                     else
                     if(*f == '$')
                     {
                       ++f;
                       working->adoptNode( new StringExpr::PidNode() );
                     }
                     else
                     if(*f == '(')
                     {
                        std::auto_ptr<StringExpression> subshell( parseSubshellBody(f,l,')', discoverRedirections) );

                        if(!subshell.get())
                           return 0;

                        if(f == l || *f != ')')
                           return 0;   // missing trailing )


                        // check to see the terminator of the subshell body wasn not '))'

                        ++f;  // skip the trailing ')'


                        if(f != l && *f == ')' )
                        {
                            return 0;
                        }


                        working->adoptNode( new StringExpr::SubShellNode(subshell.release(), false, true) );
                         
                     }
                     else
                     if( *f == '*' || *f == '@' || *f == '#')
                     {
                        std::string name;

                        name += *f++;

                        working->adoptNode( new StringExpr::VariableExpansionNode(name) );
                     }
                     else
                     if( isName(*f) )
                     {
                        // simple variable name

                        std::string name;

                        while(     (f != l)
                              && isName(*f)
                             )
                        {
                           c = *f;
                           name += c;

                           ++f;
                        }

                        working->adoptNode( new StringExpr::VariableExpansionNode(name) );
                        
                     }
                     else
                     if(*f == '{')
                     {
                       // parseDollarExpression will eat the { for us

                         if(f == l)
                             return 0;

                         std::auto_ptr<StringExpr::Node> dollarExpression( parseDollarExpression(f,l, discoverRedirections) );

                         if(f != l && *f == '}')
                            ++f;

                         if(dollarExpression.get())
                         {
                             working->adoptNode( dollarExpression.release() );
                         }
                         else
                         {
                             return 0;
                         }

                     }
                     else
                     {
                        return 0;
                     }

                  }
                  else
                  if(c == '`')
                  {
                    // handle subshells

                    ++f;

                    Iterator1 start = f;

                    while(f != l)
                    {
                        // scan until end of sequence or until we find `.  
                        // handle escape characters.

                        if(*f == '`')
                        {
                           break;
                        }
                        else
                        if(*f == '\\')
                        {
                           ++f;

                           if(f !=l)
                             ++f;
                        }
                        else
                        {
                           ++f;
                        }
                    }

                    if(f == l)
                       return 0; // we didn't find the matching character, return
                                 // no expression.

                    if(*f != '`')
                       return 0; // we didn't find the matching character, return
                                 // no expression.


                    std::auto_ptr<StringExpression> command(StringExpr::parseSubshellBody(start, f, '`', discoverRedirections));

                    if(command.get() == 0)
                      return 0;

                    ++f; // skip the trailing `

                    working->adoptNode(new StringExpr::SubShellNode(command.release(), false, false));

                  }
                  else
                  {
                     // handle constant blocks

                     std::string value;

                     while(   c != '"' 
                           && c != '$'
                           && c != '`'
                          )
                     {
                        if(c == '\\')
                        {
                           ++f;

                           if(f == l)
                           {
                              break;
                           }
                           else
                           {
                             c = *f;        // get the second character in the escape sequence

                             switch(c)
                             {
                                 case '"':
                                 case '$':
                                 case '`':
                                 case '\\':
                                    break;   // for these sequences:    \", \$', and \`
                                             // only output the second character in the sequences
                                             // but for all other cases, output both \ and the following character

                                 case '0':
                                 {  // octal


                                     int tmp = 0;

                                     while(f != l)
                                     {
                                         c = *f++;

                                         if(c < '0' || c > '7')
                                         {
                                            --f;  // leave the terminating character in the stream
                                            --f;  // and the character before the terminator due to the way
                                                  // increments occur outside this loop
                                            break;
                                         }

                                         tmp *= 8;  

                                         tmp += c - '0';
                                     }

                                     c = tmp;  // let the value +=c, below include it in the output

                                 }  // octal
                                 break;

                                 case 'x':
                                 {  // octal


                                     ++f; // skip the x

                                     if(f == l || !isxdigit(*f))
                                     {
                                        // \x  must be followed by a hex digit, or it is not hex
                                        // so lets put the \ and the x in the output

                                        value += '\\';  // include

                                        --f;  // leave f point to the x, for later skip over

                                        c='x';
                                        break;
                                     }

                                     int tmp = 0;

                                     while(f != l)
                                     {
                                         c = *f++;

                                         if(!isxdigit(c))
                                         {
                                            --f;  // leave the terminating character in the stream
                                            --f;  // we need to leave both the terminating character
                                                  // and the character before it in the stream due to the
                                                  // way increments occur outside this loop
                                            break;
                                         }

                                         tmp *= 0x10;  

                                         if(c >= 'a' && c <= 'f')
                                            tmp += ( c - 'a' + 10);
                                         else
                                         if(c >= '0' && c <= '9')
                                            tmp += (c - '0');
                                         else
                                         {
                                            tmp += ( c - 'A' + 10);
                                         }

                                     }

                                     c = tmp;  // let the value +=c, below include it in the output

                                 }  // hex
                                 break;


                                 default:              // unknown escape sequence -- leave the \ in the stream
                                    value += '\\';   
                                    break;
                             }

                           }
                           
                        }


                        value += c;
                        ++f;
                         
                        if(f == l)
                           return 0;   // we should have hit " not end of line

                        c = *f;
                     }

                     if(!value.empty())
                     {
                          working->adoptNode( new StringExpr::LiteralNode(value, true) ); // pre-encoded literal node
                     }

                  }

               }

          }
          else
          if(c == '\'')
          {
             // non-expandable constant

             ++f;

             std::string value;

             while(f != l)
             {
                 c = *f;

                 if(c == '\'')
                 {
                    ++f;
                    break;
                 }

                 value += c;

                 ++f;

             }

             rv->adoptNode( new StringExpr::RawLiteralNode(value) );


          }
          else
          if(c == '$')
          {
             // unprotected expansion

             ++f;

             if(f == l)
             {
                rv->adoptNode( new StringExpr::LiteralNode("$") );
             }

             if(*f == '$')
             {
                 ++f;
                 rv->adoptNode( new StringExpr::PidNode() );
             }
             else
             if(*f == '(')
             {
                // handle subshells of the form $(...)

                std::auto_ptr<StringExpression> subshell(parseSubshellBody(f,l,')', discoverRedirections));

                if(f != l && *f == ')')
                {
                  Iterator1 tmp(f);

                  std::auto_ptr<StringExpression> endingToken(parseStrExpr(f,l,discoverRedirections));  

                  static std::string closingParen(")");

                  if(closingParen != endingToken->stringify())
                  {
                     f = tmp; // if we get "))" instead of ")", leave it in the stream,
                              // this will trigger an error by some calling routine
                              // who won't like it
                     if (f != l && *f == ')')
                       ++f;
                  }
                }

                if(!subshell.get())
                    return 0;

                rv->adoptNode( new StringExpr::SubShellNode(subshell.release(), true, false) );
             }
             else
             if(*f == '{')
             {
               //handle fancy variable expansions of the form ${name type parm separator otherparm }
                
                 if(f == l)
                     return 0;
               
                 std::auto_ptr<StringExpr::Node> dollarExpression( parseDollarExpression(f,l,discoverRedirections) );

                 if(f != l && *f == '}')
                   ++f;
               
                 if(dollarExpression.get())
                 {
                     rv->adoptNode( dollarExpression.release() );
                 }
                 else
                 {
                     return 0;
                 }

             }
             else
             if (*f == '*' || *f == '@' || *f == '?' || *f == '#')
             {
                std::string name;

                name += *f++;

                rv->adoptNode( new StringExpr::VariableExpansionNode(name) );

             }
             else
             if( isName(*f) )
             {
                 // handle simple name expansions:  $name

                 std::string name;
                 
                 while(f != l)
                 {
                    c = *f;
                 
                    if(!isName(c))
                       break;
                 
                    name += c;
                 
                    ++f;
                 }
                 
                 if(name.empty())
                 {
                    rv->adoptNode( new StringExpr::LiteralNode("$") );
                 }
                 else
                 {
                    rv->adoptNode( new StringExpr::VariableExpansionNode(name) );
                 }
             }
             else
             {
                // syntax error
                // nyi;

                return 0;
             }
          }
          else
          if(c == '`')
          {
             // subshell invocation, expandable, etc

              ++f;
              
              Iterator1 start = f;

              while(f != l)
              {
                  if(*f == '`')
                  {

                     break;
                  }
                  else
                  if(*f == '\\')
                  {
                     ++f;
              
                     if(f !=l)
                       ++f;
                  }
                  else
                  {
                     ++f;
                  }
              }
              
              if(f == l)
                return 0;  // missing trailing ` is a syntax error

              std::auto_ptr<StringExpression> command(StringExpr::parseSubshellBody(start, f, '`',discoverRedirections));

              skipSeparators(start,f);

              if(start != f)
                {
                  return 0;  // parseSubshellBody did not completely consume the text range [start,f)
                }

              ++f; // skip the trailing '
              
              if(command.get() == 0)
                return 0;
              
              rv->adoptNode(new StringExpr::SubShellNode(command.release(), true, false));
              
          }  // back tick
          else
          if(c == '*' || c == '?' || c == '[' )
          {
              // repetitions of shell expansion characters

parseGlobSubexpression:
              std::string glob;

              while(f != l)
              {
                 c = *f;

                 if(c == '[')
                 {
                     glob += c;   // put the [ in the expression
                     ++f;

                     while(f != l && *f != ']')
                     {
                         // copy all characters into the glob string until we reach ]

                         c = *f;

                         if(c == '\\')
                         {
                            Iterator1 next = f; ++next;

                            if(next == l)
                            {
                               glob += '\\';
                               ++f;
                            }
                            else
                            {
                               ++f;
                               c = *f;

                               switch(c)
                               {
                                   case 'n':  c = '\n'; break;
                                   case 'r':  c = '\r'; break;
                                   case 't':  c = '\t'; break;
                                   case 'b':  c = 0x08; break;
                                   case 'a':  c = 0x07; break;

                                   default:  break;
                               }

                               glob += c;

                               ++f;
                            }

                         }
                         else
                         if(c == ']')
                         {
                            glob += c;
                            ++f;
                            break;
                         }
                         else
                         {
                            glob += c;
                            ++f;
                         }
                     }
                 }
                 else
                 {
    
    
                     if (isspace(c) || c == '"' || c == '$' || c == '\'' || c == '`')
                        break;
    
                     if (c == ')' || c == '{')
                        break;
    
                     if(c != '/')
                     {
                        if(c == '\\')
                        {
                           ++f;
    
                           if(f == l)
                           {
                              glob += '\\';
                              continue;  // f has already been incremented
                           }
                           else
                           {
                               switch(c)
                               {
                                   case ' ':  c = ' ';  break;
                                   case 's':  c = ' '; break;
                                   case 't':  c = '\t'; break;
                                   case 'e':  c = (char)(0x1b); break;
                                   case 'n':  c = '\n'; break;
                                   case 'b':  c = (char)(0x08); break;
                                   case 'a':  c = (char)(0x07); break;
                                   case 'r':  c = '\r'; break;
                               
                                   default:
                                      break;
                               }
                              
    
                             // the code below will add c to glob
                           }
                        }
    
                        glob += c;
                        ++f;
                     }
                     else
                        break;
                  }
              }

              rv->adoptNode( new StringExpr::GlobNode(glob) );


          }
          else
          { // expandable unprotected constant, 

            std::string value;

            bool isSplittable = true;

             while(f != l)
             { // parseStrExpr till special character, expand escape characters

                c = *f;

                if(   isSpecial(c) 
                   || isSeparator(c) 
                   || (c == '$')
                   || (c == '`')
                   || (c == '"')
                   || (c == '\'')
                   || (c == '*')
                   || (c == '?')
                  )
                {
                   if( c != '&' || !discoverRedirections || !rv->isRedir() )
                       break;

                   // if we get here, we have something like this:
                   //
                   //     2>&1

                }

                if(   !discoverRedirections
                   && ( c == '<' || c == '>' )
                  )
                {
                    value += c;
                    ++f;

                    if(f != l) 
                    {
                       if(*f == c)
                       {
                           value += c;
                           ++f;
                       }
                       else
                       if(*f == '=')
                       {
                          value += *f++;
                       }
                    }

                    rv->adoptNode(new StringExpr::LiteralNode(value));
                    return rv.release();
                }

                if(c == '<' || c == '>')
                {

                    if(value.empty())
                      {
                        //
                        //  this is the beginning of a redirection like this:  >file
                        //
                        rv->setRedir();
                        value += c;
                        ++f;  

                        bool isInlineFileRedir = false;

                        if (f != l && *f == c)
                        {
                            ++f;  // skip the second < or > if 2 are found in a row
                            value += c;
                            isInlineFileRedir = true;
                        }
                        
                        
                        while(f != l)
                        {
                           if(isSeparator(*f))
                              ++f;
                           else
                              break;
                        }

                        if (isInlineFileRedir)
                        {
                            while (f != l && !isSeparator(*f))
                            {
                               value += *f++;
                            }
                        }

                      }
                    else
                    {
                       // normally, redirections occur at the
                       // start of the string, but under these
                       // circumstances they don't:
                       //
                       //    2>something
                       //    0</dev/null
                       //    1>&2

                       if(value.find_first_not_of("0123456789") < 
                          value.size()
                         )
                       {
                          break;  // this is not a standard file redirection
                                  // so the > or < marks the begining of the
                                  // next token
                       }

                       // in this case, we one of the valid prefixes for a redirection.

                       rv->setRedir();

                       value += c;
                       ++f;

                        while(f != l)
                        {
                           if(isSeparator(*f))
                              ++f;   
                           else
                              break;
                        }



                    }
                }
                else
                if(c == '\\'  && (f !=l))
                { // handle the current escaped character


                    ++f;

                    c = *f;

                    switch(c)
                    {
                        case ' ':  c = ' ';  isSplittable=false; break;   // escaping a normal IFS character means turnning of splittability
                        case 's':  c = ' ';  isSplittable=false; break;   // ditto
                        case 't':  c = '\t'; isSplittable=false; break;   // ditto
                        case 'e':  c = (char)(0x1b); break;
                        case 'n':  c = '\n'; isSplittable=false; break;   // escaping normal IFS characters turns off splittability
                        case 'b':  c = (char)(0x08); break;
                        case 'a':  c = (char)(0x07); break;
                        case 'r':  c = '\r'; isSplittable=false;  break;  // escaping normal ifs characters turns off splittability
                        case '\\':  c = '\\'; break;

                        case '(':  c = '('; break;
                        case ')':  c = ')'; break;

                        case '0':
                        {  // octal
                        
                        
                            int tmp = 0;
                        
                            while(f != l)
                            {
                                c = *f++;
                        
                                if(c < '0' || c > '7')
                                {
                                   --f;  // leave the terminating character in the stream
                                   --f;  // leave the terminating character in the stream
                                   break;
                                }
                        
                                tmp *= 8;  
                        
                                tmp += c - '0';
                            }
                        
                            c = tmp;  // let the value +=c, below include it in the output
                        
                        }  // octal
                        break;

                        case 'x':
                        {  // octal


                            ++f; // skip the x

                            if(f == l || !isxdigit(*f))
                            {
                               // \x  must be followed by a hex digit, or it is not hex
                               // so lets put the \ and the x in the output

                               value += '\\';  // include

                               --f;  // leave f point to the x, for later skip over

                               c='x';
                               break;
                            }

                            int tmp = 0;

                            while(f != l)
                            {
                                c = *f++;

                                if(!isxdigit(c))
                                {
                                   --f;  // leave the terminating character in the stream
                                   --f;  // leave the terminating character in the stream
                                   break;
                                }

                                tmp *= 0x10;  

                                if(c >= 'a' && c <= 'f')
                                   tmp += ( c - 'a' + 10);
                                else
                                if(c >= '0' && c <= '9')
                                   tmp += (c - '0');
                                else
                                {
                                   tmp += ( c - 'A' + 10);
                                }

                            }

                            c = tmp;  // let the value +=c, below include it in the output

                        }  // hex
                        break;



                        default:
                           value += '\\';
                           break;
                    }

                    value += c;

                    ++f;

                }  // handle the current escaped character
                else
                {  // regular characters

                   value += c;
                   ++f;

                }  // regular characters

             } // parseStrExpr till special character, expand escape characters

             if (isSplittable)
                 rv->adoptNode( new StringExpr::SplitableLiteralNode(value, true) );
             else
                 rv->adoptNode(new StringExpr::LiteralNode(value, true));  // we only get here if the string contains an escaped space (or other escaped IFS character)

             if ( (f != l) && (*f == '<' || *f == '>') )
                break;   // non-redirection <> chars in the middle of a word is end of word


             if(f != l && *f == '[')
                goto parseGlobSubexpression;

          }  // expandable unprotected constant

       }   // gather all nodes

       return rv.release();

    }


    //==============================================================

    template<class Iterator1, class Iterator2>
    StringExpression *parseSubshellBody(Iterator1&f, Iterator2 const &l, char terminator, bool discoverRedirections)
        /// parseStrExpr until the terminator character is found, leave it in the stream for the caller
        /// to analyze.   return the subshell string expression.
        /// a subshell body syntax looks like this:
        ///
        ///@begincode
        ///
        ///    $( command body )
        ///     ^
        ///
        ///    stuff ;
        ///    ^
        ///
        ///    stuff | morestuff
        ///    ^
        ///
        ///@begincode

    {

    // void skipTo(Iterator1 &f, Iterator2 const &l, char terminator)


        {
           if(terminator == ')')
               ++f;  // skip (

            Iterator1 org = f;

            skipTo(f, l, terminator);

            if(terminator == ')' && f == l)
               return 0; // invalid syntax

            // really, even in the "`" case, f should not equal l, but some bug somewhere causes it to be
            // Sooo, I need the above test.  too lazy to track down and fix all the places that could cause it

            std::string value(org,f);

            std::auto_ptr<StringExpression> rv(new StringExpression);

             rv->adoptNode( new StringExpr::RawLiteralNode(value) );

             rv->protect();
            
             rv->separateWithSpaces();

             return rv.release();
        }

        
#if 0

        std::auto_ptr<StringExpression> rv(new StringExpression);

        rv->protect();
        rv->separateWithSpaces();

        if(terminator == ')')
          ++f;

        while(f != l)
        {
            skipSeparators(f,l);

            if(f != l && *f == terminator)
            {
              break;  // leave the terminator in the stream to let the caller decide
                      // if an error has occurred or not.
            }

            std::auto_ptr<StringExpression> cur(StringExpr::parse(f,l));

            if(!cur.get())
            {
              // we didn't get a sub-expression

              if(f == l)
                 break;

              if(*f == ')')
              {
                 ++f;

                 break;
              }

            }
            else
            {
               rv->adoptNode(new StringExpr::SubExpressionNode(cur.release()));

               if(f == l)
               {
                  break;
               }

            }
           
        }

        return rv.release();
#endif

    }



    //==============================================================
    template<class Iterator1, class Iterator2>
    bool parseDollarWord(Iterator1&start, Iterator2 const &f, StringExpression *pattern, bool discoverRedirections)
        ///
        ///  Consume characters out of the text range, [start,f).  Convert those characters
        ///  into nodes in a StringExpression which will implement a "word" from either
        ///  the pattern side or the replacement side of a string expression of the following
        ///  form:
        ///  
        ///@begincode
        ///
        ///     ${variable/pattern/replacement}
        ///
        ///@endcode
        ///
    {
        std::string tmp;
    
        while(start != f)
        {
           char c = *start;
    
           if(c == '$' || c == '"' || c == '`')
           {
              if(!tmp.empty())
              {
                 pattern->adoptNode(new StringExpr::LiteralNode(tmp) );
    
                 tmp.resize(0);
              }
    
              std::auto_ptr<StringExpression>  tmp(parseStrExpr(start,f, discoverRedirections));
    
              if(tmp.get() == 0)
                 return false;
    
              pattern->adoptNode( new StringExpr::SubExpressionNode(tmp.release()) );
    
           }
           else
             {
                 tmp += c;
                 ++start;
             }
              
        }

        if(!tmp.empty())
        {
           pattern->adoptNode(new LiteralNode(tmp) );
        
        }

        return true;

    }
    //==============================================================

    template<class Iterator1, class Iterator2>
    StringExpr::Node *parseDollarExpression(Iterator1&f, Iterator2 const &l, bool discoverRedirections)
        ///
        /// parseStrExpr a string expression beginning with a $.
        ///
        /// A dollar expression looks like this:
        ///
        ///  ${name<options>}
        ///        ^
        ///
        /// The options include things like this:
        ///
        ///
        ///   /firstMatch/replaceWith
        ///   //allMatches/replaceWith
        ///   %removeShortestLeadingMatch
        ///   %%removeLongestLeadingMatch
        ///   #removeShortestTrailingMatch
        ///   ##removeLongestTrailingMatch
        ///   :offset:length

    {

        std::auto_ptr<StringExpression>  rv(new StringExpression);  // this expression will be returned as a subexpression node

        std::string name;
        std::auto_ptr<StringExpression> pattern(new StringExpression);
        std::auto_ptr<StringExpression> replacement(new StringExpression);

        bool first=false;  // see VariableExpansionNode::executeHelper() for an understanding of these variables.
        bool last=false;
        bool all=false;
        bool beginOnly = false;  // replace string only if match appears at beginning 
        bool endOnly  = false;   // replace string only if match appears at end

        ++f;

        bool length=false;

        if(f != l && *f == '#')
        {
          ++f;

          length = true;

          if(    (f != l)
             &&  (*f == '*' || *f == '@')
            )
          {
            name += *f++;
       
          }
        }


        if(   (f != l)
           && (    (*f == '*')
               ||  (*f == '@')      
              )
          )
          {
            name.resize(0);
            name += *f++;
          }
        else
        while(f != l)
        {
           if(!isName(*f))
              break;
           name += *f++;
        }

        if(f == l)
            return 0; // syntax error, missing }


        if(length)
        {
           if(f != l && *f != '}')
           {
              return 0;   // we were looking for:
                          // 
                          //   ${#name}
                          //          ^

           }

           return new VariableExpansionNode(name, 
                                            pattern.release(), 
                                            replacement.release(), 
                                            first, 
                                            last, 
                                            all, 
                                            beginOnly, 
                                            endOnly, 
                                            0, 
                                            0,
                                            true  // request for length as a string
                                            );



        }

        char c = *f;

        if(c != '}')
        {
           // handle some form of substituion:  leading, trailing, or middle

            if(c == '%' || c == '#')
            {
               // remove trailing pattern matches

               if(c == '%')
                    last = true;
               else
                    first = true;

               ++f;

               if(f != l && *f == c)
               {
                  all = true;
                  ++f;
               }

               Iterator1 start=f;

               skipTo(f, l, '}');

               if(f == l)
               {
                  return 0;
               }

               if(!parseDollarWord(start, f, pattern.get(), discoverRedirections))
                   return 0; // syntax error


            }
            else
            if(c == '/')
            {
               // perform a text substitution
               ++f;
               // neither first nor last is true for substitutions

               if(f != l)
               {
                   if(*f == '/')
                   {
                      all = true;
                      ++f;
                   }
                   else
                   if(*f == '#')
                   {
                     beginOnly = true;
                     ++f;
                   }
                   else
                   if(*f == '%')
                   {
                      endOnly = true;
                      ++f;
                   }
               }


               Iterator1 start=f;

               skipTo(f, l, '/');

               if(f == l)
               {
                  return 0;
               }

               if(!parseDollarWord(start, f, pattern.get(), discoverRedirections))
                   return 0; // syntax error

               ++f; // skip past the delimiter in the pattern half of the
                    // statement.

               start = f;

               skipTo(f,l, '}');

               if(f == l)
               {
                  return 0;  // missing the trailing }?  How?
               }

               if(!parseDollarWord(start, f, replacement.get(), discoverRedirections))
                   return 0; // syntax error

            }
            else
            if(c == '^')
            {
               // we have something like:  ${var^
               //                               ^

               ++f;  // skip first ^

               if(f == l)
               {
                   return 0;  // missing the trailing }
               }

               int upperCount=1;

               if(*f == '^')
               {
                 ++f;

                 ++upperCount;   //here we have ${var^^}
                                 //                    ^

               }
             
               if(*f != '}')
                  return 0;  // syntax error

               return new VariableExpansionNode( name, pattern.release(), replacement.release(), first, last, all, beginOnly, endOnly, upperCount);

            }
            else
            if(c == ',')
            {
               // we have something like:  ${var,
               //                               ^

               ++f;  // skip first ,

               if(f == l)
               {
                   return 0;  // missing the trailing }
               }

               int lowerCount=1;

               if(*f == ',')
               {
                 ++f;

                 ++lowerCount;   //here we have ${var,,}
                                 //                    ^

               }
             
               if(*f != '}')
                  return 0;  // syntax error

               return new VariableExpansionNode( name, pattern.release(), replacement.release(), first, last, all, beginOnly, endOnly, 0, lowerCount);

            }
            else
            if(c == ':')
            {
               // expecting ${varName:$(offsetExpression):$(lengthExpression))

               ++f;  // skip the colon

               if(f != l)
               {
                   Iterator1 of = f;  // first character of offset expression
                   Iterator1 ol = f;  // end of offset expression
                   Iterator1 lf = f;  // first character of length expression
                   Iterator1 ll = f;  // end of length expression;

                   // skip till end of the offset expression

                   skipTo(f,l,':');  

                   ol = f;

                   // at this point, the offset expressions runs from of to ol.

                   if(f != l && *f != '}')
                   {

                       // if we get here, we have   :lengthExpression}


                       ++f;   // skip th e:

                       lf = f;
                       ll = f;

                       skipTo(f,l,'}');

                       ll = f;
                   }

                   // at this point, we may or may not have a length expression, but regardless
                   // we can safely perform lf == ll

                   // we now have 3 things:  the name, and 2 substrings (possibly empty) which can be
                   // used to create the offset and length sub-expressions

                   if(of != ol)
                   {
                      // now return the needed expression node

                      pattern.reset(parseStrExpr(of,ol, false));

                      if(lf != ll)
                      {
                         replacement.reset(parseStrExpr(lf,ll, false));
                      }

                      return new VariableExpansionNode(name, true, pattern.release(), replacement.release());

                   }

                   // since we are not returning prior to this point, this syntax will produce an error
               }


            }
            // else syntax error, but that will be handled below

        }

        if(f == l)
           return 0;   // some form of syntax error

        if(*f != '}')
           return 0;   // some form of syntax error

        return new VariableExpansionNode( name, pattern.release(), replacement.release(), first, last, all, beginOnly, endOnly);

        // nyi;
    }


    template<class Iterator1, class Iterator2>
    StringExpression *parse(Iterator1& f, Iterator2 const &l)
        ///
        /// Consume the text in the range [f,l) and convert that
        /// text into a single string expression.  The parsing may
        /// stop if the string expression terminates early.
        /// leading spaces will be removed but no trailing spaces.
        /// The expression will also terminate if one of the special
        /// characters is encountered.  See function isSpecial().  
        ///
        /// If a syntax error occurs, this function returns null.
        ///
        /// Otherwise, it returns a StringExpression that you must delete.
        ///
        /// You should call the empty() method of the return value to see
        /// if the expression can be ignored.  You'll still have to delete
        /// it though.
        ///
    {
       return parseStrExpr(f,l, true);
    }




    template<class Iterator1, class Iterator2>
    StringExpression *parseEmbeddedExpressions(Iterator1& f, Iterator2 const &l)
        ///
        /// Consume the text in the range [f,l) and convert that
        /// text into a single string expression.  
        ///
        /// This function converts the input text range into an expression
        /// that will replicate all the text.  If there are imbedded expressions
        /// within the outer text range, these will be recognized and made
        /// executable as part of the resultant expression.  That is, 
        /// The embedded expressions will be executed separately and their
        /// results returned in its appropriate place within the final result
        /// expression value (upon execution of course).
        ///
    {
        std::auto_ptr<StringExpression>  rv(new StringExpression);  

        std::string leadingPortion;

        while(f != l)
        {
           char c = *f;
           
           if(c == '$' || c == '`')
           {
               // handle variable expansions

               if(!leadingPortion.empty())
               {
                   std::auto_ptr<LiteralNode> node(new LiteralNode(leadingPortion));
           
                   rv->adoptNode(  node.release() );

                   leadingPortion.resize(0);
                  
               }

               std::auto_ptr<StringExpression>  subExpression(parseStrExpr(f, l, false));

               if(subExpression.get())
               {
                   // keep all 4 statements together or exception saftey.

                   std::auto_ptr<SubExpressionNode> subNode(new SubExpressionNode( subExpression.get() ) );
                   subExpression.release();

                   rv->adoptNode( subNode.get() );
                   subNode.release();
               }
           }
           else
           {
               leadingPortion += c;
               ++f;
           }
        }

        if(!leadingPortion.empty())
        {
            std::auto_ptr<LiteralNode> node(new LiteralNode(leadingPortion));
    
            {
                // for exception saftey, keep these together
                rv->adoptNode(  node.get() );
                node.release();
            }
           
        }
       
        return rv.release();
    }

}  // namespace StringExpr
}  // namespace cxxtls
#endif
