//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/scriptfunctions.h>
#include <cxxtls/ScriptLanguageException.h>
#include <cxxtls/AutoDestroyingPointer.h>
#include <cxxtls/thread.h>
#include <iomanip>

///@file
///
///  This file implements the bodies fo functions defined in file include/cxxtls/scriptfunctions.h
///
namespace cxxtls
{
namespace ScriptFunctions
{

    static std::string logLineHeader("+ ");
    static std::string equal("=");

#ifdef _MSC_VER
    static std::string nl("\r\n");
#else
    static std::string nl("\n");
#endif

    Statement::~Statement()
    {
        CXXTLS_HEAVYDEBUG("~Statement(" << std::hex << (long )(this) << std::hex << ")");  
    }
    

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

     StatementReturnCode
     StatementList::
     execute(ScriptThread::ThreadContext *executionContext,
             FunctionScope               *scope,
             JobController               *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
            ) const
     {
         AutoDestroyingPointer<ScriptThread::ThreadContext> subShellContext;

         StringExpr::ExecutionVariables localCopy;  // empty unless we are in a subshell

         if(subShell_)
         {
             // make a local copy of the variables

             localCopy = executionContext->variables().variables();

             subShellContext.reset(ScriptThread::ThreadContext::createForNewJob(  executionContext->variables().scriptName(),
                                                                    executionContext->variables().arguments(),
                                                                    &localCopy,
                                                                    executionContext->fileHandles_->clone()
                                                                 )
                                  );

             subShellContext->variables().setGlobbing( executionContext->variables().globbing() );

             executionContext = subShellContext.get();

         }


         CXXTLS_FOREACH(ThreadSafeHandle<Statement> const &statement, code_)
         {

             StatementReturnCode rc;
             
             if(!alreadyDetached_ && statement->detached())
             {
                  ThreadSafeHandle<Statement> tmpStatement(statement);

                  jobController->executeDetached(executionContext,
                                                 scope,
                                                 jobController,
                                                 tmpStatement
                                                );

                  rc = SRC_Success;
             }
             else
             {
                 rc = statement->execute(executionContext, scope, jobController, skipArg0);
             }


             switch(rc)
             {
                 default:
                     throw cxxtls::ScriptLanguage::Exception("StatementExecutionError", "return code=SRC_", int(rc));
                     ;

                 case SRC_Success:
                     break;

                 case SRC_Break:
                     return SRC_Break;
                     break;

                 case SRC_Continue:
                     return SRC_Continue;
                     break;

                 case SRC_Throw:
                     return SRC_Throw;
                     break;

                 case SRC_Return:
                     return SRC_Return;
                     break;

             }
         }

         return SRC_Success;  // nothing weird happened.
     }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////


    StatementReturnCode 
    VariableAssignmentStatement::
    execute(ScriptThread::ThreadContext *executionContext,
            FunctionScope                *functionScope,
            JobController                *jobController,
            bool                                          skipArg0            // for use by the command "builtin"
           ) const
    {
        StringArray output;

        if(value_.get())
            value_->executeHelper(&executionContext->variables(), output, false, false);  // var assigns neither glob nor split

        executionContext->variables().assign( varname_, output.combine() );

        if(executionContext->loggingExecution())
        {
            executionContext->fileHandles_->writeBytes(2, logLineHeader + varname_ + equal + output.combine() + nl);
        }

        return SRC_Success;
    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    VariableDefinitionStatement::
    ~VariableDefinitionStatement()
    {
       // does nothing
    }

    StatementReturnCode 
    VariableDefinitionStatement::
    execute(ScriptThread::ThreadContext *executionContext,
            FunctionScope                *functionScope,
            JobController                *jobController,
            bool                                          skipArg0            // for use by the command "builtin"
           ) const
    {
        StringArray output;

        value_.execute(&executionContext->variables(), output) ;

        executionContext->variables().assign( varname_, output.combine());

        return SRC_Success;
    }



///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
    StatementReturnCode 
    StatementListStatement::
    execute(ScriptThread::ThreadContext *executionContext,
            FunctionScope                *functionScope,
            JobController                *jobController,
            bool                          skipArg0            // for use by the command "builtin"
           ) const
    {
        if(statements_.get())
        {
            return statements_->execute(executionContext, functionScope, jobController, skipArg0);
        }

        return SRC_Success;
        
    }



} // ScriptFunctions
} // cxxtls
