#include <cxxtls/controlledprocess.h>
#include <stdlib.h>
#include <string.h>
#include <memory>

//
// This test program demonstrates the act of controlling a process.  The
// process being controlled is the 'cat' program.  cat reads from stdin and
// writes to stdout.  This test program launches it -- writing to its stdin
// and reading from its stdout.
//
// This test demonstrates the complexities of controlling a running process.
// See the STEP comments before each step to see a simple but realistic
// example.
//

using namespace cxxtls;

int main()
{
  // STEP 1:  start the process running

  std::auto_ptr<ControlledProcess> cpPtr(ControlledProcess::start("cat"));

  ControlledProcess &cp(*cpPtr);

  // STEP 2:  check for errors

  if(cp.status() != 0)
  {
    fprintf(stderr, "couldn't start process\n");
    exit(1);
  }

  char buffer[256];

  // STEP 3:  write data to the controlled process

  if(!fputs("line1\nline2\nline3\n", cp.process_stdin()))
  {
    fprintf(stderr,"couldn't write to process stdin\n");
    exit(1);
  }

  // STEP 4: read data from the controlled process -- read until
  //         it producess output which you know means that this
  //         particular output data block is finished -- you can't
  //         just loop forever reading unless you know you will
  //         not be sending any more data to the process.  Since
  //         fgets() blocks, you won't wake up from it if you read
  //         by mistake

  int readCount;

  readCount = 0;

  while(fgets(buffer, sizeof(buffer), cp.process_stdout()))
  {
     ++readCount;

     fputs(">>", stdout);
     fputs(buffer, stdout);

     if(strncmp(buffer,"line3",5) == 0)
       break;
  }

  // STEP 5: write more stuff then close the stdin file to signal
  //         end of input to the child process

  if(!fputs("line4\nline5\nline6\n", cp.process_stdin()))
  {
    fprintf(stderr,"couldn't write to process stdin\n");
    exit(1);
  }

  fflush(cp.process_stdin());
  fclose(cp.process_stdin());

  // STEP 6: read the remaining output from the child process

  while(fgets(buffer, sizeof(buffer), cp.process_stdout()))
  {
     ++readCount;

     fputs(">>", stdout);
     fputs(buffer, stdout);
  }

  fclose(cp.process_stdout());

  // STEP 7: done!

  if(readCount != 6)
  {
     printf("ERROR:  Expected to have read 6 lines from the child process but found only:  %d\n", readCount);
  }
  else
  {
     printf("YEAH BABY YEAH!   s'all good.\n");

  }



  exit(0);

}
