
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.



#include <cxxtls/cursorwindow.h>
#include <iostream>
#include <vector>

using namespace std;
using namespace cxxtls;

//@File
//
///   This test program runs the window through a sequence of
///   split screen layouts -- creating and deleting windows
///   whose appearance is described in the output so that you
///   can verify that it is correct.  You press the space bar to
///   step to the next screen state -- other keys are ignored --
///   you must run through all of them.


class PortVector
  ///  A PortVector is a vector<port*>, but which acts like it wasn't.
  ///  That is, it acts like a vector<port>.  This container exists
  ///  so that erase() is guaranteed to delete the items being erased,
  ///  not just manipulate pointers to include cells in the vector in and
  ///  out of the begin()->end() range as is done in some vector
  ///  implementations.
  ///
  ///  Note that all vector methods aren't implemented and that in order to
  ///  push an item onto the PortVector, that item must be an auto_ptr<viewport>.
  ///  The kind of thing returned by CursorWindow::new_viewport() as well as
  ///  CursorWindow::viewport::vsplit and hsplit.
{
public:

   typedef std::vector< CursorWindow::viewport *> rep_type;

   rep_type rep_;

   struct iterator
   {
       rep_type::iterator  rep_;

       CursorWindow::viewport& operator*()
       {
	  CursorWindow::viewport* p = *rep_;
	  return *p;
       }

       iterator( rep_type::iterator r )
       : rep_(r)
       {
       }

       iterator( iterator const &r )
       : rep_(r.rep_)
       {
       }

       iterator &operator++()
       {
	 ++rep_;
	
	 return *this;
       }

       iterator operator++(int)
       {
	 rep_type::iterator rv = rep_;
	
	 ++rep_;
	
	 return iterator(rv);
       }


       iterator &operator--()
       {
	 --rep_;
	
	 return *this;
       }

       iterator operator--(int)
       {
	 rep_type::iterator rv = rep_;
	
	 --rep_;
	
	 return iterator(rv);
       }

       iterator operator+ (int i)
       {
	 rep_type::iterator rv = rep_ + i;
	
	 return rv;
       }

       iterator operator- (int i)
       {
	 rep_type::iterator rv = rep_ - i;
	
	 return rv;
       }


       bool operator==( iterator const &r)
       {
	 return rep_ == r.rep_;
       }

       bool operator!=( iterator const &r)
       {
	 return rep_ != r.rep_;
       }


   };

   void erase(iterator a)
   {
      rep_type::iterator it = a.rep_;
      CursorWindow::viewport *p = *it;
      delete p;
      rep_.erase(a.rep_);
   }

   void erase(iterator a, iterator b)
   {
      iterator t1=a, t2=b;

      while(t1 != t2)
      {
	rep_type::iterator it = t1.rep_;
	CursorWindow::viewport *p = *it;
	delete p;
	++t1;
      }

      rep_.erase(a.rep_, b.rep_);
   }

   CursorWindow::viewport& operator[] (int i)
   {
      CursorWindow::viewport* ptr = rep_[i];

      return *ptr;
   }

   void push_back(std::auto_ptr<CursorWindow::viewport> r)
   {
      CursorWindow::viewport* tmp = r.release();

      rep_.push_back( tmp );
   }

   ~PortVector()
   {
     rep_type::iterator f = rep_.begin(),
			l = rep_.end();
		
     while(f != l)
     {
       CursorWindow::viewport* p = *f++;
       delete p;
     }

   }

   iterator begin() { return rep_.begin(); }
   iterator end()   { return rep_.end();   }

   CursorWindow::viewport* back() { return rep_.back(); }

   size_t size() const { return rep_.size(); }


};


static void box_viewport(CursorWindow::viewport* w)
{
      //
      // put a border around it using # as the fill character
      //

      w->set_curpos(0,0);
      w->fill_to_eol('#');

      w->set_curpos(w->size().row_-1,0);
      w->fill_to_eol('#');

      int i;

      for(i=0; i < w->size().row_; ++i)
      {
	 w->set_curpos(i, 0);
	 w->write('#');
	 w->set_curpos(i, w->size().col_-1);
	 w->write('#');
      }
}


class RepaintHandler
: public CursorWindow::viewport::repaint_handler
{
public:

   char const **layout_;  // pictures of what the screen should
			  // look like
			
   RepaintHandler(char const **l)
   : layout_(l)
   {
   }

    void operator() (CursorWindow::viewport* w, int cmd)
    {
      //
      // erase the whole viewport
      //

      w->set_curpos(0,0);
      w->fill_to_eos();

      box_viewport(w);

      // now, draw a description of what the screen should
      // look like in its current state

      char const **p = layout_;

      int i = 2;

      while(*p)
      {
	char const *line = *p;
	
	w->set_curpos(i,2);
	w->write(line);
	++i;
	++p;
	
      }

      w->set_curpos(i,2);

      {
	std::list<CursorWindow::viewport> vps;
	
	int count = w->window()->get_viewports(&vps);
	
	*w << "viewport count is " << count;
	
      }


    }
};

class Vhandler
: public CursorWindow::viewport::repaint_handler
{
public:

    void operator() (CursorWindow::viewport* w, int cmd)
    {
      w->set_curpos(0,0);
      w->fill_to_eos();

      box_viewport(w);

      w->set_curpos(w->size().row_/2,w->size().col_/2);
      w->write("V");
    }
};

class Ohandler
: public CursorWindow::viewport::repaint_handler
{
public:

    void operator() (CursorWindow::viewport* w, int cmd)
    {
      w->set_curpos(0,0);
      w->fill_to_eos();

      box_viewport(w);

      w->set_curpos(w->size().row_/2,w->size().col_/2);
      w->write("O");
    }
};


int main()
{
  CursorWindow w;
  w.open();

  {
     // create a single viewport and display its expected appearance
     //

     static char const *l[] =
     {
	 "+---+  Press space to delete this viewport",
	 "|   |  and precede to the next test.",
	 "+---+",
	 "       Keep pressing space to continue through the",
	 "       create/delete of more viewports.",
	 "       After you press space now, you should see NO viewports",
	 "       and when you press space again, you will be given examples",
	 "       of what the screen should look like AFTER you press space",
         0
     };

     PortVector ports;

     RepaintHandler h(l);

     ports.push_back(w.new_viewport(&h) );

     ports.back()->activate();

     while( w.read_input().value_ != ' ');  // wait for a blank
  }

   w.set_curpos(0,0);
   w.fill_to_eos();

   w.set_curpos(2,0);
   w.write("You should only see this text");

   {
      std::list<CursorWindow::viewport> vps;

      int count = w.get_viewports(&vps);

      w.set_curpos(3,0);

      w << "The count of viewports is " << count;
   }

   while( w.read_input().value_ != ' ');  // wait for a blank



  {
     // create a single viewport and display its expected appearance
     //

     static char const *l[] =
     {
	 " +--+   You should see windows as described to the left.",
	 " |  |   If so, press space to continue, and the viewport labeld V",
	 " +--+   will be deleted",
	 " |V |",
	 " +--+",
	 0
     };

     PortVector ports;

     RepaintHandler h(l);
     Vhandler       vh;

     ports.push_back(w.new_viewport(&h));
     ports.push_back(ports[0].vsplit(&vh));

     ports[0].activate();
     ports[1].activate();

     while( w.read_input().value_ != ' ');  // wait for a blank

     static char const *after[] =
     {
	 " +--+  Now, you should just see this",
	 " |  |",
	 " +--+",
	 0
     };

     h.layout_ = after;

     ports.erase( ports.begin() + 1 );

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back(ports[0].hsplit(&vh));
      ports.back()->activate();

     static char const *hsplit1[] =
     {
	 " +-+-+  Now, you should just see this, until",
	 " | |V|  you press space, then you should see",
	 " +-+-+  only the V -- verify that then press space",
	 "        again",
	 0
     };

     h.layout_ = hsplit1;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin() );

     while( w.read_input().value_ != ' ');  // wait for a blank

     static char const *hsplit2[] =
     {
	 " +-+-+  Now, you should just see this, until",
	 " |V| |  you press space, then you should see",
	 " +-+-+  only the V -- verify that then press space",
	 "        again",
	 0
     };

     h.layout_ = hsplit2;

     ports.push_back(ports[0].hsplit(&h));
      ports.back()->activate();

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin() + 1);

     while( w.read_input().value_ != ' ');  // wait for a blank

     Ohandler  o;

     ports.push_back(ports[0].hsplit(&h));
      ports.back()->activate();
     ports.push_back(ports[0].vsplit(&o));
      ports.back()->activate();

     static char const *vhsplit1[] =
     {
	 " +---+---+   You should see this until you press space",
	 " | V |   |",
	 " +---+   |",
	 " | O |   |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit1;

     while( w.read_input().value_ != ' ');  // wait for a blank

     static char const *vhsplit2[] =
     {
	 " +---+---+   Now ou should see this",
	 " |   |   |",
	 " | V |   |",
	 " |   |   |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit2;

     ports.erase( ports.end() -1 );

     while( w.read_input().value_ != ' ');  // wait for a blank

     h.layout_ = vhsplit1;  // again

     ports.push_back( ports[0].vsplit(&o) );

     while( w.read_input().value_ != ' ');  // wait for a blank

     static char const *vhsplit3[] =
     {
	 " +---+---+   Now you should see this",
	 " |   |   |",
	 " | O |   |",
	 " |   |   |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit3;

     ports.erase( ports.begin() );

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.end() -1 );

     h.layout_ = after;  // again

     while( w.read_input().value_ != ' ');  // wait for a blank


     ports.push_back( ports[0].hsplit(&o) );
     ports.push_back( ports[0].vsplit(&vh) );

     static char const *vhsplit4[] =
     {
	 " +---+---+   Now you should see this",
	 " |   | O |   (press space)",
	 " +---+   |",
	 " | V |   |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit4;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin() +1 );

     static char const *vhsplit5[] =
     {
	 " +-------+   Now you should see this.",
	 " |       |   When you press space you should",
	 " +-------+   be left with just the V.  Press",
	 " |   V   |   space, verify this, then press",
	 " +-------+   space again.",
	 0
     };

     h.layout_ = vhsplit5;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin());

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back( ports[0].hsplit(&o) );
     ports.push_back( ports[1].vsplit(&h) );


     static char const *vhsplit6[] =
     {
	 " +---+---+   Now you should see this.",
	 " |   | O |   Press space",
	 " | V +---+   (vsplit6)",
	 " |   |   |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit6;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin() );

     static char const *vhsplit7[] =
     {
	 " +---+---+   Now you should see this.",
	 " |   O   |   Press space -- then you should",
	 " +-------+   just see the O.  Then press space",
	 " |       |   again. (vhsplit7)",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit7;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.end() -1 );

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back( ports[0].hsplit(&vh) );
     ports.push_back( ports[0].vsplit(&h)  );

     static char const *vhsplit8[] =
     {
	 " +----+----+ Now you should see this",
	 " | O  |    | until you press space",
	 " +----+  V |",
	 " |    |    |",
	 " +----+----+",
	 0
     };

     h.layout_ = vhsplit8;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin() + 1 );

     static char const *vhsplit7a[] =
     {
	 " +---+---+   Now you should see this.",
	 " |   O   |   Press space (vhsplit7a)",
	 " +-------+",
	 " |       |",
	 " +---+---+",
	 0
     };

     h.layout_ = vhsplit7a;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin(), ports.end() );

     w.set_curpos(0,0);
     w.fill_to_eos();

     w.set_curpos(0,0);
     w.write("You should see only this text -- press space");

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back( w.new_viewport(&h) );  ports.back()->activate();
     ports.push_back( ports[0].vsplit(&h));  ports.back()->activate();
     ports.push_back( ports[1].hsplit(&vh)); ports.back()->activate();
     ports.push_back( ports[1].vsplit(&o));  ports.back()->activate();

     static char const *complex1[] =
     {
	 " +----------+  You should see this, press space",
	 " |          | ",
	 " +-----+----+ ",
	 " |     |    | ",
	 " +-----+ V  | ",
	 " |  O  |    | ",
	 " +-----+----+ ",
	 0
     };


     h.layout_ = complex1;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin() + 2);

     static char const *complex2[] =
     {
	 " +----------+  You should see this, press space",
	 " |          | ",
	 " +----------+ ",
	 " |          | ",
	 " +----------| ",
	 " |     O    | ",
	 " +----------+ ",
	 0
     };

     h.layout_ = complex2;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin() + 1, ports.end());

     static char const *complex3[] =
     {
	 " +----------+  You should see this, press space",
	 " |          | ",
	 " |          | ",
	 " +----------+ ",
	 0
     };

     h.layout_ = complex3;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back(ports[0].hsplit(&h));  // 2 ports
       ports.back()->activate();
     ports.push_back(ports[0].vsplit(&o));  // 3 ports
       ports.back()->activate();
     ports.push_back(ports[0].hsplit(&vh)); // 4 ports
       ports.back()->activate();

     static char const *complex4[] =
     {
	 " +--+--+----+  You should see this, press space",
	 " |  |V |    | ",
	 " +--+--+    | ",
	 " |  O  |    | ",
	 " +----------+ ",
	 0
     };

     h.layout_ = complex4;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin() + 2);

     static char const *complex5[] =
     {
	 " +--+--+----+  You should see this, press space",
	 " |  |  |    | ",
	 " |  |V |    | ",
	 " |  |  |    | ",
	 " +--+--+----+ ",
	 0
     };

     h.layout_ = complex5;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.push_back( w.new_viewport(&h) );
       ports.back()->activate();

     static char const *complex6[] =
     {
	 " +----------+  You should see this, press space",
	 " |          |  (complex6)",
	 " |          |",
	 " |          |",
	 " +----------+",
	 0
     };

     h.layout_ = complex6;

     ports.back()->activate();


     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin());

     static char const *complex7[] =
     {
	 " +----------+  You should STILL see this, press space",
	 " |          |  (complex7)",
	 " |          |",
	 " |          |",
	 " +----------+",
	 0
     };

     h.layout_ = complex7;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase( ports.begin() + ports.size() -1 );

     static char const *complex8[] =
     {
	 " +--+----+  You should see this, press space",
	 " |  |    |  (complex8)",
	 " |V |    | ",
	 " |  |    | ",
	 " +--+----+ ",
	 0
     };

     h.layout_ = complex8;

     while( w.read_input().value_ != ' ');  // wait for a blank

     {
	int left=0;
	
	if(ports[left].origin().col_ != 0)
	{
	  left = 1;
	}



	CursorWindow::row_col origin(ports[left].origin());
	CursorWindow::row_col size  (ports[left].size());
	
	ports.push_back(w.new_viewport(&h));
	  ports.back()->activate();
	
	ports.back()->move(origin, size);
	  ports.back()->activate();

	
     }

     static char const *complex9[] =
     {
	 " +--+----+  You should see this, press space",
	 " |  |    |  (complex9)",
	 " |  |    | ",
	 " |  |    | ",
	 " +--+----+ ",
	 0
     };

     h.layout_ = complex9;

     while( w.read_input().value_ != ' ');  // wait for a blank

     ports.erase(ports.begin(), ports.end());  // make sure the viewports
					       // are deleted before the
					       // repaint handlers descope!

  }


  return 0;
}



