#ifndef __linux__
int please_ignore_this_file_ProcessViewer;
#else
// unix only app

#include <cxxtls/ProcessViewer.h>
#include <cxxtls/controlledprocess.h>
#include <cxxtls/viewermanager.h>
#include <cxxtls/file.h>
#include <cxxtls/foreach.h>
#include <ctype.h>
#include <cxxtls/deltaviewer.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

using namespace std;

namespace cxxtls
{

static Mutex processRunningGuard;

bool processRunning=false; // only one allowed at a time.  Flag means "process SHOULD be running", not IS running


//===========================================================================
//===========================================================================
//===========================================================================

ProcessViewer::ProcessViewer(ViewerManager *vm, std::string const & cmdLine)
:  Viewer(vm)
,  vm_(vm)
,  cmdLine_(cmdLine)
,  topVisibleLine_(0)
,  leftOffset_(0)
,  repaintMode_(0)
,  threadRunning_(0)
,  threadComplete_(0)
,  handlingGdbEmacsAnnotation_(false)
,  prevOutputCharacter_(0)
,  running_(false)
{
    memset(&mostRecentChars_[0],0, sizeof(mostRecentChars_));

    bool gdbMode = false;

    if(cmdLine_.empty())
    {
       cmdLine_ = "/bin/bash";

       startDir_ = "/tmp";

    }
    else
      {
        FileName f(cmdLine_);

        startDir_ = f.dirname();

        if(f.extension() == ".gdbMode")
          {
            size_t lastDot = cmdLine_.find_last_of('.');
            cmdLine_ = cmdLine_.substr(0,lastDot);

            cmdLine_ = "gdb " + cmdLine_ + " 2>&1";

            gdbMode = true;
          }
      }

    process_.reset(  ControlledProcess::start(cmdLine_, ControlledProcess::LINE_BUFFERED, ControlledProcess::CHARACTER_BUFFERED) );

    if(!process_->status() == ControlledProcess::good)
    {
    }
    else
    {
        {
            ScopedMutexLock lock(processRunningGuard);
        
            processRunning = true;
        }

        if(gdbMode)
        {
            fputs("set pagination off \n",  process_->process_stdin());
            fputs("set prompt (<gdb>) \n",  process_->process_stdin());  // keep the spaces exactly as they are in this command!
            fputs("set annotate 1\n",       process_->process_stdin());  // make gdb send us \n\032\032/path/file:line...\n msgs

            char buffer[80];

            snprintf(buffer, 80, "echo GDB PID=%d\n", process_->process_id());

            fputs(buffer, process_->process_stdin());

            running_ = true;
        }

        startThread();
        
    }

}

//===========================================================================
//===========================================================================
//===========================================================================


ProcessViewer::~ProcessViewer()
{
   ScopedMutexLock lock(processRunningGuard);

   processRunning = false;

}

//===========================================================================
//===========================================================================
//===========================================================================

void 
ProcessViewer::
operator() ( CursorWindow::viewport *     vp,
             int                          cmd
           )
{
    repaintMode_ = cmd;

    // paint the entire screen on command

    paintPage(vp);

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
paintPage( CursorWindow::viewport *vp)
{
    paintTitleRegion(vp);
    paintOutputRegion(vp);
    paintCmdRegion(vp);

}

//===========================================================================
//===========================================================================
//===========================================================================

int
ProcessViewer::
pageHeight(CursorWindow::viewport *vp)
{
   CursorWindow::row_col  size = vp->size();

   return size.row_;
}


//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
paintTitleRegion(CursorWindow::viewport *vp)
{
    vp->set_curpos(0,0);
    
    vp->set_text_attribute(CursorWindow::inactive_title_att);
    if(running_)
         vp->write("Running:");
    else
         vp->write("Stopped:");

    
    vp->set_text_attribute(CursorWindow::active_title_att);
    vp->write("  ");
    vp->write(cmdLine_);
    vp->fill_to_eol();
    vp->set_text_attribute(CursorWindow::normal_att);

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
paintOutputRegion(CursorWindow::viewport *vp)
{
    int i;

    int usable = usableLines(vp);

    for(i = 0; i < usable; ++i)
    {

        size_t lineToDraw = topVisibleLine_ + i;

        if(lineToDraw >= outputLines_.size())
          break;

        string const &line = outputLines_[lineToDraw].first;

        vp->set_curpos(CMD_REGION_HEIGHT+i,0);

        vp->write(line); 
        vp->fill_to_eol();

    }

    curpos_.row_ = CMD_REGION_HEIGHT+i;
    curpos_.col_ = 0;

    while(i < usable)
    {
        vp->set_curpos(CMD_REGION_HEIGHT+i,0);
        vp->fill_to_eol();
        ++i;
    }
   
}


//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
paintCmdRegion(CursorWindow::viewport *vp)
{
     int cmdRow = pageHeight(vp) - CMD_REGION_HEIGHT;

     vp->set_text_attribute(CursorWindow::bottom_att);

     vp->set_curpos(cmdRow,0);
     vp->fill_to_eol();
     
     vp->set_curpos(cmdRow,0);

     {
       char buffer[140];

       LineRange commandRange = findMostRecentCommandRange();

       commandRange = commandRange; // suppress unused variable warning during the debug process.

       size_t curLineIndex = outputLines_.size();

       if(curLineIndex > 0)
           curLineIndex -= 1;

       // snprintf(buffer, sizeof(buffer), "%ld[%ld-%ld] cmd: ", curLineIndex, commandRange.first, commandRange.second);

       snprintf(buffer, sizeof(buffer), "cmd: ");


       vp->write(buffer);
     }

     vp->write(inputBuffer_);

     CursorWindow::row_col here = vp->curpos();

     vp->fill_to_eol();

     vp->set_curpos(here);
     vp->refresh();

     vp->set_text_attribute(CursorWindow::normal_att);

     

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
addOutputRegionCharacter(CursorWindow::viewport* vp, char c, bool interpretCharacter)
{

    if(handlingGdbEmacsAnnotation_)
    {
        // Ee have read \n \032 \032 from gdb which means that we are about to receive
        // a gdb  emacs annotation line.  This means that we must either accumulate new
        // characters into the annotation buffer, or if we have hit the end of the line,
        // we will process the annotation and go back to the default state.

        if(c == ' ')  // scoop up everything to the "(<gdb>) "
        {
            // this annotation line is complete

            handlingGdbEmacsAnnotation_ = false;
            prevOutputCharacter_ = 0;

            // populate currentFile_ and currentLine_ from handlingGdbEmacsAnnotation_

            StrTool::stringlist_t words;

            StrTool::Is_Delim  delim(':');

            int rc= StrTool::parse_words(gdbEmacsAnnotation_, &words, 100, delim);

            gdbEmacsAnnotation_ = "";


            if(rc < 2)
            {
                vp->beep();
                return;
            }

            string filename = words.front();

            StrTool::stringlist_t::iterator f = words.begin();

            ++f;

            int line=0;

            sscanf(f->c_str(), "%d", &line);


            if(line != 0)
            {

                if(filename != currentFile_)
                   vp = vm_->unify();

                currentLine_ = line;
                currentFile_ = filename;

                running_ = false;
                paintTitleRegion(vp);


                showCurrentFile();

            }
            else
                vp->beep(); 

            return;

        }

        gdbEmacsAnnotation_ += c;  // just accumulate the annotation line
        return;

    }
    else
    {
       // we are not currently handling gdb emacs annotations, but we might (or might not) have received a character that is poignant with respect
       // to that process.  We will detect that situation here.

       if(c == '\032')
       {
           if(prevOutputCharacter_ == '\n')
           {
               prevOutputCharacter_ = c;    // we have found \n \032 -- so we should now begin waiting for a trailing \032
               return ;
           }
           else
           if(prevOutputCharacter_ == '\032')
           {
               // here, we have the complete sequence

               handlingGdbEmacsAnnotation_ = true;
               prevOutputCharacter_ = 0;
               return;

           }  
           else
           {
               prevOutputCharacter_ =  c;
               return;  // don't display it
           }
       }
       else
           prevOutputCharacter_ = c;  // don't know what it is, so save it for later

    }



    int cmdRow = pageHeight(vp) - CMD_REGION_HEIGHT;

    if(curpos_.row_ >= cmdRow)
    {
       ++topVisibleLine_;
       curpos_.row_ = cmdRow-1;
       paintPage(vp);
    }

    if(interpretCharacter)
        pushMRC(c);                 // handle_event can call this function but its characters should not be
                                    // interpreted as true program output

    vp->set_curpos(curpos_);        // position the curswor properly for painting
    
    if(c == '\t')
    {
       addOutputCharacter(' ');

       vp->write(' ');
       ++curpos_.col_;
    
       while((curpos_.col_ % 8) != 0)
       {
          vp->write(' ');
          lastLine() += ' ';

          ++curpos_.col_;
       }
    
    }
    else
    if(c == '\n')
    {

       addLineInfo("");

       vp->fill_to_eol();

       if((curpos_.row_ + 1) == pageHeight(vp) - CMD_REGION_HEIGHT)
       {
           ++topVisibleLine_;
           curpos_.col_ = 0;

           paintPage(vp);
       }
       else
       {
           curpos_.row_++;
           curpos_.col_ = 0;

           vp->set_curpos(curpos_);
           vp->fill_to_eol();
           curpos_.col_ = 0;
           vp->set_curpos(curpos_);
   

       }

    }
    else
    {
       addOutputCharacter(c);

       vp->write(c);
       ++curpos_.col_;
   
       if(interpretCharacter && mostRecentCharsIsCommandPrompt())
       {
           running_ = false;
           paintTitleRegion(vp);

           markOutputLineAsCommand();

           addLineInfo("");

           ++curpos_.row_;
           curpos_.col_ = 0;
   
           if(curpos_.row_ >= (pageHeight(vp) - CMD_REGION_HEIGHT))
           {
               ++topVisibleLine_;
               --curpos_.row_;       // we are staying on the last line but scrolling
                                     // the text around it
               paintPage(vp);
           }
           else
               paintCmdRegion(vp);  // just to get the cursor back where
                                    // it belongs


           interpretCommandRegion(vp);
       }
           
    }

}


//===========================================================================
//===========================================================================
//===========================================================================

bool 
ProcessViewer::
handle_event( CursorWindow::input_event const * event,
              CursorWindow::viewport          * vp
            )
{

    // handle events from the user, as well as events coming from the controlled
    // process.  The event->type_ field tells us where the data is coming from.


   if(event->type_ == CursorWindow::input_event::ForceExitKey)
   {
      vp->set_curpos(0,0);
      vp->fill_to_eos('@');
      vp->refresh();

      kill();
      return true;
   }


   char c = event->value_;

   if(c == '\r')
       c = '\n';

   bool processOutputData = (event->type_ == CursorWindow::input_event::ProcessOutput);

   // we must echo the characters ourselves.

   if(processOutputData)
   {
       addOutputRegionCharacter(vp, c);
       return false;
   }
   else
   {
     // the character we are processing came from the user

     if(threadComplete())
       return true;               // the process died.  maybe we don't want to just quit here, but
                                  // we should at least set a flag letting us know that no more output
                                  // is coming.

     if(event->type_ == CursorWindow::input_event::FunctionKey)
     {
         switch(event->value_)
         {
             case CursorWindow::key_bs:
                 c = 0x08;
                 break;

             case CursorWindow::key_next:
                   pageDown(vp);
                 return false;

             case CursorWindow::key_prior:
                   pageUp(vp);
                 return false;

             case CursorWindow::key_up:
                   lineUp(vp);
                 return false;

             case CursorWindow::key_down:
                   lineDown(vp);
                 return false;

         }
     }



     if(c == 0x03)
     {

         inputBuffer_.clear();
         
         paintCmdRegion(vp);

         addOutputRegionCharacter(vp, '^', false);
         addOutputRegionCharacter(vp, 'C', false);

         ::kill(process_->process_id(), 1);


     }
     else
     if(c == 0x0b)        // ^K
     {
          // kill the running process and exit

          vp->set_curpos(0,0);
          vp->fill_to_eos('@');
          vp->refresh();

          kill();
          return true;
     }
     else
     if(c == '\n')
     {
         if(!inputBuffer_.empty())
         {
             CXXTLS_FOREACH(char ch, inputBuffer_)
             {
                addOutputRegionCharacter(vp, ch, false);  // copy the user's command region into the output region
                                                          // to create a permanent (kind of anyway) record of the
                                                          // command so that when you scroll back through the output
                                                          // region, you can see the command intermingled with gdb's behavior
             }
             
             addOutputRegionCharacter(vp, '\n', false);
         }

         if(inputBuffer_ == "?")
         {
             LineRange  range = findMostRecentCommandRange();

             while(range.first < range.second)
             {
                 string line = outputLines_[range.first++].first;

                 addOutputRegionCharacter(vp, '>', false);

                 CXXTLS_FOREACH(char ch, line)  addOutputRegionCharacter(vp, ch, false);

                 addOutputRegionCharacter(vp, '\n', false);
             }

         }
         else
         {
      
              // write the user's command input region to the controlled process as a command.
              running_ = true;
              paintTitleRegion(vp); // to show the program as "running"

              fputs(inputBuffer_.c_str(), process_->process_stdin());
              fputc('\n', process_->process_stdin());
              fflush(process_->process_stdin());
      
              inputBuffer_.resize(0);
      
              // redraw the screen to show user all this crap
         }

         paintCmdRegion(vp);

     }
     else
     if(c == 0x08)
     {
         if(!inputBuffer_.empty())
         {
            inputBuffer_.erase( inputBuffer_.size() - 1 );
         
            paintCmdRegion(vp);
         
         }
     }
     else
     if(c >= ' ' && c <= '~')
     {
         // this is just one more character in the text to be sent as a command to the
         // child process

         inputBuffer_ += c;

         paintCmdRegion(vp);

     }
     else
         vp->beep(); // unknown character

     return false;
   }


   return false;
}


//===========================================================================
//===========================================================================
//===========================================================================

Viewer* ProcessViewer::app(ViewerManager*vm, std::string &fullname)
{
    // create a new viewer of this type and return it

    return new ProcessViewer(vm, fullname);
}

//===========================================================================
//===========================================================================
//===========================================================================

std::string ProcessViewer::app_name("ProcessViewer");

//===========================================================================
//===========================================================================
//===========================================================================

std::string ProcessViewer::description() const { return "Run a child process"; }

//===========================================================================
//===========================================================================
//===========================================================================

void ProcessViewer::help() 
{
}

//===========================================================================
//===========================================================================
//===========================================================================

void*  
ProcessViewer::
consumeSlaveData(void *parm)
{
    // read bytes from the slave process and append them to the
    // stdout_ queue.
    //
    // Note:  the stop() function will close the process_stdout()
    // so when that happens, this thread

    ProcessViewer *f = reinterpret_cast<ProcessViewer*>(parm);


    {
       ScopedMutexLock lock(f->threadDataGuard_);

       f->threadRunning_ = true;   // thread is actually running

    }

    char c;

    // the following code will be fine under normal circumstances
    // but if the user closes the process, this code may get a 
    // false read or other race condition.
    //
    // we should really be using an atomic<FILE*>. 


    for(;;)
    {
        ScopedMutexLock lock(processRunningGuard);

        if(!processRunning)
           break;

        if(!(EOF != (c = f->process_->readStdoutByte())))
           break;

        f->stdout_.push(c);
    }

    f->setThreadComplete();

    return 0;
}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::startThread()
{
   // this function is called by ViewerManager::registerProcessViewer();

   createThread(consumeSlaveData, this, false); /// thread is detached,
                                                /// no joinable.
}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
setThreadComplete()
{
   ScopedMutexLock lock(threadDataGuard_);

   threadRunning_ = false;
   threadComplete_= true;

}

//===========================================================================
//===========================================================================
//===========================================================================

bool
ProcessViewer::
threadRunning()
{
   ScopedMutexLock lock(threadDataGuard_);

   return threadRunning_;

}

//===========================================================================
//===========================================================================
//===========================================================================

bool
ProcessViewer::
threadComplete()
{
   ScopedMutexLock lock(threadDataGuard_);

   return threadComplete_;

}


//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
pageDown(CursorWindow::viewport *vp)
{
    size_t availableLines = outputLines_.size();

    size_t nextPageStart = topVisibleLine_ + usableLines(vp);

    if(nextPageStart >= availableLines)
    {
       // paged off the bottom of the existing data

       if(outputLines_.empty())
         return ;  // nothing to show

       nextPageStart = availableLines - usableLines(vp);

       if(nextPageStart > availableLines)
       {
          // page size is bigger than the number of lines available

          nextPageStart = availableLines -1;
       }

    }

    topVisibleLine_ = nextPageStart;

    paintPage(vp);

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
pageUp(CursorWindow::viewport *vp)
{
    size_t availableLines = outputLines_.size();
    size_t usableHeight   = usableLines(vp);

    size_t nextPageStart = topVisibleLine_ - usableHeight;

    if(nextPageStart > availableLines)
    {
       // we have paged off the top

       nextPageStart = 0;
    }

    topVisibleLine_ = nextPageStart;

    paintPage(vp);

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
lineDown(CursorWindow::viewport *vp)
{
    if(outputLines_.empty())
        return;

    size_t availableLines = outputLines_.size();

    size_t nextPageStart = topVisibleLine_ + 1;

    if(nextPageStart >= availableLines)
    {
       // paged off the bottom of the existing data

       if(outputLines_.empty())
         return ;  // nothing to show

       nextPageStart = topVisibleLine_ + 1;

       if(nextPageStart > availableLines)
       {
           return;
       }

    }

    topVisibleLine_ = nextPageStart;

    paintPage(vp);

}

//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
lineUp(CursorWindow::viewport *vp)
{
    size_t availableLines = outputLines_.size();

    if(topVisibleLine_ == 0)
       return;

    size_t nextPageStart = topVisibleLine_ - 1;

    if(nextPageStart > availableLines)
    {
       // we have paged off the top

       nextPageStart = 0;
    }

    topVisibleLine_ = nextPageStart;

    paintPage(vp);

}


//===========================================================================
//===========================================================================
//===========================================================================

ProcessViewer::LineRange
ProcessViewer::findMostRecentCommandRange()
{
    
       size_t cur = outputLines_.size();

       size_t i;

       // find the most recently containing the (<gdb>) string

       for(i = cur;  i > 0; )
       {
          --i;  

          if(outputLines_[i].second)
             break;  // found the most recent (<gdb>) line -- should be one less than the most recent line, but who knows

       }

       cur = i;  // save this line as the upper bound of command interpretation

       // now search to the previous (<gdb>) line

       for(i = cur; i > 0; )
       {
           --i;

           if(outputLines_[i].second)
              break;
       }

       if(i < outputLines_.size())
       {
           ++i;
       }
       else
       if(i == 0)
       {
          i = cur;
       
       }

       return LineRange(i, cur);


}


//===========================================================================
//===========================================================================
//===========================================================================

void 
ProcessViewer::
interpretCommandRegion(CursorWindow::viewport *vp)
{

    size_t oldDisplayVariableCount = displayVariables_.size();

    displayVariables_.clear();

    LineRange  range = findMostRecentCommandRange();

    narrowLineRange(range);

    bool updateCompanionPage = false;

    while(range.first != range.second)
    {
        updateCompanionPage |= parseAndInterpretLineRange(vp, range);
    }

    if(oldDisplayVariableCount || !displayVariables_.empty())
         updateDisplayVariablesCompanionWindow();

    if(updateCompanionPage)
    {
         showCurrentFile();
    }

}

//===========================================================================
//===========================================================================
//===========================================================================

void 
ProcessViewer::
narrowLineRange(LineRange &r)
{

    bool currentLineFound = false;

    size_t f = r.second;

    while(f != r.first)
    {
       --f; // work backwards from the end of the range include only those lines which are guaranteed to be
            // part of the output from gdb itself.
            //
            // scan backwards through only the well understood outputs from gdb
            //
            //  We include lines like the following as possible command outputs:
            //
            //   ^anything   ) at filename:lineNumber$
            //
            //   ^number .*$   // except for breakpoint and diplay variable line output.  
            //
            //   ^Breakpoint number.*$
            //
            ///  ^Misssing.*  (not really a command but doesn't narrow the range)

       StrTool::stringlist_t  words;

       string const &line = outputLines_[f].first;

       int rc = StrTool::parse_words(line, &words);

       if(rc < 2)
       {
           ++f;
           break;
       }

       string const &firstWord = words.front();

       if(firstWord.empty())
          break; // some kind of wierd null string

       char fc = firstWord[0];       // first char of first word

       if(    (fc != 'B')  // as in breakpoint
          &&  !isdigit(fc) // as in a number
          &&  (fc != 'M')  // as in Missing debug infos
         )
       {
           // this is probably some random output from the program being debugged, but there is one exception:
           // When you single step, the debugger both does the command and then immediately prints the location
           //
           //    >step
           //     somefunction (params) at somefile:someline
           //
           // So, before we abandon the line and overly-narrow the range, lets check for ')' 'at' and the colon.
           //       
           
           bool isSingleStepOutput=false;

           if(rc >= 3)
           {
              // possibly single step output

              StrTool::stringlist_t::iterator l = words.end();

              --l; 

              string word = *l; // word is last word right now


              if(    (word.find_first_of(':') < word.size())
                 &&  isdigit(word[word.size()-1])
                )
              {
                 // ok, the last word on the line has a : in it and the last character
                 // of the word is a digit.

                 --l;

                 word = *l;

                 if(word == "at")
                 {
                    // ok, mostly there!  

                    --l;

                    word = *l;

                    if(!word.empty() && word[word.size()-1] == ')')
                    {
                       // we are here:   funcname (parms) at file:number
                       //                               ^

                       isSingleStepOutput = true;
                    }

                 }
              }
           }


           if(!isSingleStepOutput)
           {
               size_t equal = line.find_first_of('=');
               
               if(equal < line.size())
               {
                  // maybe we have a display variable value:
                  //
                  //   name = value

                  addDisplayVariable(line);

                  continue;


               }
               else
               {
                   ++f;  // not a valid command and not a single step output line either
                   break;
               }
           }
       }

       if(isdigit(fc))
       {
          if(currentLineFound)
          {
              // apparently we have found a second single line number in the stream between the
              // previous and current command boundaries (ie lines delimited by (<gdb>)...(<gdb>)
              //
              // so we must have run out of real commands and have now back up up from end of command
              // block into random output from the program.
              // so lets bail here
              ++f;
              break;
          }


          // normally line numbers are presented by gdb whenever it stops, and you can tell whether it is
          // a real line number of a display label by checking for a trailing : on the leading number
          // on the line:
          //
          //     1        print("fred")
          //     1:       y  fred
          //     1:       a = b  
          //
          // However, if the user has requested a list of the breakpoints, then you get stuff like this, which
          // confuses the issue:
          //
          //     1    y   n   maybe     main(asdf)  at   file:line
          //
          // So we must check for this pattern before giving up and calling this a good line number
          //
          // However, it could also be a display variable value:
          //
          //     1: a = b

          bool isLineNumber = true;

          if(rc > 4)
          {
              StrTool::stringlist_t::iterator l = words.end();

              --l;

              if(l->find_first_of(':') < l->size())
              {
                  --l;

                  if(*l == "at")
                  {
                       --l;

                       string word = *l;

                       if(!word.empty() && word[word.size()-1] == ')')
                       {
                          // ok, we have the main(asdf) string in the above example

                          ++f;  
                          break;  // this line is not a usable command output
                       }

                  }
              }

          }

          if(firstWord[ firstWord.size() - 1 ] == ':')
          {
              isLineNumber = false; // this line is a display variable output line

              addDisplayVariable(line);

          }

          if(isLineNumber)
               currentLineFound = true;
       }

       
    }

    r.first = f; // clamp to the first valid command output detected
    

}

//===========================================================================
//===========================================================================
//===========================================================================

//
//  interesting output lines from gdb:
//
//   Breakpoint 1 at 0x40362c: file cpptagdb.cxx, line 117.
//   Breakpoint 1, main (argc=1, argv=0x7fffffffdb78, environ=0x7fffffffdb88) at cpptagdb.cxx:117
//   117     calling(fred)
//



bool 
ProcessViewer::
parseAndInterpretLineRange(CursorWindow::viewport *vp, LineRange &range)
{


    bool updateCompanionPage   = false;

    while(range.first != range.second)
    {
        string line = outputLines_[range.first].first;

        ++range.first;

        // eat the range and ignore it for the moment.

        StrTool::stringlist_t  wordlist;

        int rv = StrTool::parse_words(line, &wordlist);

        if( (rv>2) 
            && wordlist.front() == "Breakpoint"
          )
        {
            // handle setting and reaching breakpoints
            //
            //   1.  "Breakpoint 1 at ..." indicates a setting, not a reach
            //   2.  "Breakpointt 1, main(...) at" indicates a read
            

            StrTool::stringlist_t::iterator p = wordlist.begin();

            ++p; // skip Breakpoint

            string w = *p;

            if(    !w.empty()               // bp number exists
               &&  isdigit(w[0])            // it starts with a digit
               &&  w[w.size()-1] == ','     // and ends with a comma
              )
            {
                string l = wordlist.back();  

                size_t colon = l.find_first_of(':');

                if(colon > 0 &&
                   colon < l.size())
                {
                    string line = l.substr(colon+1);

                    if(!line.empty()
                       && isdigit(line[0])
                      )
                    {
                        size_t number=0;

                        sscanf(line.c_str(), "%ld", &number);

                        if(number)
                        {
                             currentFile_ = l.substr(0,colon);
                             currentLine_ = number;
                             updateCompanionPage = true;
                        }

                    }


                }

            }

            // end of breakpoint keyword handling section
        }
        else
        if( (rv >= 1)
            && !wordlist.front().empty()
            && isdigit(wordlist.front()[0])
            && (':' != wordlist.front()[wordlist.front().size()-1])

          )
        {
           // handle line numbers achieved using the 'next' or 'continue' commands
           // and also when breakpoints are reached.
           //
           // Note that lines that beging with "<number>:" are display statement outputs and they don't count

           updateCompanionPage = true;

           size_t number=0;

           sscanf(wordlist.front().c_str(), "%ld", &number);

           currentLine_ = number;

           // end of process of lines beginning with numbers
        }
        else
        if(rv > 3)
        {
           // let us now handle the output from single step operations.  They typically look like this:
           //  >step
           //  funcname (parms) at file:line

           StrTool::stringlist_t::iterator l = wordlist.end(); 
           --l;

           string last = *l;

           size_t colon = last.find_first_of(':');

           if(colon < last.size())
           {
             // there's a colon in the last word

             --l;

             if(*l == "at")
             {
                // we have at filename:line

                --l;

                string word = *l;

                if(!word.empty() && word[ word.size()-1] == ')')
                {
                  // ok, the whole pattern is matched: 
                  //
                  //   func   (parms)   at   file:line
                  //                ^

                  currentFile_ = last.substr(0, colon);

                  word = last.substr(colon+1);

                  size_t line = 0;

                  sscanf(word.c_str(), "%ld", &line);

                  if(line)
                  {
                     currentLine_ = line;
                     updateCompanionPage = true;
                  }


                }

             }
           }

        }

    }




    return updateCompanionPage;
}


void 
ProcessViewer::
showCurrentFile()
{
  FileName name = vm_->find_file_in_standard_locations(currentFile_, startDir_);

  if(name.empty())
  {
    vm_->window()->beep();

    return ;

  }

  Viewer* oldie = vm_->find_viewer(name);

  if(oldie)
  {
    oldie->set_row_col_hint(currentLine_-1,0);

    if(!vm_->isVisible(oldie))
        vm_->vsplit(oldie, name);


    vm_->activate(oldie);
  }
  else
  {
    if(!name.exists())
    {
       CursorWindow::Message m(name + " does not exist");

       m += "Sorry, can't find it";
       m += "";
       m += "Press Enter to continue";

       m.popup(vm_->window());

       return ;
    }

    Viewer *newbie = vm_->edit(name);

    if(!newbie)
      vm_->window()->beep();
    else
    {
      newbie->set_row_col_hint(currentLine_-1, 0);  // 0 based line numbers!
      vm_->vsplit(newbie, name);
    }

  }

  vm_->activate(this);
}


//===========================================================================
//===========================================================================
//===========================================================================

void
ProcessViewer::
kill()
{
          fputs("kill\n", process_->process_stdin());
          fputs("quit\n", process_->process_stdin());
          fflush(process_->process_stdin());
          sleep(1);
       
          process_->stop();

          ::kill(process_->process_id(), 9);

          running_ = false;

}



//===========================================================================
//===========================================================================
//===========================================================================
void
ProcessViewer::
addDisplayVariable(std::string const &line)
{
 
    // this line is a display variable output.  It should look like this:
    //
    //    number:  display expression = value
 
    size_t colon = line.find_first_of(':');
 
    if(colon < line.size())
    {
       // found us one of them thar colons on the darn line, by crackie!
 
       size_t equal = line.find_first_of('=');
 
       if(equal < line.size())
       {
           // so like, dude, like, I mean like, like you got one of them equal signs on your line, or something...
 
           string name  = line.substr(colon+1, equal - colon-1);
           string value = line.substr(equal+1);
 
           displayVariables_.push_back( DeltaViewer::NameValuePair(name,value) );
 
       }
 
    }
    else
    {
       size_t equal = line.find_first_of('=');

       if(equal < line.size())
       {

            string name = line.substr(0, equal);
     
            string value = line.substr(equal+1);
     
            displayVariables_.push_back(DeltaViewer::NameValuePair(name, value));

       }
    }
 
 
}

//===========================================================================
//===========================================================================
//===========================================================================


void
ProcessViewer::
updateDisplayVariablesCompanionWindow()
{

      static char const *dvPage = "GDB Display Variables";

      Viewer* oldie = vm_->find_viewer(dvPage);
    
      if(oldie)
      {
    
        if(!vm_->isVisible(oldie))
            vm_->hsplit(oldie, dvPage);
    
        DeltaViewer *d = dynamic_cast<DeltaViewer*>(oldie);

        if(d)
        {
            d->replace(displayVariables_);
        }
    
        vm_->activate(oldie);
      }
      else
      {
        DeltaViewer *newbie = new DeltaViewer(vm_, dvPage);
    
        if(!newbie)
          vm_->window()->beep();
        else
        {
          vm_->hsplit(newbie, dvPage);
        }
    
      }
    
      vm_->activate(this);
        
 }

//===========================================================================
//===========================================================================
//===========================================================================

  //    NO_BACKGROUND_PROCESS,
  //    REQUESTING_VIEWER_TERMINATION,
  //    NO_BACKGROUND_DATA_AVAILABLE,
  //    BACKGROUND_DATA_AVAILABLE

Viewer::BackgroundProcessStatus 
ProcessViewer::
getBackgroundData(CursorWindow::input_event &e)
{
    if(! stdout().empty() )
    {
        e.type_ = CursorWindow::input_event::ProcessOutput;
             
        char c=0;
      
        stdout().waitPop(c);
      
        e.value_ = c;

        return BACKGROUND_DATA_AVAILABLE;

    }
    else
    if(threadComplete())
    {
         return REQUESTING_VIEWER_TERMINATION;
    }

    return NO_BACKGROUND_DATA_AVAILABLE;
}

//===========================================================================
//===========================================================================
//===========================================================================

}; // namespace cxxtls
#endif
