#ifndef directory_viewer_included
#define directory_viewer_included

#include <cxxtls/cpptagdb.h>
#include <cxxtls/viewermanager.h>
#include <cxxtls/file.h>
#include <portable_io.h>
#include <vector>
#include <list>
#include <set>

namespace cxxtls
{

///@file
///
///   This file defines, DirectoryViewer.  A class used to
///   instantiate a user interface providing directory viewing and
///   modification.

struct DirectoryViewer
: public ListViewer::policies
  //
  ///
  /// An object of the DirectoryViewer class is passed to a ListViewer
  /// at the time of its construction so that the list viewer implements
  /// a directory viewing command interface.
  ///
  /// DirectoryViewer derives from ListViewer::policies and specializes the
  /// event handler function.  It uses the strings_ member of
  /// policies_ to hold the displayable version of text describing
  /// files and their status.  It keeps other another vector of
  /// strings, names_, which contains the raw data of file names.
  /// The strings_ array contains other text beside the file names.
  ///
  /// To use add this class to the list of valid applications for a
  /// a ViewerManager object, do this:
  ///
  ///   ViewerManger vm;
  ///   ...
  ///   vm.add_app(DirectoryViewer::app, DirectoryViewer::app_name);
  ///
  /// This makes the directory viewer avaiable in the application chooser
  /// menu brought up when you split the screen or replace the current
  /// viewer.  The default key bindings for these activities are
  ///
  /// @code
  ///   ^X 2   -- split vertically
  ///   ^X 3   -- split horizontally
  ///   ^X ^N  -- replace the current viewer
  ///
  /// @endcode
  ///
  /// There is of course, no reason to have a DirectoryViewer object
  /// unless you are also creating a ListViewer object because a
  /// DirectoryViewer is nothing more than a set of policies for controlling
  /// a ListViewer's behavior.  There is really no reason to have either
  /// unless you have a ViewerManager because it declares global variables
  /// used by ListViewer.
  ///
  /// Here's how to manually create a DirectoryViewer and pop it up in the
  /// viewer manager:
  ///
  /// @code
  ///
  ///   ViewerManger vm;
  ///   ...
  ///   vm.add_app(DirectoryViewer::app, DirectoryViewer::app_name);
  ///   ...
  ///   FileName curdir="*";
  ///   curdir.convert_to_absolute_path();
  ///   vm.add(new ListViewer(new DirectoryViewer(curdir), &vm), curdir);
  ///
  /// @endcode
  ///
  /// This example both adds the DirectoryViewer class as a valid application
  /// type _and_ it creates a directory viewer for the current directory and
  /// pops it up in the viewer manager.
  ///
  /// The directory viewer prefers absolute pathnames on directories.
  ///
  /// Here are the command keys available as part of the directory viewer
  /// itself.  The scrolling and system control key bindings are handled
  /// by the viewport manager.
  ///
  /// @code
  ///   'q'    -  quit showing this directory, close the vieweport
  ///   Enter  -  show the object under the cursor
  ///   '^'    -  show the parent directory
  ///   'm'    -  mark the item under the cursor
  ///   'M'    -  mark files by filename pattern
  ///   'd'    -  delete the file under the cursor
  ///   'D'    -  delete marked files
  ///   'c'    -  copy the file under the cursor
  ///   'C'    -  copy the marked files
  ///   'r'    -  rename the file under the cursor
  ///   'R'    -  rename marked files
  ///   'a'    -  lets the user change the attributes of the file under
  ///             the cursor
  ///   'l'    -  re-read the directory and redisplay it
  ///   'L'    -  same as 'l', but sort on size
  ///   '.'    -  search for a symbol in a TAGSPP.tagspp file.
  ///   '!'    -  search for a symbol in an etags TAGS file.
  ///   '0'    -  scan the subdirectory for symbols and display them hiearchically
  ///   ^G     -  execute a command on the current file and display the results in a show session
  ///
  /// @endcode

{
  typedef FileName::sorted_names_t   file_name_list;     ///< Storage type for lists of file names
  typedef std::vector<FileName>      file_name_vector;   
  typedef std::vector<std::string>   string_vector;      
  typedef std::vector<FileStatus>    file_status_vector; ///< Vector of operating system file information objects
  typedef std::set<FileName>         marked_files_set;   ///< Set of file names marked by the user
  typedef CursorWindow::input_event  input_event;        ///< user input events


  size_t             marked_line_;
  FileName           fullname_;
  file_status_vector status_;     ///< status_.size() == names_.size() == strings_.size()
  file_name_vector   names_;
  marked_files_set   marks_;

  int                vp_width_;

  bool               user_names_;  ///<  show user id as a name (group too)
  bool               show_ids_;    ///< do not put any id info

  bool is_marked(size_t index);    ///< Is this item marked?
  bool is_special(size_t index);   ///< returns true on directories

  enum sortCode 

    ///  sortCode's are symbolic names for ways in which the directory can be
    ///  sorted.
    { 
      onName=0,   ///<  sort by the file's name
      onSize=1,   ///< sort by the file's size in descending order
      onTime=2
    };
  
  sortCode           sortCode_;    ///< Sort specification for read_directory.

  void remove_item(size_t index);
    ///< delete the file whose name appears at location "index"


  DirectoryViewer(std::string const &fullname,
                bool show_user_names=false,
                bool show_ids=false );

   ///< Construct a DirectoryViewer's command handler -- and read the
   /// corresponding directory.


  void read_directory();
     ///< This is the function that really reads the directory.  It
     ///  eliminates in extant directory info, and rereads it all.  
     ///  Directory elements are sorted based on the most recent
     ///  change to the class member sortCode_ whose default value
     ///  is onName.
                

  void display_symbolic_link_file_target(size_t index, ViewerManager *vm);
    ///<  Produce a popup message containing the name of the the
    ///   link file target for the specified line index.  If the line
    ///   is not a symbolic link file, then you get an empty string


  void generate_row_text();

    ///< generate a string_ for each name_ in the directory

  void generate_row_text(size_t index);
    ///< update strings_[index] with the correct text for its row


  void handle_resize(CursorWindow::row_col const &vp_size);
    ///< Respond to resize commands.  Note that you may receive extraneous
    ///  resize commands, so save vp_size and ignore duplicate size
    ///  settings.

  int etags_first(ViewerManager* vm);
    ///< Prompt the user for the name of a symbol and bring up an edit session
    ///  on that file using the standard etags database

  int cpptags_first(ViewerManager* vm);
    ///< Prompt the user for the name of a symbol and bring up an edit session
    ///  on that file using the cpptags database, TAGPP.tagpp

  int cpptagviewer(ViewerManager* vm);
    ///< Discover the CppTagDB for the current directory and show it.

  int etags_all(ViewerManager* vm);
     ///< Prompt the user for a symbol name and list all symbols that match it.

  int cpptags_all(ViewerManager* vm);
     ///< Prompt the user for a symbol name and list all file locations that match it.

  int copy_file(size_t index, ViewerManager* vm, FileName target_filename = "");

    ///< Handle user interaction needed for copying a single file.

  int rename_file(size_t index, ViewerManager* vm, FileName target_filename = "");

    ///<  Handle user interaction needed for renaming a single file.

  int copy_files(ViewerManager*vm);

    ///<  Handle user interaction needed for copying all marked files.

  int rename_files(ViewerManager*vm);

    ///<  Handle user interaction needed for renaming all marked files.

  int delete_file(size_t index, ViewerManager* vm, bool verify=true);

    ///<  Handle user interaction needed for deleting the file under the
    ///  cursor.  Returns 0 if everything went ok.  Return 1 if the file
    ///  really didn't get deleted


  int change_mode(size_t index, ViewerManager* vm);

    ///<  Prompt the user for file permissions changes.

  void view_file_or_dir(std::string const &dir, ViewerManager* vm, bool is_dir);

    ///< Ensure that there is a viewer containing the specified directory or file
    /// search criteria.

  void view_dir(size_t index, ViewerManager* vm);

    ///<  Ensure that there is a viewer containing the directory whose name
    ///  is in under the cursor.

  int mark_files_by_pattern(ViewerManager* vm);
    ///< prompt the user for a file name pattern and mark files whose names
    /// match that pattern.

  int delete_files(ViewerManager* vm);
    ///<  Delete all the marked files.


  int go_file(int index, ViewerManager* vm);
     ///<  execute a command on the current file based on its extension
     ///   and display the results in a show session.
     ///
     ///   @param index The index with the names list to the current file.
     ///   @param vm    The viewer manager object to interact with

  int handle_event(CursorWindow::input_event const *e,
                   size_t index,
                   ViewerManager* vm);

    ///<   Handle user events -- and turn them into commands to be
    ///    executed.


  static Viewer* app(ViewerManager*vm, std::string &fullname);

    ///<  Application "hook" function for the DirectoryViewer class.
    ///   Use it as a parameter to add_app, like this:
    ///   @code
    ///    vm.add_app(DirectoryViewer::app, "DirectoryViewer");
    ///   @endcode

  static std::string app_name;

    ///< The name to prompt the user with in the ViewerManager::appChooser
    ///  menu.

  virtual std::auto_ptr< std::list<std::string> > help_text();

  std::string const& application_name() const { return app_name; }

    ///< Return the name of the application to use when describing this
    ///  viewer in the ViewerManager::appChooser or in the
    ///  ViewerManager::ViewerSelector.

  std::string description() const { return fullname_; }


  virtual bool fetchAllLines(std::list< std::string > &output );
         ///<  @see cxxtls::Viewer::fetchAllLines().


  virtual void removeTemporaryFile();
      ///<  Deletes the directory referred to by this object -- called only when
      ///   the isTemporary() flag is set in the ListViewer that owns this object

  int grep(int index, ViewerManager* vm);
      ///<  Prompt the user for grep options and a search string, then perform
      ///   the search and prompt the user with the search results
      ///
      ///   @param index The index with the names list to the current file.
      ///   @param vm    The viewer manager object to interact with

};

} // namespace cxxtls

#endif
