#ifndef gdbviewer_header_included_p
#define gdbviewer_header_included_p

#include <cxxtls/viewer.h>
#include <cxxtls/gdbcontroller.h>

#include <memory>


///@file
///
///  This file declares class cxxtls::GdbViewer.

namespace cxxtls
{

class GdbStackViewer;             // declared below.
class GdbLogViewer;              // declared below
class GdbBreakpointViewer;       // declared below
class GdbDisplayVariableViewer;  // declared below

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class GdbViewer
:  public Viewer
    ///
    ///  A viewer that allows you run gdb as a "mode" of the cursor window terminal that you are using.
    ///  A "mode" is a forced layout of multiple viewers.  That is, when you start the GdbViewer, you allow it to control
    ///  the position and size of windows all windows on the screen.  You can change the layout yourself, 
    ///  but whenever you allow the GdbViewer to run a debugged application, it will re-layout the screen 
    ///  each time the child process stops.
    ///
    ///  As mentioned earlier, the purpose of the GdbViewer is control the execution both of the gdb
    ///  debugger and also a child process that it is debugging.  This is accomplished uging
    ///
    ///  Note that this viewer deals with the following broadcast messages:
    ///
    ///  @li "fileLocation"  -- received and interpreted as a location where a breakpoint should be set.
    ///  @li "breakpoint"    -- sent to indicate that a breakpoint has been set
    ///  @li "stopPoint"     -- sent to inddicate to all interested viewers that the debugger has stopped.
    ///
{
    std::string fullname_; ///<  fullname of the executable being debugged.
    std::string runArgs_;  ///<  target run arguments

    Viewer*     breakpoints_;   ///<  ptr to viewer (owned by ViewerManager) where breakpoints are listed
    Viewer*     variables_;     ///<  ptr to viewer (owned by ViewerManager) where expressions are displayed
    Viewer*     stack_;         ///<  ptr to viewer (owned by ViewerManager) where the stack is displayed
    Viewer*     terminal_;      ///<  ptr to viewer (owned by ViewerManager) where the user can interact with the debugged child.

    bool        active_;        ///<  we have received a 0 command through operator()()
    bool        runnable_;      ///<  fullname_ refers to a runnable executable
    bool        argsSent_;      ///<  arguments have been sent to the executable.  Also implies executable name has been sent.
    bool        firstStart_;    ///<  defaults to true, goes false on first display of the current source file

    GdbController gdb_;         ///<  debugger control interface.

    GdbController::ChildState savedState_; ///<  GdbController::state() value saved after most recent action by user.
                                           ///   When the current gdb controller state is not equal to this, the
                                           ///   function, getBackgroundData, return true and saves the new state.
                                           ///   Note that this function sets 0 as the event value.  The value means nothing,
                                           ///   the fact of the state change is all that matters.

    std::vector<std::string> messages_;    ///<  String to print while we are stopped.  Cleared on any kind of run.

    ListViewer               *stackViewer_;           // display stack frame at stop.          NOT an auto_ptr!  We don't own this.
    GdbStackViewer           *gdbStackViewer_;        // policy object for stackViewer_;       NOT an auto_ptr!  we don't own this
    ListViewer               *logViewer_;             // display program output log file.      NOT an auto_ptr!  We don't own this
    GdbLogViewer             *gdbLogViewer_;          // policy object for logViewer_;         NOT an auto_ptr!  We don't own this
    ListViewer               *breakpointViewer_;      // display breakpoints.                  NOT an auto_ptr!  We don't own this
    GdbBreakpointViewer      *gdbBreakpointViewer_;   // Policy object for breakpointViewer_;  NOT an auto_ptr!  We don't own this
    GdbDisplayVariableViewer *gdbVariableViewer_;     // Policy object for variableViewer_;    NOT an auto_ptr!  We don't own this
    ListViewer               *variableViewer_;        // Display watched variables.            NOT an auto_ptr!  We don't own this

    typedef std::map<std::string /*bpname*/ , std::pair< std::string /*file*/, std::string /*line*/> > ActiveBreakpoints; 

    ActiveBreakpoints activeBreakpoints_;
        ///<  keeps track of file/line of bps by name.  bp names are a string representation of a bp number in gdb


    std::string  showFilename_;   ///<  most recent file "shown" to highlight stop point location
    int          showLine_;       ///<  line within showFilename_

    std::string  stopFilename_;   ///<  most recent stop point filename
    std::string  stopLine_;       ///<  most recent stop point line number

    int          stopFrames_;     ///<  count of stack frames at the most recent stop.  May be zero to indicate
                                  ///   that we don't currently know.

    bool         attachMode_;    ///<   True if this run is an attachment to a running process.  Which means
                                 ///    that we can't display the target output or feed it's stdin.


    std::vector<std::string> types_;        ///<  Types available in this executable

    std::string bpItem_;  ///<  The number of a breakpoint which should appear displayed in the
                          ///   breakpoint viewer.

    typedef std::list< std::pair< std::string, std::string > > history_t;

    history_t  history_; ///< the history of stop points:  (fullname,lineNo) both as strings

public:

    void clearLogViewerPointer()
       // called by the log viewer destructor to make sure we have no pointer to the log viewer.
    {
       logViewer_ = 0;
       gdbLogViewer_ = 0;
    }

    void clearStackViewerPointer()
       // called by the stack viewer destructor to make sure we have no pointer to the log viewer.
    {
       stackViewer_ = 0;
       gdbStackViewer_ = 0;
    }

    void clearBreakpointViewerPointer()
       // called by the breakpoint viewer destructor to make sure we have no pointer to the log viewer.
    {
       breakpointViewer_ = 0;
       gdbBreakpointViewer_ = 0;
    }


    void clearVariableViewerPointer()
       // called by the breakpoint viewer destructor to make sure we have no pointer to the log viewer.
    {
       variableViewer_ = 0;
       gdbVariableViewer_ = 0;
    }



    virtual bool is_dirty() const; ///<  Return true if running or stopped.

    GdbViewer(ViewerManager *vm, std::string const &fullname);
        ///<  Construct a GdbViewer.
        ///
        ///  @param vm         A pointer to the executable's viewer manager
        ///  @param fullname   The full pathname of the executable to debug.

    ~GdbViewer();
        ///<  destructor

    virtual void operator() ( CursorWindow::viewport * vp, int cmd);
        ///< @see cxxtls::Viewer member of the same name.

    virtual BackgroundProcessStatus getBackgroundData(CursorWindow::input_event &e);
        ///< @copydoc cxxtls::Viewer::getBackgroundData

    virtual bool handle_event( CursorWindow::input_event const * e, CursorWindow::viewport* vp);
        ///< @see cxxtls::Viewer member of the same name.

    virtual std::string const &application_name() const;
        ///<  @copydoc cxxtls::Viewer::application_name

    virtual std::string description() const;
        ///<  @copydoc cxxtls::Viewer::description

    virtual void help();  ///< @copydoc  cxxtls::Viewer::help

    static Viewer* app(ViewerManager*vm, std::string &fullname);
          ///<
          ///  Application "hook" function for the GdbViewer class.
          ///  Use it as a parameter to add_app, like this:
          ///
          ///    vm.add_app(TreeViewer::app, "GdbViewer");

     void handleBreakpointViewerDeath();  ///< recreate this viewer if the user kills it
     void handleStackViewerDeath();       ///< recreate this viewer if the user kills it 
     void handleLogViewerDeath();         ///< recreate this viewer if the user kills it 
     void handleVariableViewerDeath();    ///< recreate this viewer if the user kills it 

     void removeBreakpoint(std::string const &bpId);  
         ///<  orchestrate the bp removal and screen udpates.
         ///
         ///   This is accomplished by queing up the activity until the next background data query.  At which point,
         ///   if there is no background data and if the target is stopped, the delete of the breakpoint will be
         ///   triggered -- and then a 1 second delay will occur before the screen gets painted.

     void enableBreakpoint(std::string const &bpId, bool enabled);  
         ///<  orchestrate the setting of the breakpoint's enabled flag and screen udpates.
         ///
         ///   This is accomplished by queing up the activity until the next background data query.  At which point,
         ///   if there is no background data and if the target is stopped, the delete of the breakpoint will be
         ///   triggered -- and then a 1 second delay will occur before the screen gets painted.

     virtual void listen(std::string const &type, std::string const &msg); 
         ///<  respond to "fileLocation" broadcast messages by attempting to set a breakpoint


     void showFileLocation(std::string const &file, int line);
         ///<  Force the display of the specified fiel in the screen location reserved for the current source location.
         ///   This function is meant to be used by daughter viewers to bring up source code.
         ///   Note that this function causes a window refresh.
         ///   
         ///   @param file    The name of some file to show in a window.
         ///   @param line    The line number within the file to show.

      std::string stackFrameParms(std::string const &frameNumber) { return gdb_.stackFrameParms(frameNumber); }
         ///<  Fetch the parameters that were passed to a given stack frame.  Only works if target is stopped.

      std::string stackFrameVariables(std::string const &frameNumber) { return gdb_.stackFrameVariables(frameNumber); }
         ///<  Fetch the local variables in the specified stack frame.  Only works if target is stopped.

      static void displayGdbValue(ViewerManager* vm, std::string const &title, std::string const &itemValue);
                                      ///<  Display a popup window containing user data described by the encodedData string.
                                      ///   @param vm            The viewer manager pointer
                                      ///   @param title         The title for the popup window
                                      ///   @param itemValue     A possibly structured string value produced as the 'value'
                                      ///                        VarObj member from gdb command -data-evalue-expression

    static void displayGdbVariableList(ViewerManager* vm, std::string const &title, std::string const &encodedData);
                                      ///<  Display a popup window containing user data described by the encodedData string.
                                      ///   @param vm            The viewer manager pointer
                                      ///   @param title         The title for the popup window
                                      ///   @param encodedData   The stack-args or variables lists from gdb command -stack-list-*

    std::string evaluateTmpExpression(std::string const &frameDepth, std::string const &expression); 
        ///<  Invoke the 'evaluateTmpExpression' on the gdb_ member object and return its value.
        ///
        ///   @param frameDepth   The stack frame of interest ("0", "1", ...)
        ///   @param expression   The expression to be evaluated.
        ///

    bool getBreakpointCondition(std::string const &bpID, std::string &condition);
        ///<  Get the breakpoint condition expression for the specified breakpoint number.
        ///   @return true to indicate success, and false for failure.
        ///
        ///   @param bpID       The number of the breakpoint (as a string)
        ///   @param condition  A possibly empty string containing the breakpoint expression.

    void  setBreakpointCondition(std::string const &bpID, std::string const &condition);
        ///<  Set or remove the breakpoint condition expression for a specified breakpoint number.
        ///
        ///   @param bpID       A string containing the breakpoint number to be modified.
        ///   @param condition  A possibly empty string containing the breakpoint condition expression.
        ///
        ///   @note  When the condition string is empty, the condition is removed from the breakpoint.

    void watchLocation(std::string const &frame, std::string const &expression, char const *access="");
        ///<  @copydoc cxxtls::GdbController::watchLocation

    void addDisplayExpression(std::string const &frame, std::string const &expression);
        ///<  @copydoc cxxtls::GdbController::createDisplayExpression

    bool  deleteDisplayExpression(std::string const &expression, std::vector<std::string> &remainingStrings);
        ///<  @copydoc cxxtls::GdbController::deleteDisplayExpression
        ///
        /// @param remainingStrings  A vector of strings defining the remaining display expression
        ///                          values after the delete has occurred.



    std::string whatisHelper(std::string const &frame, std::string const &expr)
        ///<  call gdb_.whatis(string) and return the results
        ///
        ///   @return empty string on fail, otherwise the type of the expression
    {
        if(gdb_.state() == GdbController::stopped)
            return gdb_.whatis(frame, expr);

        return "";
    }

    void runHelper(int keycode); 
        ///<  When invoked by a daughter viewer, this function makes the main viewer active and also
        ///   calls the run() function which makes the program run with the specified keycode.
        ///   The allowed keycode values are F5, F10, and F11.

    void focusOnOtherViewer(int keycode);   ///<  A helper function used by daughter classes to redirect focus to
                                            ///   other windows.  The allowed keycodes are as follows:
                                            ///   @li b  -- The breakpoint window
                                            ///   @li d  -- The display variables
                                            ///   @li s  -- The stack
                                            ///   @li v  -- The source code viewer
                                            ///   @li g  -- The main gdb window itself.

    void feedStdin(CursorWindow::viewport *vp);   ///<  Prompt the user for text to write to stdin of the debugged target.
                                                  ///   Also allows the sending of end of file to the target stdin.

    void feedStdinHelper();                       ///<  Called by daughter viewers to both activate the main gdb viewer and
                                                  ///   feed data to stdin.

    void cdHelper();                              ///<  Let daughter viewers change the target process current directory.
    void haltHelper();                            ///<  Let daughter viewers halt the target process.

    void setSaveFile(std::string const &sf);      ///<  Set the GdbController's "save file name".  That is the file in which
                                                  ///   breakpoints are stored on exit.  They are read on entry.

    void setRunArgs(std::string const &args)
       ///<  Set the arguments that will be passed to the debugged target executable.
    {
       runArgs_ = args;
    }

private:
    bool run(CursorWindow::viewport *vp, int keycode, int pid=0);
        ///<  Handle the running of the executable either from an unstarted, started, terminated, or stopped state.
        ///
        ///
        ///   @param vp       a drawing handle
        ///
        ///   @param keycode  The key that the user pressed that led to the
        ///                   decision to run.  One of F11, F10, or F5.
        ///
        ///   @param pid      An optional parameter which defaults to 0.  If the pid is non-zero,
        ///                   and if keycode == CursorWindow::key_f5, then this is assumed to be
        ///                   an attachment to a running process rather than the starting of a new
        ///                   process.
        ///
        ///   @return exitDebugger  If true, the debugger should exit 
        ///                         immediatly and clean up after itself.

    bool isRunning();  
        ///<   @return true if the debugged child process is currently running

    bool isStopped();
        ///<   @return true if the debugged child process is currently stopped at a 
        ///                 breakpoint or because of a signal.

    bool isTerminated();
        ///<    @return true if the debugged child process has finished or died.

    bool isRunnable();
        ///<    @return true if the debugged child process is not running and is not in some error state.

    void repaintTitle(CursorWindow::viewport *vp);
        ///<    Redraw the title region of the debugger main window (basically all of the window since it
        ///     really only a title window anyway...
        ///
        ///  @param vp  The viewport to draw in

    bool handleStateChange(CursorWindow::viewport *&vp);
        ///<   This function is called by handle_event() to deal with state
        ///    changes as reported by gdb.
        ///
        ///    Note that vp is passed by reference so that it can be changed by the function as it runs.
        ///    
        ///    @return true if the window is supposed to close

     void showCurrentFile(CursorWindow::viewport *&vp, std::string const &name, int lineNumber);
        ///<   Bring up an edit sessions on the named file positioned such that the specified line number
        ///    is visible.
        ///
        ///    @param[in,out] vp          A reference to the pointer to the viewport to draw in.
        ///    @param[in]     name        The name of the file to bring edit.
        ///    @param[in]     lineNumber  The number of a line in the file to make visible.
        ///
        ///    @warning  This function makes vp invalid -- then creates a new one.  Don't save the vp then
        ///              expect it to be the same after the call to showCurrentFile.

      void displayTargetOutput();     ///<  populate the gdb target output field with any new output
      void discardDaughterViewers();  ///<  unify screen on *this, then discard logViewer_ and stackViewer_, et al.
      void displayBreakpoints(std::string const &item);      ///<  populate the breakpoints viewer.  Make sure item is visible
      void displayVariables();        ///<  populate the GdbDisplayVariables viewer

      void removeStopPointMarker();   ///<  signal other viewers to remove the stop point markers
      void displayStopPointMarker();  ///<  signal other viewers to display the current stop point marker

     void changeDirectory();  ///< prompt to change directory
     void performHalt();      ///< halt or kill the target process

     void promptAttachPID(CursorWindow::viewport*, ViewerManager*);  ///<  prompt for a pid number to attach to.

};  // class GdbViewer

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class GdbStackViewer
:  public ListViewer::policies
{
     std::string executableName_;                         ///<  name of executable being debugged

     GdbViewer* gdbViewer_;                               ///<  our owner
public:

    ~GdbStackViewer();                                    ///<  destructor

     GdbStackViewer(std::string const &executableName, GdbViewer*viewer);   
         ///<  constructor
         ///
         ///   @param executableName  The name of the executable being debugged
         ///   @param viewer          The gdb viewer that owns *this

     void resetContent(std::vector<std::string> &newContent, Viewer*v); 
         ///< replace the existing stack frame list with new text
         ///
         ///  @param[in]  newContent  The new strings which should be displayed as the program stack frame
         ///  @param[in]  v           The viewer in which to display the content (if viewer visible)
         ///  @param[in]  vp          The viewport in which to display the content (iv viewer visible, nul otherwise)

     std::string const&  application_name() const;        ///<  @copydoc ListViewer::policies::application_name
     std::string         description()      const;        ///<  @copydoc ListViewer::policies::application_name

     virtual int handle_event(CursorWindow::input_event const *e,
                               size_t                           index,
                               ViewerManager*                   vm  
                              );
         ///<  @copydoc ListViewer::policies::handle_event

     bool is_marked(size_t index); ///< @copydoc ListViewer::policies::is_marked

     virtual std::auto_ptr< std::list<std::string> > help_text(); ///<  @copydoc ListViewer::policies::help_text

     std::string const &executableName() const { return executableName_; }


    
 
}; // class GdbStackViewer

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class GdbBreakpointViewer
:  public GdbStackViewer
{
    GdbViewer *gdbViewer_;
public:
    virtual ~GdbBreakpointViewer();

    GdbBreakpointViewer(std::string const &executableName, GdbViewer *gdb)
    :  GdbStackViewer(executableName, gdb)
    ,  gdbViewer_(gdb)
    {
        title_ = "GdbBreakpoints";
    }

     void resetContent(std::vector<std::string> &newContent, Viewer*v, size_t itemIndex); 
         ///< replace the existing stack frame list with new text
         ///
         ///  @param[in]  newContent  The new strings which should be displayed as the program stack frame
         ///  @param[in]  v           The viewer in which to display the content (if viewer visible)
         ///  @param[in]  itemIndex   The index of a breakpoint to guarantee appears in the display

     virtual std::auto_ptr< std::list<std::string> > help_text(); ///<  @copydoc ListViewer::policies::help_text

     virtual int handle_event(CursorWindow::input_event const *e,
                               size_t                           index,
                               ViewerManager*                   vm  
                              );
         ///<  @copydoc ListViewer::policies::handle_event

     virtual void set_row_col_hint(ListViewer* v, size_t line, size_t column, bool repaint, size_t &display_top, size_t &current_row);
         ///<  called to make sure a given location is visible.
         ///   @param[in,out]  v            The viewer being modified
         ///   @param[in]      line         Row offset to make sure is visible with this->strings_.
         ///   @param[in]      column       should be zero
         ///   @param[in]      reapint      Controls whether or not repainting occurs
         ///   @param[in,out]  display_top  a writable refence to ListViewer::display_top_
         ///   @param[in,out]  current_row  a writable refence to ListViewer::current_row_


};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class GdbLogViewer
:  public ListViewer::policies
{
     std::string executableName_;                         ///<  name of executable being debugged
     GdbViewer *gdbViewer_;

public:

    ~GdbLogViewer();                                    ///<  destructor
     GdbLogViewer(std::string const &executableName, GdbViewer *vm);   ///<  constructor

     void appendContent(std::vector<std::string> &newContent, Viewer*v); ///< Append lines to btoom of viewer
                                                                         ///  window and move the cursor to
                                                                         ///  the bottom line.

     std::string const&  application_name() const;        ///<  @copydoc ListViewer::policies::application_name
     std::string         description()      const;        ///<  @copydoc ListViewer::policies::application_name

     int handle_event(CursorWindow::input_event const *e,
                      size_t                           index,
                      ViewerManager*                   vm  
                     );
         ///<  @copydoc ListViewer::policies::handle_event

     bool is_marked(size_t index); ///< @copydoc ListViewer::policies::is_marked

     virtual std::auto_ptr< std::list<std::string> > help_text(); ///<  @copydoc ListViewer::policies::help_text

     std::string const &executableName() const { return executableName_; }

     virtual void set_row_col_hint(ListViewer* v, size_t line, size_t column, bool repaint, size_t &display_top, size_t &current_row);
         ///<  called to make sure a given location is visible.
         ///   @param[in,out]  v            The viewer being modified
         ///   @param[in]      line         Row offset to make sure is visible with this->strings_.
         ///   @param[in]      column       should be zero
         ///   @param[in]      reapint      Controls whether or not repainting occurs
         ///   @param[in,out]  display_top  a writable refence to ListViewer::display_top_
         ///   @param[in,out]  current_row  a writable refence to ListViewer::current_row_
    
 
}; // class GdbLogViewer


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class GdbDisplayVariableViewer
:  public ListViewer::policies
{
     std::string executableName_;                         ///<  name of executable being debugged

     GdbViewer* gdbViewer_;                               ///<  our owner
public:

    ~GdbDisplayVariableViewer();                                    ///<  destructor

     GdbDisplayVariableViewer(std::string const &executableName, GdbViewer*viewer);   
         ///<  constructor
         ///
         ///   @param executableName  The name of the executable being debugged
         ///   @param viewer          The gdb viewer that owns *this

     void resetContent(std::vector<std::string> &newContent, Viewer*v); 
         ///< replace the existing stack frame list with new text
         ///
         ///  @param[in]  newContent  The new strings which should be displayed as the program stack frame
         ///  @param[in]  v           The viewer in which to display the content (if viewer visible)
         ///  @param[in]  vp          The viewport in which to display the content (iv viewer visible, nul otherwise)

     std::string const&  application_name() const;        ///<  @copydoc ListViewer::policies::application_name
     std::string         description()      const;        ///<  @copydoc ListViewer::policies::application_name

     virtual int handle_event(CursorWindow::input_event const *e,
                               size_t                           index,
                               ViewerManager*                   vm  
                              );
         ///<  @copydoc ListViewer::policies::handle_event

     bool is_marked(size_t index); ///< @copydoc ListViewer::policies::is_marked

     virtual std::auto_ptr< std::list<std::string> > help_text(); ///<  @copydoc ListViewer::policies::help_text

     std::string const &executableName() const { return executableName_; }

    
 
}; // class GdbDisplayVariableViewer

}      // namespace cxxtls
#endif // gdbviewer_header_included_p
