#ifndef key_viewer_included
#define key_viewer_included

#include <cxxtls/viewermanager.h>
#include <cxxtls/file.h>
#include <vector>
#include <list>
#include <set>

namespace cxxtls
{

///@file
///
///   Defines KeyViewer a textual application promptin the user for keyboard maps.
///   instantiate a user interface providing display of keys the user has
///   pressed.

struct KeyViewer
: public ListViewer::policies
  //
  ///
  /// An object of the KeyViewer class is passed to a ListViewer
  /// at the time of its construction so that the list viewer implements
  /// a way to see how keys are interpreted.
  ///
  /// KeyViewer derives from ListViewer::policies and specializes the
  /// event handler function.  It uses the strings_ member of
  /// policies_ to hold the displayable version of text describing
  /// keys.
  ///
  /// To use add this class to the list of valid applications for a
  /// a ViewerManager object, do this:
  ///
  ///   ViewerManger vm;
  ///   ...
  ///   vm.add_app(KeyViewer::app, KeyViewer::app_name);
  ///
  /// This makes the directory viewer avaiable in the application chooser
  /// menu brought up when you split the screen or replace the current
  /// viewer.
  ///
  /// There is of course, no reason to have a KeyViewer object
  /// unless you are also creating a ListViewer object because a
  /// KeyViewer is nothing more than a set of policies for controlling
  /// a ListViewer's behavior.  There is really no reason to have either
  /// unless you have a ViewerManager because it declares global variables
  /// used by ListViewer.
  ///
  /// Here's how to manually create a KeyViewer and pop it up in the
  /// viewer manager:
  ///
  ///   ViewerManger vm;
  ///   ...
  ///   vm.add_app(KeyViewer::app, KeyViewer::app_name);
  ///   ...
  ///   FileName curdir="*";
  ///   curdir.convert_to_absolute_path();
  ///   vm.add(new ListViewer(new KeyViewer(curdir), &vm), curdir);
  ///
  /// This example both adds the KeyViewer class as a valid application
  /// type _and_ it creates and pops up a viewer that lets you look at keys.
  ///
{
  bool is_marked(size_t index);    ///< Is this item marked?  Never returns true

  KeyViewer();

  virtual ~KeyViewer();

  void handle_resize(CursorWindow::row_col const &vp_size);
    //
    ///  This is a no-op for the key viewer

  int handle_event(CursorWindow::input_event const *e,
		   size_t index,
		   ViewerManager* vm);
    //
    ///  Handle user events -- no interpretation is done.  All events
    ///  just get logged to the window along with their categorization.
    //


  static Viewer* app(ViewerManager*vm, std::string &fullname);
    //
    ///  Application "hook" function for the KeyViewer class.
    ///  Use it as a parameter to add_app, like this:
    ///
    ///    vm.add_app(KeyViewer::app, "KeyViewer");

  static std::string app_name;
    //
    ///  The name to prompt the user with in the ViewerManager::appChooser
    ///  menu.

  std::string const& application_name() const { return app_name; }
    //
    ///  Return the name of the application to use when describing this
    ///  viewer in the ViewerManager::appChooser or in the
    ///  ViewerManager::ViewerSelector.

  std::string description() const { return "Viewing keys"; }

  virtual std::auto_ptr< std::list<std::string> > help_text();

};

} // namespace cxxtls

#endif
