#ifndef SCRIPT_TABLE_VIEWER_HEADER_INCLUDED_p
#define SCRIPT_TABLE_VIEWER_HEADER_INCLUDED_p
//
// Copyright 2010, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/tableeditor.h>
#include <map>

namespace cxxtls
{

class ScriptTableViewer
: public TableEditor
    ///
    ///  A editor for tables produced by a user defined shell script program.
    ///  See ../STV/README for an explanation of how to write such scripts.
    ///
    ///  The script that generates the tables must maintain a set of
    ///  state variables which are sent to the ScriptTableViewer object
    ///  along with the table contents when the user executes commands.  
    ///  The script is responsible for executing whatever operating system 
    ///  shell commands are necessary to populate tables shown to the user 
    ///  and to save any needed data to files.
    ///
    ///  The script table viewers are not designed for session or connection
    ///  based applications.  Transaction based applications are about all 
    ///  that it can do (and then only with hard work -- or by following a
    ///  well defined implementation pattern).
    ///
    ///  At startup, the script is invoked with no command line parameters to
    ///  indicate that this is the first time through.  The script must generate
    ///  an initial table and initial state variable values.  The user is then 
    ///  presented with that table which he can view, dismiss, enter cell data,
    ///  and issue commands on table cells.  Commands are entered by pressing
    ///  the enter key (on any cell), or by pressing other keys in cells which
    ///  are not editable.  That is, if a cell is not editable, key strokes are
    ///  interpreted as commands.
    ///
    ///  This kind of viewer is best for relatively low performance editing
    ///  because command interpretation requires a script invocation and
    ///  a delete and recreate of the whole table.  To speed things up a bit,
    ///  popup dialogs are handled by the ScriptTableViewer object based on
    ///  state variables provided by the script.  
    ///
    ///  Example scripts can be found in the STV directory.  See example.sh in
    ///  particular because of its comments.
    ///
{
    std::string        scriptname_;
    std::string        fileError_;
    bool               keyKludge_;  // previous input key was ^X
    std::string        title_;      // displayed in top row of viewport

    std::vector<int>   invocationKeys_;  // keys that trigger script reinvocations 

    std::string        directory_; // current directory where script should run

    struct SessionInfo
    {

	 std::map<std::string,std::string> options_;  ///< the script sends us options and we send them back

	 CursorInfo  startCursorPosition_;

	 std::string commandLine_;  ///< Text sent to the script in this
				    ///  invocation as its command line.
				    ///  The script will be re-invoked with this
				    ///  same command line when user presses q
				    ///  to get out of sub-pages.

	 std::vector< std::pair<size_t, size_t> > returnedCells_; 
	     ///<
	     ///  returnedCells_ is a vector of table cell indices (row,col).
	     ///  
	     ///  If this vector is not empty, it means that the script page
	     ///  we are displaying is a "pop up style" page, not a normal "edit" style
	     ///  page.  Popup style pages have the property that when they exit, 
	     ///  the returnedCells are converted into command line parameters and are
	     ///  send back to the page from which this popup was derived as options
	     ///  whose names are of the following form:
	     ///
             ///     --CELL_VALUE_\<row\>_\<col\>="\<cell contents\>"
	     /// 
	     ///  There are no leading zeros.
	     ///
	     ///  Also, if a session is a popup, there are only two kinds of commands
	     ///  that can be executed:  
	     ///
	     ///   @li Escape
	     ///   @li Enter
	     ///
	     ///  Escape can be done using any of the following key sequences:
	     ///
	     ///   @li Esc
	     ///   @li ^X^K
	     ///   @li ^XK
	     ///
	     ///  The ScriptTableViewer::populateTable() function detects popups when
	     ///  it sees a header variable declaration like this:
	     ///
	     ///@code
	     ///
	     ///  --POPUPRETURN= (<row1>,<col1>) (<row2>,<col2>), ...
	     ///  
	     ///@endcode


	 SessionInfo()
	 :  startCursorPosition_(0,0,false)
	 {
	 }

	 bool isPopup() const { return !returnedCells_.empty(); }
	    ///<  Determine if this page is a popup window.   @return true if it is.


    };

    std::vector<SessionInfo> sessions_;  // Each level of script invocation is a session.
					 // For example, the top level session is the invocation of the
					 // script with parameters.  The table entries returned are selectable.
					 // When a section of made from this table, session two is born.
					 // When the user presses q from session 2, session 1 is restored.
					 // 

    std::vector<std::pair<TableViewer, std::string> > prevScreenLayouts_;  
        // copies of the TableViewer contents, used to quickly
        // restore screen layouts after pages are exited.


    void handleReturnedCells();
	///<  Handle passing cells back from a popup style page to its caller.

public:

    ScriptTableViewer(ViewerManager *vm,
	      std::string const &scriptName,
	      std::string        directory=std::string()
	     );
	///<  Construct a ScriptTableViewer and open its file and read
	///   it into memory.  The lines in the file are 
	///   assumed to be delimited by the specified delimiter character.
	///   This constructor assumes that the first line of the file
	///   contains the names of the columns rather than the
	///   data for a row of text.  The column headers must be delimited
	///   as well.
	///   
	///   @param vm          The viewermanager controlling this viewer.
	///   
        ///   @param scriptName  The name of the script to invoke to populate the table
	///
	///   @param directory   Where the script should run.  The script must be visible from
	///                      that directory!
	///
    
    bool ok() const { return fileError_.empty(); }
	///<  Return true if no file read errors have occurred.
	///   See error() for details of any error.
    
    std::string const &error() const { return fileError_; }
	///<  Returns the most recent file i/o error.
    
    virtual ~ScriptTableViewer();



    virtual std::string const &application_name() const; ///<  Return the application name
    virtual std::string description() const;             ///<  Return a description of the application
    static Viewer* app(ViewerManager*vm, std::string &fullname);
	///< Application "hook" function for the ScriptTableViewer class.
	///  Use it as a parameter to add_app, like this:
	///  @code
	///   vm.add_app(ScriptTableViewer::app, "ScriptTableViewer");
	///  @endcode


    virtual bool handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           );
        ///< Handle keystrokes and other events by superceding the
        ///  the behavior of the TableViewer.

    virtual void help();  ///<  Display help about this application

    void populateTable(std::string const &scriptParms, CursorWindow::viewport *vp=0);
	///<
	///   PopulateTable is invoked at various times to re-invoke the
	///   script associated with the ScriptTableViewer.  The goal is to have
	///   the script produce the table.  
	///
	///   @param[in]      scriptParms   The command line arguments to the script
	///   @param[in,out]  vp            Optional view port for painting popups
	///
	///   If the viewport pointer is null, popup processing will not work
	///   correctly.
	///

    void invokeCommand(CursorWindow::viewport *vp, int cmdKey);

    void restartSession(CursorWindow::viewport *vp);

    virtual void repaintTitle(CursorWindow::viewport *vp);

    void setDir(std::string const &dir) { directory_ = dir; } ///< Set the directory where the
							      ///  script should run.
                                                              ///  @param dir  Directory name.
    bool fieldExecutable(size_t tableRow, size_t tableCol) const
        ///
	///  Determine if a specified row and column is and executable
        ///  column.
        ///
	///  @return true if the specified row and column is executable.
        ///
        ///  @param tableRow -- the row within the table, has nothing
        ///                     to do with screen coordinates.
        ///  @param tableCol -- the column within the table, has nothing
        ///                     to do with screen coordinates.

    {
       if(    tableRow >= rows_.size()
          ||  tableCol >= cols_.size()
         )
            return false; // invalid table location

       return rows_[tableRow].executable_ && cols_[tableCol].executable_;
    }

    bool fieldExecutable(CursorInfo const &cur) const
       ///
       ///  Determine if the specified table cursor cell is
       ///  editable or not.
       ///
       ///  @return true if the table row and column, specified in @b cur
       ///               refers to an editable location.
       ///
       ///  @param cur The table cursor position (not screen, table)
       ///
    {
      return fieldExecutable(cur.row_, cur.col_);
    }

    bool showPopup(CursorWindow::viewport *vp,
		   std::string const &definition, 
		   std::vector< std::pair<std::string,std::string> > &userInputs
		  );
	///<
	///    Convert a string definition of a popup into an actual popup dialog
	///    and prompt the user with it.  If the user aborts, return true, else
	///    return false;
	///
	///    User selections are returned in the userInputs vector.  Each user
	///    input field is returned as a <name,value> pair.
	///    
        ///    @param[in]  vp            The viewport to draw in
        ///    @param[in]  definition    The definition of the popup title and fields.
	///    @param[out] userInputs    The user's selections.
	///
	///    The definition string looks like this:
	///
	///@code
        ///
	///    =%==%--POPUP=title;label1;value1;type1;content1;...
        ///
	///@endcode
	///
	///    Invalid formatting of the definition will be treated as a user abort.
	///
					      
    void saveScreenLayout()
        /// Save the screen layout in a stack
    {
       prevScreenLayouts_.push_back( std::pair<TableViewer,std::string>(*this,title_) );
    }

    void restoreSavedScreenLayout()
        /// Pop the screen layout stack and make its
        /// contents active.
    {
       TableViewer &me = *this;
       
       me     = prevScreenLayouts_.back().first;
       title_ = prevScreenLayouts_.back().second;

       prevScreenLayouts_.pop_back();

    }

    void discardSavedScreenLayout()
       /// pop the screen layout stack and discard the data.
    {
       prevScreenLayouts_.pop_back();
    }

};
} // namespace cxxtls

#endif
