#ifndef SYMBOL_VIEWER_HEADER_INCLUDED_p
#define SYMBOL_VIEWER_HEADER_INCLUDED_p
//
// Copyright 2011, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <cxxtls/treeviewer.h>
#include <cxxtls/cpptagdb.h>

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------

using namespace std;

namespace cxxtls
{

//---------------------------------------------------------------------------------

struct SymbolViewerNodeInformation
:  public TreeViewerNodeInfo
   /// Information that can be associated with a TreeViewerNode which is specific
   /// to symbols as defined in the SymbolViewer.
{
   string filename_;
   int    lineNumber_;


   SymbolViewerNodeInformation(string file, int line)
   :  filename_(file)
   ,  lineNumber_(line)
   {
   }

   ~SymbolViewerNodeInformation()
   {
   }

   bool operator() (ViewerManager*vm, CursorWindow::viewport *vp, int command, FileName const &);
   

};

//---------------------------------------------------------------------------------

class SymbolViewer
: public TreeViewer
    ///  A Viewer for the symbol table stored in a CppTagDB as created by the
    ///  program, cpptagdb.exe.
{
   FileName databaseName_;

   CppTagDB::Info database_;

   bool linkTreeMode_;

public:

   SymbolViewer(ViewerManager *vm, std::string databaseFilename);
       ///< Note that the databaseFilename can be either of the following
       /// forms:
       ///
       /// @li a normal pathname for the host you are running on
       /// @li dir/!.cpptagpp
       /// @li dir/filename.cpptagpp
       /// @li *.lnktree
       ///
       /// These formats are interpreted as follows:
       ///
       /// <ol>
       /// <li> Normal pathnames are just files which will be read
       ///      as if they were C++ source.
       /// <li> dir/!.cpptagpp means that *.h* and *.c* in directory
       ///      dir will be read as if they were C++ source
       /// <li> dir/file.cpptagpp means to read dir/file as if it were
       ///      normal C++ source
       /// <li> *.lnktree files are assumed to lists of be linkages to specific
       ///      line numbers in specific files.  These kinds of files are
       ///      useful for automatic creation by scripts.  The symbol viewer
       ///      lets you use the script to generate a link tree and have it animated
       ///      for you.  *.lnktree files are formatted as follows:
       ///@begincode
       ///
       ///      line1:  directory name
       ///      line2:  kind of links being displayed
       ///      line3: filename:lineNumber,DescriptionOfLink1
       ///      line4: filename:lineNumber,DescriptionOfLink2
       ///      line*: filename:lineNumber,DescriptionOfLinkN
       ///
       ///@endcode
       /// </ol>
       ///

   ~SymbolViewer();
       ///<  If this viewer refers to a temporary file, delete it.

   void addChild(TreeViewerNode &parent, CppTagDB::SymbolInfo const &);
       ///<  Add a child node to the tree

   static Viewer* app(ViewerManager*vm, std::string &fullname);
        ///< Application "hook" function for the SymbolViewer class.
        ///  Use it as a parameter to add_app, like this:
        ///  @code
        ///   vm.add_app(SymbolViewer::app, "CppTagDBViewer");
        ///  @endcode


   void helpHeader(std::list<std::string> &help_text); 
       ///<  Add symbol viewer specific help info to the help
       ///   text list.


   void populateLinkTree();
       ///<  Read a .lnktree file and populate the symbol viewer with
       ///   it's contents.  A .lnkTree file will have the following
       ///   contents:
       ///
       ///@begincode
       ///
       ///   line1 :  The name of a file or a directory
       ///   line2 :  A description of what kind of links are in the tree
       ///   line3+:  Linkage information
       ///
       ///@endcode
       ///
       ///  Linkage information consists of the following fields:
       ///  @li filename -- ending in ':'
       ///  @li lineNumber -- ending ','
       ///  @li text -- the rest of the line
       ///
       ///  The user will be presented with a tree of links organized like this
       ///
       ///@begincode
       ///
       ///    FileOrDirectory
       ///       Description of the links
       ///          Directory1
       ///             file1
       ///               text1
       ///               text2
       ///               ...
       ///             file2
       ///                 text3
       ///                 text4
       ///                 ....
       ///         Directory2
       ///
       ///@endcode

};





//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------

} // namespace cxxtls
#endif
