#ifndef tableviewer_included_h_
#define tableviewer_included_h_
#include <cxxtls/viewer.h>

namespace cxxtls
{

class TableViewer
: public Viewer
    ///
    ///  A TableViewer is a base class that lets you implement some kinds of viewers that show 
    ///  rows and columns and allows the user to move around within the table.  The TableViewer
    ///  class is meant primarily to be derived from.  It handles the basics of table editing
    ///  without doing anything really useful.  You should derive from a TableViewer and implement
    ///  various methods as needed to do something useful.  In particular, the Tableviewer::handle_event
    ///  method returns with an "exit" code whenver a function key that it doesn't understand is pressed.
    ///  This was orignally done to enable users to program function keys behavior in derived classes.
    ///  It turns out to make the implementation of the CSV Viewer somewhat tedious....
    ///
    ///  Note that the @ref TableEditor class allows the user to edit table columns and rows.
    ///
    ///  Instructions for creating a useful table viewer sub-class:
    ///
    ///    @li Derive from TableViewer.
    ///    @li In your constructor, populate the TableViewer using addColumnHeaders(), addRow(),
    ///        and addColumnData() for the cells of interest.
    ///
    ///    @li Superseed the handle_event virtual method to deal with user events:
    ///
    ///        <blockquote>
    ///            @li handle cell editing
    ///            @li handle command interpretation
    ///        </blockquote>
    ///  
    ///
    ///
{
public:

  static void tester(long);  // do nothing function tha that you can pass integers to so the debugger won't optimize
                             // variables out.

  TableViewer(ViewerManager*vm)
  :  Viewer(vm)
  ,  originRow_(0)         // ul corner row offset (zero based)
  ,  originCol_(0)         // ul corner col offset (zero based)
  ,  displayTitles_(false) // always display row and column titles.  See setDisplayTitles()
  ,  active_(false)        // not currently the active window
  ,  layoutDirty_(true)    // need to recompute layout
  ,  cursorRow_(-1)        // current cell row within layout_
  ,  cursorCol_(-1)        // current cell column within layout_
  ,  cursorDisplayed_(1)   // true if the current cursorRow_ and cursorCol_ are shown highlighted
  ,  displayOrigin_(0,0)   // for drawing during and after a paint (vp coords)
  ,  viewerManager_(vm)    // saved for the clear function
  ,  dataDirty_(0)         // field edits have not been saved
  ,  pageWidth_(0)         // world columns set painting process     (layout_ columns)
  ,  pageHeight_(0)        // world rows set by the painting process (layout_ rows)
  ,  mixed_case_(true)     // search parameter
  ,  is_regex_(false)      // search parameter
  ,  match_words_(false)   // search parameter
  ,  search_text_("")      // text to search for
  ,  topMarkedRow_(size_t(-1))     // visibly top, but lowest row number
  ,  bottomMarkedRow_(size_t(-1))  // >= topmarkedRow_
  {
  }

  // WARNING:  Copies of this object are made, so don't leak memory!  Right now, now
  //           raw pointers that we own are kept in this class, if you add one make sure you 
  ///          implement a copy constructor!

  void triggerLayoutRecompute() { layoutDirty_ = true; }
      ///<
      ///  Called by field editing functions to inform the table
      ///  repaint mechanism, operator()(), that the column sizes
      ///  need to be recomputed.
  
  virtual ~TableViewer();

  virtual void repaintTitle(CursorWindow::viewport *vp);
      ///<
      /// Repaint the title (presumably after it has changed).
      ///
      /// @param vp The viewport where this table is drawn.
      ///
      /// By default, this function prints the application name,
      /// but derived classes may wish to superseed this behavior.

  virtual void operator() ( CursorWindow::viewport *     vp,
                            int                          cmd
                          );
     ///< Handle screen repaint commands

  virtual bool handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           );
     ///< Handle keystrokes and other events.

  virtual std::string const &application_name() const; ///<  Return the application name
  virtual std::string description() const;             ///<  Return a description of the application
  virtual void help();                                 ///<  Display help about this application
  virtual bool is_dirty() const;                       ///<  Return true if the application has modifications

  struct ColumnInfo
  {
     std::string title_;
     size_t      width_;       // size hint, not actual value.  See TableViewer::colWidths_
     bool        leftBorder_;
     bool        bold_;
     bool        editable_;    // not used in table viewer, but derived classes can use it
     bool        executable_;  // ditto

     ColumnInfo()
     : title_("")
     , width_(0)
     , leftBorder_(false)
     , bold_(false)
     , editable_(true)
     , executable_(true)
     {
     }

     ColumnInfo(std::string const &title, size_t width, bool leftBorder)
     : title_(title)
     , width_(width)
     , leftBorder_(leftBorder)
     , bold_(false)
     , editable_(true)
     , executable_(true)
     {
     }

     ColumnInfo(std::string const &title, size_t width, bool leftBorder, bool bold, bool edit, bool executable)
     : title_(title)
     , width_(width)
     , leftBorder_(leftBorder)
     , bold_(bold)
     , editable_(edit)
     , executable_(executable)
     {
     }

  };

  struct RowInfo
  {
     std::string title_;
     size_t      height_;    // size hint, not actual value
     bool        bottomBorder_;
     bool        bold_;
     bool        editable_;   // not used in table viewer, but derived classes can use it
     bool        executable_;

     RowInfo()
     : title_("")
     , height_(0)
     , bottomBorder_(false)
     , bold_(false)
     , editable_(true)
     , executable_(true)
     {
     }

     RowInfo(std::string const &title, size_t height, bool bottomBorder)
     : title_(title)
     , height_(height)
     , bottomBorder_(bottomBorder)
     , bold_(false)
     , editable_(true)
     , executable_(true)
     {
     }

     RowInfo(std::string const &title, size_t height, bool bottomBorder, bool bold, bool edit, bool executable)
     : title_(title)
     , height_(height)
     , bottomBorder_(bottomBorder)
     , bold_(bold)
     , editable_(edit)
     , executable_(executable)
     {
     }

     void setRowTitleNumber(size_t row); // Set title_ to be a numeric string

  };

  struct CellInfo
  {

     size_t worldRow_;       // imaginary coordinate system (bigger than screen)
     size_t worldColumn_;    // diito
     std::vector< std::string > text_;  // one vector element for each row (usually 1)

     size_t widestRow() const;  // returns max of string lengths in text_

     CellInfo()
     :  worldRow_(0)
     ,  worldColumn_(0)
     {
     }

     CellInfo(size_t row, size_t col, std::vector< std::string > const &text)
     :  worldRow_(row)
     ,  worldColumn_(col)
     , text_(text)
     {
     }

  };

    struct UndoElement
       // objects that get pushed onto the undo stack
    {
       enum type { 
                   restore_rowcol,  // restore a single column in a single row
                   restore_row,     // including all columns
                   restore_col,     // add the column back including all row values
                   delete_row,      // and all column data in one or more rows
                   delete_col,      // delete one or more columns from all rows and the title line
                 };

      type type_;  // what kind of undo operation is this

      std::list<UndoElement*>  recursive_elements_;  // vector of undo elements in a block delete

      size_t row_;
      size_t col_;
      std::vector< std::string > cell_value_;
      std::vector< std::vector<std::string> >  col_values_; // for restore_row, restore_col, and restore_rowcol
                                                            // wherein there will be only the one entry populated.
      std::string col_title_;

#if 0
      RowInfo    rowInfo_;
#endif
      ColumnInfo colInfo_;

      bool       renumberColumns_; // generally, no, but for the csv viewer yes.

      size_t count_; // rows or columns to delete

      typedef std::vector< std::vector< std::string > >  BlockRows;

      BlockRows block_;  // for restoring marked blocks and single deleted rows.

      UndoElement(size_t row, size_t count=1)
          /// Construct an undo element whose job it is to delete a row or rows
      :  type_(delete_row)
      ,  row_(row)
      ,  col_(0)
      ,  renumberColumns_(false)
      ,  count_(count)
      {
      }

      UndoElement(type t, size_t col=0, size_t count=1)
      :  type_(t)
         ///  Create a mostly uninitialized UndoElement but with the specified type.
         ///  typically, you only use this for creating Column delete_col elements --
         ///  which are used to un-create a column.
      ,  row_(0)
      ,  col_(col)
      ,  renumberColumns_(false)
      ,  count_(count)
      {
      }


      UndoElement(size_t row, size_t col, std::vector< std::string > const &oldValue)
          ///  Create an undo element that will restore the old state of the string value
          ///  of  row and column
      : type_(restore_rowcol)
      , row_(row)
      , col_(col)
      , cell_value_(oldValue)
      ,  renumberColumns_(false)
      ,  count_(0)
      {
      }


      UndoElement(size_t row, BlockRows &block)
          ///  Create an undo element that will restore the old state of the string value
          ///  of an entire block of rows.  The input block gets emptied by this process
          ///
          ///  @param[in] row  the start row of the block whose deletion will be undone by
          ///                  by this element.
          ///  @param[in,out] block  the data that will be restored by this operation.
      : type_(restore_row)
      , row_(row)
      , renumberColumns_(false)
      , count_(0)
      {
         block_.swap(block);
      }

      UndoElement(size_t col, std::vector< std::vector< std::string > > const &oldValues, ColumnInfo const &colInfo)
          ///  Create an undo element that will restore the old state of the string value
          ///  of an entire column -- including its title
      : type_(restore_col)
      , col_(col)
      , col_values_(oldValues)
      , colInfo_(colInfo)
      ,  renumberColumns_(false)
      ,  count_(0)
      {
      }


      ~UndoElement(); ///< clean up memory, etc

    }; // struct UndoElement


  struct FieldAttributes
  {
     bool bold_;
     bool underlined_;
     bool editable_;
     bool executable_;

     FieldAttributes(bool b=false, bool u=false, bool e=false, bool x=false)
     : bold_(b)
     , underlined_(u)
     , editable_(e)
     , executable_(e)
     {
     }

     template<class Iterator>
     static FieldAttributes
     skip(Iterator &first, Iterator const &last)
         /// Parse the [x!_] string leading a row field header 
         /// or a column field header.
         ///
         /// @return a FieldAttributes object and move the first iterator past
         ///         the field attribute string.
         ///
         /// @param first defines the beginning of the text range of interest.
         /// @param last  defines the end of the text range.

     {
        FieldAttributes fa;
     
        if(first == last || *first != '[')
        {
          return fa;
        }

        ++first;

        while(first != last && *first != ']' )
        {
           char c = *first++;

           switch(c)
           {
              case '!':  fa.bold_         = true; break;
              case 'x':  fa.executable_   = true; break;
              case 'e':  fa.editable_     = true; break;
              case '_':  fa.underlined_   = true; break;
           }
        }

        if(first != last)
          ++first; // skip trailing ]

        return fa;


     }


  };


  size_t rows() const { return rows_.size(); } ///< return count of table rows
  size_t cols() const { return cols_.size(); } ///< return count of table columns

  size_t originRow() const { return originRow_; } ///< top left corner row
  size_t originCol() const { return originCol_; } ///< top left corner column

  void setDisplayTitles(bool flag) { displayTitles_ = flag; }
      ///<
      ///    Set the flag that controls the permanancy of the displaying of the row and column
      ///    titles.
      ///
      ///    @param flag if true, means that the titles should always be displayed.
      ///

  size_t longestRowTitle() const; ///< returns the number of chars in longest title

  void setRowsAndColumns(size_t rows, size_t cols);
      ///< 
      ///  Set the maximum number of rows and columns in the layout.  Adding or deleting
      ///  as necessary to make it happen.
      ///
      ///  @param rows is the exact number of rows to establish
      ///  @param cols is the exact number of cols to establish.

  size_t columnWidth(size_t column) const; 
      ///<
      ///  Consult the layout_ and the cols_ tables to determine with maximum
      ///  of the column header and the widest cell in that column.  Does not
      ///  include gutters.

  void setRowInfo(size_t row, std::string const &title, size_t height, bool border)
      /// Set the row header information
      /// @param row    defines the row number of interest (world row, not screen row)
      /// @param title  defines the title for the row
      /// @param height defines the number of screen rows to reserve vertically for "row"
      /// @param border defines the bottom border drawing option
  {
     if(row < rows_.size())
     {
        layoutDirty_ = true;

        rows_[row].title_        = title;
        rows_[row].height_       = height;
        rows_[row].bottomBorder_ = border;
     }
  }

  void setRowInfo(size_t row, std::string const &title, size_t height, bool border, bool bold, bool edit)
      /// Set the row header information
      /// @param row    defines the row number of interest (world row, not screen row)
      /// @param title  defines the title for the row
      /// @param height defines the number of screen rows to reserve vertically for "row"
      /// @param border defines the bottom border drawing option
      /// @param bold   defines the bold drawing flag for the row
      /// @param edit   defines the editable flag for this row
  {
     if(row < rows_.size())
     {
        layoutDirty_ = true;

        rows_[row].title_        = title;
        rows_[row].height_       = height;
        rows_[row].bottomBorder_ = border;
        rows_[row].bold_         = bold;
        rows_[row].editable_     = edit;
     }
  }


  void setColInfo(size_t col, std::string const &title, size_t width, bool border)
      /// Set the column header information
      /// @param col    defines the col number of interest (world col, not screen col)
      /// @param title  defines the title for the row
      /// @param width  defines the number of screen columns to for this world column
      /// @param border defines the left border drawing option
  {
     if(col < cols_.size())
     {
        layoutDirty_ = true;

        cols_[col].title_ = title;
        cols_[col].width_ = width;
        cols_[col].leftBorder_ = border;
     }
  }

  void setColInfo(size_t col, std::string const &title, size_t width, bool border, bool bold, bool edit)
      /// Set the column header information
      /// @param col    defines the col number of interest (world col, not screen col)
      /// @param title  defines the title for the row
      /// @param width  defines the number of screen columns to for this world column
      /// @param border defines the left border drawing option
      /// @param bold   defines the bold display attribute for this column
      /// @param edit   defines the editable flag for this colum
  {
     if(col < cols_.size())
     {
        layoutDirty_ = true;

        cols_[col].title_ = title;
        cols_[col].width_ = width;
        cols_[col].leftBorder_ = border;
        cols_[col].bold_       = bold;
        cols_[col].editable_   = edit;
     }
  }

  void addColumnHeader(std::string const &title, size_t width, bool border)
      /// Add a column to the column header table.
      /// @param title  defines the title for the row
      /// @param width  defines the number of screen columns to for this world column
      /// @param border defines the left border drawing option
  {
      layoutDirty_ = true;

      cols_.push_back( ColumnInfo(title, width, border) );
  }

  void addColumnHeader(std::string const      &title, 
                       size_t                  width, 
                       FieldAttributes const  &a, 
                       bool                    border)
      /// Add a column to the column header table.
      /// @param title  defines the title for the row
      /// @param width  defines the number of screen columns to for this world column
      /// @param a      defines the field attributes
      /// @param border defines the left border drawing option
  {
      layoutDirty_ = true;

      cols_.push_back( ColumnInfo(title, width, border, a.bold_, a.editable_, a.executable_) );
  }


  void addColumnData(std::string const &text)
  {
      size_t rows = layout_.size();

      if(rows == 0)
      {
        addRow("",1,false);

        rows = 1;
      }

      std::vector<CellInfo> &row = layout_[rows-1];

      row.push_back(CellInfo());

      CellInfo &cell = row.back();

      cell.text_.push_back(text);  // 1 row in this cell
  }

  void addRow(std::string const &title, size_t height, bool bottomBorder, bool bold=false, bool editable=true, bool executable=true)
  {
     rows_.push_back( RowInfo(title, height, bottomBorder, bold, editable, executable ) );

     layout_.push_back(std::vector<CellInfo>());
  }

  RowInfo const &rowInfo(size_t row) const { return rows_[row]; }
     ///< get a copy of a specified row titles, etc.
     ///  @param row   The row to modify
     ///
     ///  @return The specified row's information.

  void setRowInfo(size_t row, RowInfo const &r) { rows_[row] = r; }
     ///< Set the titles and other row specific information givin a row.
     ///  @param row   The row to modify
     ///  @param r  The RowInfo to change


  void insertColumn(size_t beforeColumn, std::string const &title, size_t width, std::string const &defaultValue=std::string(), bool border=true);
      /// insert a new column before the specified beforeColumn.  The column
      /// will be given the specified title.  All rows will be modified to include this new
      /// column and the column will be populated with the specified default value.

  void deleteColumn(size_t column);
     /// delete the specified column from all rows as well as the title line

  CellInfo const *cellInfo(size_t row, size_t col) const;
      ///< return a null or a const pointer to the specified cell information.
      ///
      ///  @param row defines the zero based row number of interest
      ///  @param col defines the zero based column number of interest
      ///  @return a const pointer to the desired cell or null if the cell
      ///          does not exist.

  CellInfo *cellInfo(size_t row, size_t col);
      ///< return a null or a modifiable pointer to the specified cell information.
      ///
      ///  @param row defines the zero based row number of interest
      ///  @param col defines the zero based column number of interest
      ///  @return a modifiable pointer to the desired cell or null if the cell
      ///          does not exist.

   void setOrigin(size_t row, size_t col) { originRow_ = row; originCol_ = col; }
      ///<  Set the upper left corner column in the world world of the layout_
      ///   no the screen coordinate world.
      ///
      ///   @param row is the row in the virtual world
      ///   @param col is the column in hte virtual world

   virtual void setCursorInfo(CursorWindow::viewport *vp,
                              size_t worldRow, 
                              size_t worldCol, 
                              bool displayed=true);
      ///<
      ///  Set the "edit cell" for the table in world coordinates.  If the cursor is currently visible
      ///  and not already at the specified location it will be moved.
      ///
      ///  If the current edit cell has been modified, its modifications will be kept but no action
      ///  will be triggered as a result of the move to another cell.
      ///
      ///  @param vp       defines the viewport to draw in
      ///  @param worldRow defines the row in the table to be highlighted
      ///  @param worldCol defines the column in the table to be highlighted
      ///  @param displayed indicates that the screen should be updated to show
      ///         the highlighting.
      ///
      ///  Note:  the displayed parameter is optional
      ///
      ///  This function is virtual so that derived classes can detect cursor movement
      ///  and handle input field editing resetting, flushing data, etc.


   virtual void makeVisible(CursorWindow::viewport *vp,
                            size_t worldRow, 
                            size_t worldCol, 
                            bool displayed=true);
      ///<  Same as SetCursorInfo but this function adds the behavior that the
      ///   particular row and column are brought to visible region of the table.
      ///   That is, the origin row and column are adjusted as well as the cursor
      ///   row and column.

   struct CursorInfo
      ///  Information about the table cursor, as opposed to the screen cursor.
      ///  The row_ and col_ are refer to table entries, not screen locations.
      ///  The displayed_ member refers to the cursor being actually displayed 
      ///  at this time.
   {
       size_t row_;
       size_t col_;
       bool   displayed_;

       CursorInfo(size_t r, size_t c, bool d)
       : row_(r)
       , col_(c)
       , displayed_(d)
       {
       }

   };

   CursorInfo cursorInfo() const { return CursorInfo(cursorRow_, cursorCol_, cursorDisplayed_); }
       ///<
       ///   Return the current cursor information

   void setCursorInfo(CursorWindow::viewport *vp, 
                      CursorInfo const &i
                     ) 
       ///
       /// Set the cursor information given a CursorInfo structure.
       ///
   { 
       setCursorInfo(vp, i.row_, i.col_, i.displayed_); 
   }

   void moveUp(CursorWindow::viewport *vp);      
                       ///< Move the cursor up one row. 
                       ///  Normally bound to the up arrow key.
                       
   void moveDown(CursorWindow::viewport *vp);    
                       ///< Move the cursor down one row. 
                       ///  Normally bound to the down arrow key.
                       
   void moveLeft(CursorWindow::viewport *vp);    
                       ///< Move the cursor left one column.
                       ///  Normally bound to the left arrow key.
                       
   void moveRight(CursorWindow::viewport *vp);   
                       ///< Move the cursor right one column.
                       ///  Normally bound to the right arrow key.
                       
   void pageUp(CursorWindow::viewport *vp);      
                       ///< Move the cursor up one page.
                       ///  Keep the column the same.
                       ///
                       ///  Normally bound to the PageUp or Prior key.
                       
   void pageDown(CursorWindow::viewport *vp);    
                       ///< Move the cursor down one page.
                       ///  Keep the column the same.
                       ///
                       ///  Normally bound to the PageDown or Next key.
                       
   void pageBottom(CursorWindow::viewport *vp);  
                       ///< Move the cursor to the last valid rows, and
                       ///  keep the column the same.
                       ///
                       ///  Normally bound to the control B key.

   void pageTop(CursorWindow::viewport *vp);     
                       ///< Move the cursor to the first row in the table.
                       ///  keep the column the same.
                       ///
                       ///  Normally bound to the control B key.
                       
   void pageRight(CursorWindow::viewport *vp);   
                       ///< Move the cursor right one page.
                       ///  Keep the column the same.
                       ///
                       ///  Normally bound to the tab key.
                       
   void pageLeft(CursorWindow::viewport *vp);    
                       ///< Move the cursor left one page.
                       ///  Keep the column the same.
                       ///
                       ///  Normally bound to the shift-Tab key.
                       
   void pageHome(CursorWindow::viewport *vp);    
                       ///< Move the cursor to 0,0
                       ///
                       ///  Normally bound to the Home key
                       
   void pageEnd(CursorWindow::viewport *vp);     
                       ///< Move the cursor to maxrow,maxcol
                       ///  Keep the column the same.
                       ///
                       ///  Normally bound to the end key
                       
   void clear();       ///< Discard all rows and columns.

  void hideCursor(CursorWindow::viewport *vp);       
                       ///< If cursor is visible, repaint it to remove its highlighting.
                       ///  Also, terminate any edit state operations (if any).

  void showCursor(CursorWindow::viewport *vp);       
                           ///< If the cursor is not already displayed, repaint it to indicate
                           ///  highlighting for edit.

  virtual void paintCursor(CursorWindow::viewport *); 
      ///< If the cursor is marked for display, draw it on the screen.
      ///
      ///  The function is virtual so that derived classes can hook into the repaint
      ///  functionality so they can position the edit cursor.
      ///
      ///  @param vp is the viewport to positiont he cursor into
      ///

  void paintCell(CursorWindow::viewport *vp,
                 size_t vpRow,
                 size_t vpCol,
                 size_t tbRow,
                 size_t tbCol,
                 bool   highlight=false,
                 bool   editCursor=true

                );               
      ///< Paint the specified cell at the specified viewport location.
      ///  No validity checking is done.  The most recently painted viewport
      ///  is used.
      ///
      ///  @param vp         the viewport to draw in
      ///  @param vpRow      the viewport row
      ///  @param vpCol      the column
      ///  @param tbRow      table row
      ///  @param tbCol      the table column
      ///  @param highlight  flag indicating highlighted display
      ///  @param editCursor flag indicating the cursor should be left in edit poisiton

  void set_data_dirty(bool f) { dataDirty_ = f; }
      ///<  Set the 'unsaved modifications' flag.
      ///
      ///     @param f defines the status of the dirty flag.
      ///
      ///   See the function is_dirty().

  bool active() const { return active_; }
      ///<  Return the current "is window active" flag.
      ///
      ///  A given window is active if it has text focus, otherwise
      ///  it is not.

  bool mixed_case() const { return mixed_case_; }
      ///<  Return the state of the mixed-case-find flag.

  typedef std::vector< std::vector< CellInfo > > LayoutType;

  bool markedBlockIsEmpty() const 
     /// returns true if the marked block is empty of lines.
  {
     return topMarkedRow_ >= layout_.size();
  }

  void includeRowInBlock(CursorWindow::viewport *vp);  ///<  toggles the state of the current line.

  bool isMarked(size_t row)
      /// return true if the row is marked
  {
    return     row >= topMarkedRow_ 
           &&  row <= bottomMarkedRow_
           ;
  }

  void clearMarkedBlock()
      ///  discard the marked block.
  {
      topMarkedRow_ = bottomMarkedRow_ = size_t(-1);
  }


private:



  size_t                    originRow_;
  size_t                    originCol_;

  bool                      displayTitles_;
  bool                      active_;  // currently active window?
  bool                      layoutDirty_; // need to recompute layout

  size_t                    cursorRow_;
  size_t                    cursorCol_;
  bool                      cursorDisplayed_;


  CursorWindow::row_col     displayOrigin_; // scr coords of origin table display 

  ViewerManager*            viewerManager_;  // we don't own this!

  CursorWindow::row_col     textInputPointer_;

  bool                      dataDirty_; // see is_dirty(), set_data_dirty()


  void recomputeLayout();  ///< Regenerate the world coords in layout_.

protected:

  size_t                    pageWidth_;   // set by the painting proc
  size_t                    pageHeight_;

  std::vector<size_t>       colWidths_;     // computed widths


  LayoutType                layout_;  // a cell for each row and column

  std::vector<RowInfo>      rows_;  // titles, size hints etc
  std::vector<ColumnInfo>   cols_;

  ///////////////////////////////////////////////////////////////////////////////////////
  // search dialog parameters
  //
  bool                      mixed_case_;     ///< If true, use case insensitive searches
  bool                      is_regex_;       ///< If true, use regular expression searches
  bool                      match_words_;    ///< If true, use word searches
  std::string               search_text_;    ///< Text to search for
  ///////////////////////////////////////////////////////////////////////////////////////

  ///////////////////////////////////////////////////////////////////////////////////////
  // marked block info
  //
  size_t topMarkedRow_;     // lowest row number [0, layout_.size()-1]
  size_t bottomMarkedRow_;  // highest row numbers.  inclusive.  bottomMarkedRow_ is > top.
                            // bottomMarkedRow_ is INSIDE the marked region, not beyond it.
                            // when topMarkedRow_ >= layout_.size(), the marked block is empty
  ///////////////////////////////////////////////////////////////////////////////////////


  virtual void paintField(CursorWindow::viewport *vp,
                          size_t          vpRow, 
                          size_t          vpCol, 
                          CellInfo const *cell, 
                          size_t          cellRow, 
                          size_t          vpColumns,
                          bool            editCursor
                         );
      ///<  Paint the text of a cell's row and any trailing blanks needed
      ///   to pad the string's width to vpColumns.
      ///
      ///   @param vpRow      The upper left row where display is to start
      ///   @param vpCol      The upper left col where display is to start
      ///   @param cell       A pointer to the layout_ location being painted.
      ///   @param cellRow    The row within the cell to paint.
      ///   @param vpColumns  The width of the field.
      ///   @param editCursor Indicates the cursor should be left in feild
      ///                     edit position (if true).
      ///
      ///   This function exists to be superseeded by derived classes.
      ///
      ///   Note that the leading gutter will have been painted and the
      ///   text attributes set before this function is called.
      ///
      ///   If editCursor is false, then the cursor will be left at the end
      ///   end of the field.  This flag is ignored if the field is not 
      ///   editable and the cursor will end up at the end.

   void setTextInputPointer(size_t vpRow, size_t vpCol)
      ///  Set the screen location where text cursor goes. 
      ///  @param vpRow is the viewport relative row
      ///  @param vpCol is the viewport relative column
   {
       textInputPointer_.row_ = vpRow;
       textInputPointer_.col_ = vpCol;
   }

   void positionTextInputPointer(CursorWindow::viewport *vp);

   void forceRelayout() { layoutDirty_ = true; } ///< mark the table layout for
                                                 ///  recomputation at some later
                                                 ///  point in time

    void insertRowAbove(CursorWindow::viewport *vp, RowInfo const &info); 
        ///< Insert new row above the current row
        ///
        ///  @param vp   the viewport to draw in
        ///  @param info the title of the current row
        ///

    void insertRowBelow(CursorWindow::viewport *vp, RowInfo const &info); 
        ///< Insert new row below the current row
        ///
        ///  @param vp   the viewport to draw in
        ///  @param info the title of the current row
        ///

    void deleteRow(CursorWindow::viewport *vp);
        ///< Erase the current row.
        ///
        ///  @param vp  the viewport to draw in -- though none is none

    void checkLayoutRecompute()
    {
        CursorInfo ci = cursorInfo();

        CellInfo *cell= cellInfo(ci.row_, ci.col_);

        if(!cell || cell->text_.empty())
          return;

        std::string text = cell->text_[0];

        if(    (text.size() < colWidths_[ci.col_] / 2)
#if 0
           ||  (text.size() > colWidths_[ci.col_] * 1.5)
#else
           ||  (text.size() > colWidths_[ci.col_] )   // resize if string bigge than field
#endif
          )
        {
          triggerLayoutRecompute(); 
        }
    }

    void findFirst(CursorWindow::viewport *vp);
        ///< Prompt the user for find string parameter
        ///  and search for that string from the current edit point.

    virtual bool findNext(CursorWindow::viewport *vp, bool first);
        ///<  Search for the current search_text_ from the current cell to the end
        ///   of the table.  Update the screen to reflect the new cursor position.
        ///
        ///   @param vp     The viewport in which to draw
        ///   @param first  True if the is first time this string has been used for
        ///                 the search (and thus the current cell should be searched).
        ///
        ///   @return true to indicate that the search text was found and that the
        ///                the screen has been updated.

    std::list<UndoElement*> undo_stack_;

    void pushUndoCellChange(size_t row, size_t col);
        ///<  push an undo stack element that is a change to a single cell.

    UndoElement* pushUndoRowDelete(size_t row, size_t count=1);  
        ///< before deleting the row, call this to get it restored
        ///
        /// @return null or a pointer to the undo element.  Do not delete this element yourself!

    void pushUndoRowCreate(size_t row, size_t count);  ///< AFTER creating the row, call this with the new row number
                                                       /// to get it deleted when you press the undo key.

    void undo(void *vp);  ///< pop the undo stack and execute the change it contains
                          ///  vp is the CursorWindow::viewport pointer.

    void pushUndoColCreate(size_t col, bool renumberColumns=false);  
                                         ///< AFTER creating the column, call this with the new column
                                         ///  number to allow for the undoing of the create.

    void pushUndoColDelete(size_t col, bool renumberColumns=false);  
                                         ///<  before deleting the column, call this with the column
                                         ///   number to allow for the recreating the column with
                                         ///   the undo key.


}; // class TableViewer;
} // namespace cxxtls
#endif
