#ifndef viewer_h_included
#define viewer_h_included

//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
///
///  This file defines the class, Viewer -- which is the one of the highest
///  level text windowing application interfaces.
/// 
///  Before directly using any of the classes in this file, you should consider
///  reading @ref text_windows.

#include <cxxtls/cursorwindow.h>
#include <memory>
#include <vector>
#include <list>

namespace cxxtls
{

class ViewerManager;

class Viewer
  ///
  ///  A Viewer is an application 
  ///  fragment that knows how to draw in 
  ///  CursorWindow::viewport which is managed by class ViewerManager using drawing
  ///  calls in the viewport class.
  /// 
  ///  Before directly using any of the classes in this file, you should consider
  ///  reading @ref text_windows or @ref CXXCURSORWINDOWS
  /// 
  /// To implement a Viewer you must do the following:
  ///
  ///  @li derive a class from Viewer that handles drawing and command
  ///      interpretation for a specific kind of object which can be handled
  ///      with a text viewer.  Note that your derived class may want to support
  ///      the concept of temporary files.  That is, a file being edited might
  ///      be temporary and should be deleted as soon as the editor is destructed.
  ///      In your derived class destructor, call the base member isTemporary()
  ///      to find out -- and delete the file being edited if it is true.
  ///
  ///  @li create an instance of your class -- possibly initializing
  ///      it with important data
  ///
  ///  @li create a CursorWindow::vieweport and pass the address of your
  ///      viewer class to it.  This is done automatically if you are using
  ///      the ViewerManager class -- and you should be using it.
  ///      See editor/editor.c
  ///
  ///  @li Some part of your application must invoke the ViewerManager::run()
  ///      function to get input from the user and feed it to your activated
  ///      viewer.
  ///
  ///  Note that there are some derived Viewers which may be more easy to
  ///  use than Viewer.  Below, you will find ListViewer which is designed
  ///  to present and allow editing of a list of strings.
  ///
  ///  Maintenance note:  a viewer should not store a pointer to its viewport
  ///  because the viewport associated with a viewer will change over time.
  ///  A viewer has the responsibility of completely repainting the viewport it
  ///  given when asked to do so.  This means, you can't store the size of the
  ///  viewport either -- but must inquire it from the viewport passed during
  ///  resizes and handle_event() calls.
  ///
  ///  A convention you should follow:  when drawing the bottom line of your
  ///  viewport, always make sure that it is highlighed in some fashion --
  ///  such as making it underlined.  When drawing the left most column of
  ///  your viewerport, if the origin.col_ for your viewport is not 0, draw a
  ///  vertical bar.  That is, do something like this when painting:
  ///
  ///  @code
  ///
  ///    row_col origin = viewport->origin();
  ///
  ///    int left_column_bias = 0;
  ///
  ///    if(origin.col_ != 0)
  ///    {
  ///      left_column_bias = 1;
  ///
  ///      for each row of your viewport, draw '|' in column 0.
  ///
  ///      if this is the last row, make sure you have it underlined
  ///      or inverted from the other rows in some fashion.
  ///    }
  ///
  ///    In all subsequent drawing, don't draw in column 0.
  ///
  ///  @endcode
  ///

: public CursorWindow::viewport::repaint_handler
{

protected:

  ViewerManager* manager_;

public:

  Viewer(ViewerManager* man)
  : manager_(man)
  {
  }

  virtual ~Viewer();

  ViewerManager* manager() { return manager_; }

  virtual void operator() ( CursorWindow::viewport *     vp,
                            int                          cmd
                          ) = 0;
     ///< Handle screen repaint commands
     ///
     ///  @note The vp parameter gets deleted regularly, don't save
     ///        it!

  enum BackgroundProcessStatus
  {
      NO_BACKGROUND_PROCESS,
      REQUESTING_VIEWER_TERMINATION,
      NO_BACKGROUND_DATA_AVAILABLE,
      BACKGROUND_DATA_AVAILABLE
  };

  virtual BackgroundProcessStatus getBackgroundData(CursorWindow::input_event &e);
     ///<  Get any pending background process data and store it in the 
     ///   specified event.  Return one of the enumeration values in
     ///   BackgroundProcessStatus to indicate what has happened.
     ///
     ///   In the event that only a single character is returned, 
     ///   e.type_ will be set to CursorWindow::input_event::ProcessOutput.
     ///   
     ///   If there are multiple characters waiting, e.type_ will be set to
     ///   CursorWindow::input_event::BackgroundData
     ///
     ///   @param e  An initialized but not populated cursor window event.
     ///   @return   One of the BackgroundProcessStatus values.
     ///
     ///   The default implementation of this function always returns
     ///   NO_BACKGROUND_PROCESS.  If the viewer does implement one, it must
     ///   return some other value.  Fair warning:  only the main thread of the
     ///   application can make CursorWindow calls!  No other thread should.


  virtual bool handle_event( CursorWindow::input_event const * e,
                             CursorWindow::viewport          * vp
                           ) = 0;
     ///<
     /// Handle input events -- and return true if the event is meant to
     /// terminate the viewer -- and make it go away.  When handling
     /// key events -- either FunctionKey or DataKey, you will want to
     /// consult the CursorWindow::func mapping to figure out if the
     /// key corresponds to 'system wide' edit function.  For example,
     ///
     /// @code
     ///    int edit_func = vp->window()->func[e->value_];
     ///    switch(edit_func)
     ///    {
     ///       case CursorWindow::func_up:    // the up arrow key
     /// @endcode
     ///
     ///
     /// Good example:  see handle_event in the deltaviewer.cxx file.
     ///
     /// Choose this approach instead of looking at the key values
     /// directly to allow for consistence in system wide input remapping
     /// that occurs when the application modifies the CursorWindow::func
     /// map.  Note that the CursorWindow::func_data is returned for all
     /// keys not mapped into one of the other named func_* keys.
     ///
     /// Usage note:  If your handle event creates a new viewport that
     /// covers this viewport, make sure that your current viewport doesn't
     /// draw anything until the next repaint request.  Otherwise, your
     /// screen will get very jumbled up!

  virtual std::string const &application_name() const = 0;
     ///<
     ///  Return a const references to a string which defines the name
     ///  of the application being implemented in the viewer.  This function
     ///  must always return the same address and the string at that address
     ///  must never change;

  virtual std::string description() const = 0;
     ///<
     ///  Get a description of this viewer.  Typically, this just returns
     ///  the name of the directory or the name of the file the viewer is
     ///  being applied to.  However, with FTP viewers, POP3 viewers, etc
     ///  this may not refer to any real file.

  virtual void help();  ///< Present the user with help information that is
                        ///  specific to this viewer's operation and contains
                        ///  information about system keep bindings, etc.
                        ///  Note that there is a default implementation of
                        ///  this function, but it should be overridden.
                        
  virtual bool is_dirty() const;
    ///<  Returns true if there are unsaved edit changes in this viewer.
    ///  the default return value is false.  Override this function if
    ///  your viewer has the ability to have unsaved edit changes -- most
    ///  don't.
                        
  typedef CursorWindow::viewport        viewport;          ///< typename alias
  typedef CursorWindow::row_col         row_col;           ///< typename alias
  typedef CursorWindow::input_event     input_event;       ///< typename alias

  typedef std::list<std::string> paste_buffer_type; ///< The data type of the paste buffer

  paste_buffer_type* paste_buffer();  ///< Get a pointer to the viewer manager's paste
                                      ///  buffer.

  void commitPasteBuffer() const;  ///<  write the paste buffer to the os clipboard.
  void fetchPasteBuffer();         ///<  read the paste buffer from the os clipboard

  virtual void set_row_col_hint(size_t line, size_t column, bool repaint=false) {}
    ///<
    ///  The set_row_col_hint() method lets you specify the location where the
    ///  cursor should be positioned.  It is called a 'hint' because mistakes in
    ///  the calling parameters are silently ignored.  Also, obedience to the
    ///  hint is optional.  The 'repaint' parameter indicates that the screen should
    ///  be immediately updated to reflect the change.
    ///
    ///  Again, obedience to this function is entirely optional and most viewers
    ///  will ignore it.
    ///
    ///  The line and column numbers are 0 base!  Contrary to normal user's
    ///  expectations, so account for this in any input calculations you make.

  virtual bool fetchAllLines(std::list<std::string> &output);
    ///<  An optional method that fetches all the lines in this viewer object 
    ///   into a caller supplied list of strings.
    ///
    ///   @warning  This is an optional feature, many viewers do not support it.
    ///
    ///   @param[out] output  The location to store all the lines from this viewer.
    ///
    ///   @return true to indicate that the data was extracted and false to indicate
    ///                that this viewer does not support this feature.
    ///      

  virtual void listen(std::string const &type, std::string const &msg) {}
    ///< Listen and respond to messages of various types.  By default, do nothing.
    ///  Override this function to handle the message.  Remember that the viewer
    ///  may not be visible when this function is invoked, so call @b manager()->is_visible(this)
    ///  to decide whether or not to display anything

};

class ListViewer
: public Viewer
  //
  /// A ListViewer is a simplified version of the generic Viewer interface.
  /// It only handles the viewing/editing of lists of strings.  Each
  /// string is presented in vertical list form -- one string per line.
  ///
  /// A ListViewer is not meant for handling wide strings like files might have.
  /// It is meant rather for the display of strings where horizontal
  /// scrolling isn't really necessary.  That is, you might need to compress
  /// your strings to fit in the space available rather than relying on
  /// horizontal scroll bars.  See DirectoryViewer's implementation.  The
  /// directory viewer chooses the text to be displayed based on the amount
  /// of horizontal space available -- and it provides a command, 'i', which
  /// toggles the display of certain less important text.
  ///
  /// To construct a ListViewer, you must supply a ListViewer::policies object
  /// which will provide the list of strings and also a command handler
  /// for events not processed by the ListViewer itself.  That is, the
  /// ListViewer will handle display, paging, etc.  All other commands must
  /// be processed -- one string at a time.
  ///
  /// Note that the policies object will be deleted by the ListViewer upon
  /// its destruction!
  ///
  /// Here's how to use a ListViewer:
  ///
  ///   @li  derive from ListViewer::policies to create a list of strings
  ///        to display and implement the handle_event method and the
  ///        resize_handler method (if needed -- its optional).
  ///
  ///   @li user operator new to create an object of that type and use
  ///       that value to construct a ListViewer object.
  ///
  ///   @li give the ListViewer to the ViewerManager -- or animate it
  ///       yourself using direction in Viewer.  (strongly advised against)
  ///
  /// Note that there are two ways to give a ListViewer to the viewer manager.
  /// One way is to manually create the ListViewer and use one of
  ///
  /// @code
  ///  ViewerManager::add    -- lets you manually new up the viewer then pass it to the viewer manager
  ///  ViewerManager::edit   -- lets ViewerManager figure out the viewer type, and add it to the system
  /// @endcode
  ///
  /// The other way is to let the ViewerManager create the viewer in the first
  /// place.  To accomplish this, you use the ViewerManager::add_app function to
  /// provide the viewer manager with a function that will create the ListViewer
  /// for you.  That is, you write a function that will prompt the user for
  /// parameters to the constructor to your ListViewer, then you give that function
  /// to the viewer manager.  The viewer manager will then be able to present the
  /// user with a list of 'application' functions and let him choose which function
  /// to call -- and thus which kind of application viewer to create.  A nice way to
  /// create this function is to make it a static member of the policies class, and
  /// you could also make the application name a static member of policies as well.
  /// In this way you don't have any extraneous free functions and variables laying
  /// around.
  ///
{
public:

   struct policies
     //
     ///  An abstract base class definining the list of strings to
     ///  to be displayed/edited and a command handler.  The command
     ///  handler must not do any displaying of its own.
     ///
     ///  Policies are virtual methods of this class which are invoked
     ///  to accomplish certain tasks.  Review the documentation on
     ///  the various methods below for more details.
     ///
     ///  In general, you should also create your 'add_application' function
     ///  that will be used by the ViewerManager as a static method of
     ///  this class as well.
     //
   {
     virtual ~policies();                ///< Even policies have to destructed

     std::vector<std::string> strings_;  ///< items to display in the ListViewer
     std::string              title_;    ///< title of list display viewer

     virtual std::string const& application_name() const =0; 
       ///<
       /// name of this application

     virtual std::string description() const =0;
       ///<
       /// A description of the active session.  Might be the name of the
       /// directory or file being viewed -- might be something else.

     virtual int handle_event(input_event const *e,
                              size_t             index,
                              ViewerManager*     vm
                             ) =0;
       ///<
       /// The handle_event method is invoked to handle events
       /// not otherwise processed by the ListViewer itself.
       /// The return code is interpreted as follows:
       ///
       /// @code
       ///   0  -- the strings_ list has not been changed
       ///   1  -- the list has changed, repaint the window
       ///   2  -- done -- close the window
       /// @endcode
       ///
       /// The 'index' parameter is the index into the strings_ array
       /// of the item under the cursor.
       ///
       /// Note that you can get the window from vm->window().
       ///
       /// Usage note:  If your handle event creates a new viewport that
       /// covers this viewport, make sure that your current viewport doesn't
       /// draw anything until the next repaint request.  Otherwise, your
       /// screen will get very jumbled up!  So for example, if you call
       /// a ViewerManager function to create a viewer or split an existing
       /// one, you must be careful about doing any drawing in the current
       /// viewport.  The ListViewer itself protects against this, but if
       /// you do drawing in the policies::handle_event() routine you'll need
       /// to do your own protection.
       ///
       /// You can delete and add strings_ in this method -- just make sure
       /// that return 1 if you do -- or 2.

     virtual void handle_resize(CursorWindow::row_col const &new_size);
       ///<
       ///  This method is optional -- it will be invoked to tell you what
       ///  the current viewport size is.  If you dont' care, don't implement
       ///  this method.  If you do care, you'll want to save the size
       ///  for later use.  You might want to define this function if want
       ///  to re-architect the strings_ array based on available screen space.
       ///
       ///  Note -- do not delete or add strings_ in this method.


     virtual bool is_marked(size_t index)=0;///< returns true if the specified
                                            ///  string index is 'marked'.
                                            ///  It will be displayed in a
                                            ///  special way.  Do not delete or
                                            ///  add strings_ in this method

     virtual bool is_special(size_t index); /// return true if this item
                                            /// should be displayed bold
                                        
     virtual std::auto_ptr< std::list<std::string> > help_text()=0;
        ///<
        ///   This function returns the help text specific to this application.
        ///   The ListViewer help method will first present the user with the return
        ///   values from this function then it will present ListViewer specific help,
        ///   then it will present ViewerManager specific help.
   
   
     virtual bool fetchAllLines(std::list< std::string > &output );
         ///<  @see cxxtls::Viewer::fetchAllLines().

     virtual void removeTemporaryFile() {}
         ///<   Called by the list viewer only when the file represented by this
         ///    this policy object is a temporary file.  This function's job
         ///    is to destroy the file or directory referred to by this policy
         ///    object.  Note that the default implemenation does nothing.

     virtual void set_row_col_hint(ListViewer*v, size_t row, size_t col, bool repaint, size_t &display_top, size_t &current_row){}
         ///<  Set the row column hint for the specified viewer
         ///
         ///   @param[in]    v             The viewer to set the policy for
         ///   @param[in]   row            The row to make sure is visible
         ///   @param[in]   col            The column to make sure is visible
         ///   @param[in]   repaint        Controls whether the screen gets updated
         ///   @param[out]  display_top    A reference to the display_top member in ListViewer
         ///   @param[out]  current_row    A reference to the current_row member in ListViewer

   
   };

   friend struct policies;  ///<  let the row col hint handler from the policies object
                            ///   mess around with the privates

   int rowsAvailable(CursorWindow::viewport *vp)
       ///  @return the number of rows on which list data can actually be displayed.
   {
      CursorWindow::row_col  size = vp->size();

      return size.row_ - first_row_;   
   }


private:

    std::auto_ptr<policies> com_;

    size_t display_top_;
    size_t cur_row_;
    int    first_row_;

    bool   active_;

    std::string search_string_;


public:

  ListViewer(policies* comtable, ViewerManager* man)
  : Viewer(man),
    com_(comtable),
    display_top_(0),
    cur_row_(0),
    first_row_(1),
    active_(false)
  {
    ///  Contruct a ListViewer given a pointer to a ListViewer::policies
    ///  object and a pointer to the ViewerManager which will own this
    ///  list viewer.  The ViewerManager will orchestrate the destruction
    ///  of this list viewer and that destruction will also cause the
    ///  destruction of the policies object.  Do not manually delete either.

  }

  virtual ~ListViewer();

  void operator() ( viewport * vp,int cmd);
     ///<
     /// Handle screen repaint policies (not virtual)

  bool handle_event(input_event const * e, viewport * vp); ///< (not virtual)

  void repaint(viewport *vp) ///  wrapper around the function call operator
                             ///  that makes it easier to invoke a repaint.
  {
     (*this)(vp, activate );
  }

  void output_row_text(viewport * vp, int i) ///  output one row of the viewed text
  {
    // do not set the text attributes here!

    if(i < (int)(com_->strings_.size()) )
      *vp << com_->strings_[i];
  }

  inline int line_attribute(bool selected, size_t index, bool is_bottom);
    ///<
    ///  Compute text attributes of lines given some information about
    ///  the line.

  int find_string(size_t index, ViewerManager* vm);
    ///<
    ///  Prompt the user for a string which will then be used to
    ///  initiate a search starting at the current line.  The cursor
    ///  will be left on the line, below the current line, which
    ///  contains the string -- if a match was found.


  int find_string(size_t index);
    ///<
    ///  Repeats the find operation started by find_string(index,vm), above.
    ///  The user will not be prompted for a new string -- the old one will
    ///  be used for this new search.

  virtual void help();  ///<  See ListViewer::policies::help_text for comments.

  std::string const& application_name() const
    ///  For ListViewers, the application name command is just passed through
    ///  to the policy object for implementation.
  {
    return com_->application_name();
  }

  std::string description() const
    ///  For ListViewers, the description() method is just passed through
    ///  to the policy object for implementation.
  {
    return com_->description();
  }

  virtual bool fetchAllLines(std::list<std::string> &output);
         ///<  @see cxxtls::Viewer::fetchAllLines().

  virtual void  set_row_col_hint(size_t line, size_t column, bool repaint);
     ///< Just calls polices::set_row_col_hint(*this, line, column, repaint)

  void resetContent()
     ///<  Called by the policies:: member functions to indicate that the content
     ///   has been changed.  This function resets the current row 
     ///   and display_top_ variables to 0s.
  {
      display_top_ = 0;
      cur_row_     = 0;
  }

};

} // namespace cxxtls

#endif
