#include <cxxtls/symbolviewer.h>
#include <cxxtls/viewermanager.h>
#include <cxxtls/file.h>
#include <portable_io.h>   // for popen()
#include <cxxtls/options.h>
#include <cxxtls/foreach.h>
#include <stdlib.h>

using namespace std;

#ifdef _MSC_VER
#define SLASH '\\'
#else
#define SLASH '/'
#endif

namespace cxxtls
{
//---------------------------------------------------------------------------------


SymbolViewer::
SymbolViewer(ViewerManager *vm, std::string databaseFilename)
:  TreeViewer(vm)
,  databaseName_( (databaseFilename.empty() ? "TAGPP.tagpp" : databaseFilename.c_str() ) )
{
     using namespace CppTagDB;

    bool generateDirectly = false; // the database name is not a real file
    bool linkTree         = false; // the database is just a set of filenames and text to use a link to select from

    size_t cppTagDbNameLength = 9; // .cpptagdb
    size_t lnkTreeNameLength  = 8; // .lnktree


    if(databaseName_.size() > cppTagDbNameLength)
    {
       char const *p = databaseName_.data();

       p += databaseName_.size()-cppTagDbNameLength;

       if(memcmp(p, ".cpptagdb", cppTagDbNameLength) == 0)
       {
          generateDirectly = true; // we have something like:   /home/dir/fred.cpp.cpptagdb

          databaseName_ = databaseName_.substr(0, databaseName_.size()-cppTagDbNameLength);
       }


    }
    
    if(!generateDirectly && (databaseName_.size() > lnkTreeNameLength) )
    {
       char const *p = databaseName_.data();

       p += databaseName_.size()-lnkTreeNameLength;

       if(memcmp(p, ".lnktree", lnkTreeNameLength) == 0)
       {
          linkTree = true; // we have something like:   /tmp/fred.o.lnktree

          linkTreeMode_ = true;  // affects help behavior

          // in this case, the file, /tmp/fred.o.lnktree exists but it's contents
          // refer to another file -- whose name is found on line one

       }

    }

    if(!generateDirectly && !linkTree)
    {
        // read a normal *.tagpp file

        char const *etagsPath = getenv("ETAGSPATH");
        
        char const *sourcePath= getenv("SOURCEPATH");
        
        if(!FileName(databaseName_).exists())
        {
            FileName location(databaseName_);
        
            if(etagsPath)
                location = location.find_file_in_path(databaseName_, etagsPath);
        
            if(location == databaseName_)
            {
               if(sourcePath == 0)
               {
                  sourcePath=".:..:../..:../../..:../../../..";
               }
        
                location = location.find_file_in_path(databaseName_, sourcePath);
            }
        
            databaseName_ = location;
        
        }
    }

    TreeViewer *me = this;
    me->setApplicationName(string("SymbolViewer: ") + databaseFilename);

    me->setRootDir(databaseName_.dirname());

    int errorLine=0;

    if(generateDirectly)
    {
        FileName curdir = FileName::getcwd();   // save current directory for later return

        string cpptagdb_executable = "cpptagdb.exe";
      
        if(ProgramOptions::global_options_)
        {
           FileName myProgram = ProgramOptions::global_options_->argv[0];  // 0 is known to exist
      
           FileName myDir = myProgram.dirname();
      
           FileName exeFile = myDir;
           
           exeFile += SLASH;
           
           exeFile += cpptagdb_executable;
      
           if(exeFile.exists())
           {
              cpptagdb_executable = exeFile;
           }
      
        }

        static string quote("\"");

        
        string cmd = quote + cpptagdb_executable + quote;

        if(databaseName_.size() > 2
           && databaseName_[ databaseName_.size() - 1] == '!'
          )
        {
           // the databaseName_ is a directory reference -- 
           // instead of using the normal command, let us
           // find the names of files in the specified directory
           // that are expected to be C++ and generate the
           // tag information for all of them.

           string dirname = databaseName_.substr(0, databaseName_.size()-2); 
              // everything up to the trailing /!

#ifdef _MSC_VER

           chdir(dirname.c_str());

           string searchCmd(" dir ");
           string searchOption(" /s /b ");
           string fileTypes("*.h *.hpp *.hxx *.h++ *.c *.cpp *.cxx *.c++ | ");

           cmd = searchCmd + searchOption + fileTypes + quote + cpptagdb_executable + quote + " -stdin ";

#else
           string searchCmd("find ");
           string searchOption(" ");
           string fileTypes("\\( -name '*.h' -o -name '*.hpp' -o -name '*.hxx' -o -name '*.h++' -o -name '*.c' -o -name '*.cpp' -o -name '*.cxx' -o -name '*.c++' \\) | ");
           cmd = searchCmd + quote + dirname + quote + searchOption + fileTypes + quote + cpptagdb_executable + quote + " -stdin ";

#endif
        }
        else
        {
            cmd += " \"";
            cmd += databaseName_;
            cmd += "\"";

        }


        vm->close();


        FILE *f=popen(cmd.c_str(), "r");
        
        database_.populate(f, errorLine);

        pclose(f);

        int rv=chdir(curdir.c_str());  // return to current directory (dos in particualar)

        rv=rv; // ingore chdir return value gosh darn it!

        vm->open();


    }
    else
    if(linkTree)
    {
        populateLinkTree();
    }
    else
        database_.populate(databaseName_, errorLine);

    if(!linkTree)
    {
        
        TreeViewerNode &topNode = this->topNode();
        topNode.setName(databaseName_);
        
        { // populate node tree
        
            TreeViewerNode &Symbols           = topNode.addChild("SYMBOLS");
        
           SymbolInfo::const_iterator first= database_.symbols_.children_.begin(),
                                      last = database_.symbols_.children_.end();
        
           while(first != last)
           {
              addChild(Symbols, first->second);
        
              ++first;
           }
            
        } // populate node tree
    }

}

//-----------------------------------------------------------------------------------

void SymbolViewer::
addChild(TreeViewerNode &parent, CppTagDB::SymbolInfo const &symbol)
{
    using namespace CppTagDB;

    TreeViewerNode &child = parent.addChild(symbol.annotatedName(true));

    if(symbol.line_ != 0)
    {
        child.adoptInfo(  new SymbolViewerNodeInformation(symbol.file_,
                                                          symbol.line_
                                                          )
                        );

    }
    else
    {
       child.setHighlighted();
    }

    if(!symbol.children_.empty())
    {

    
        SymbolInfo::Children::const_iterator first = symbol.children_.begin(),
                                             last  = symbol.children_.end();

        while(first != last)
        {
           SymbolInfo::Children::value_type const &cur = *first++;

           addChild(child, cur.second);

        }

    }

}


//-----------------------------------------------------------------------------------

bool 
SymbolViewerNodeInformation::
operator() (ViewerManager*vm, CursorWindow::viewport *vp, int command, FileName const &rootDir)
{
    FileName fileName(filename_);

    if(!fileName.is_absolute_path())
    {
        fileName = rootDir + fileName;
    }


    Viewer* oldie = vm->find_viewer(fileName);

    if(oldie)
    {
      oldie->set_row_col_hint(lineNumber_-1,0);
      vm->activate(oldie);
    }
    else
    {

      Viewer* newbie = vm->edit(fileName);

      if(newbie)
      {
        newbie->set_row_col_hint(lineNumber_-1,0);

        vm->vsplit(newbie, fileName);

      }
      else
        vm->active_viewport()->beep();
    }

    return false; // don't exit the browser
}
   

//-----------------------------------------------------------------------------------

Viewer* 
SymbolViewer::
app(ViewerManager*vm, std::string &fullname)
{
   SymbolViewer *rv =new SymbolViewer(vm, fullname);  



   return rv;
}
//-----------------------------------------------------------------------------------

void
SymbolViewer::
helpHeader(std::list<std::string> &help_text)
{
   if(linkTreeMode_)
   {
       help_text.push_back("");
       help_text.push_back( "Viewing Links" );
       help_text.push_back("");
       help_text.push_back("  This viewer contains a hierarchy of links to other files.");
       help_text.push_back("  The links refer to symbols or character strings found in same.");
       help_text.push_back("");
       help_text.push_back("  Bold text lines are selectable and normal text lines are headers.");
       help_text.push_back("");
       help_text.push_back("  Press enter to view the target of links in a split screen mode.");
   }
   else
   {
       help_text.push_back( "Viewing Symbols" );
       help_text.push_back("");
       help_text.push_back("  This window contains a hierarchy of symbols read from");
       help_text.push_back("  file or directory (or subtree) by the cpptagdb.exe");
       help_text.push_back("  program.  It lists the locations at which symbols are");
       help_text.push_back("  defined in files (not just declared, but defined.)");
       help_text.push_back("");
       help_text.push_back("  Each symbol is identified by name and type:");
       help_text.push_back("");
       help_text.push_back("    type names (classes, typedefs) etc are postfixed with {}");
       help_text.push_back("");
       help_text.push_back("    enums and #defines are postfixed with #");
       help_text.push_back("");
       help_text.push_back("    variables are identified by @");
       help_text.push_back("");
       help_text.push_back("    functions are annotated with ()");
       help_text.push_back("");
       help_text.push_back("  Use the Enter key to show the actual source of the symbol.");
       help_text.push_back("");
       help_text.push_back("  Use the =/+ keys to open and close one level of hierarchy");
       help_text.push_back("");
       help_text.push_back("  See below for more key definitions");
   }
}


//-----------------------------------------------------------------------------------

SymbolViewer::
~SymbolViewer()
{
   if(isTemporary())
   {
      databaseName_.remove();
   }
}


//-----------------------------------------------------------------------------------

namespace SymbolViewerHelper
{

    typedef std::multimap<size_t, std::string>      LineLinks;      // map line numbers to the text of the link

    typedef std::multimap<std::string,  LineLinks>  FileLinks;      // map file name to lines with links on them

    typedef std::multimap<std::string, FileLinks>   DirectoryTree;  // map directory name to the files in it

    //-----------------------------------------------------------------------------------
    
    void insertLinkage(DirectoryTree &tree, FileName const &file, int line, std::string const &text)
    {
        FileName    directory = file.dirname();
        std::string basename  = file.basename();
    
        DirectoryTree::iterator dir = tree.find(directory);

        if(dir == tree.end())
        {
           dir = tree.insert(  DirectoryTree::value_type( directory, FileLinks() ) );
        }

        FileLinks &fileLinks = dir->second;

        FileLinks::iterator fl = fileLinks.find(basename);

        if(fl == fileLinks.end())
        {
           fl = fileLinks.insert( FileLinks::value_type( basename, LineLinks() ) );
        }

        LineLinks &lineLinks = fl->second;

        lineLinks.insert( LineLinks::value_type( line, text ) );
    
    }


};

//-----------------------------------------------------------------------------------


void SymbolViewer::populateLinkTree()
{
        using namespace SymbolViewerHelper;

        DirectoryTree  tree;

        TreeViewerNode &topNode = this->topNode();

        ifstream file(databaseName_.c_str());

        //
        //  Read the first line of the database file (*.lnktree) and get the name of the directory
        //  or file that this tree is meant to go with
        //

        std::string realName;

        std::getline(file, realName);   // the first line of the file contains the real name
                                        // of the database

        topNode.setName(realName);
        topNode.setHighlighted();

        //
        //  Now read the second line which is supposed to describe the kinds of links in this tree
        //

        std::string tableName;

        std::getline(file, tableName);   // the SECOND line of the file contains the name
                                         // of the TABLE being displayed

        TreeViewerNode &parent = topNode.addChild(tableName);
        parent.setHighlighted();

        //
        //  Now read all the rest of the lines in the file and assume that they look like this:
        //
        //      filename:lineNumber,text
        //
        //  These fields will be split up and the text will be displayed to the user as the link
        //  from which to select. 
        //
        //  We will first populate a DirectoryTree object so we can associate file names with 
        //  directories and line numbers with the files they come from.
        //
        //  Later, we will copy the directory tree into the symbol viewer's database

        std::string line;

        while(std::getline(file, line))
        {
             size_t firstComma = line.find_first_of(',');

             if(firstComma < line.size())
             {
                // we have a line formatted something like this:
                //
                //    filename:lineNumber, symbol name

                size_t firstColon = line.find_first_of(':');

                size_t firstNonBlank = line.find_first_not_of(" \t", firstComma+1);

                string text;

                if(firstNonBlank < line.size())
                  text = line.substr(firstNonBlank);
                else
                  {
                    // all the characters in the current line are blank or maybe the line is empty.  

                    text=" ";  // pretend we have a line of interest;
                  }

                string file;
                size_t lineNumber=0;

                if(firstColon < firstComma)
                {
                   // s'all good

                   file  = line.substr(0, firstColon);

                   sscanf(line.c_str() + firstColon+1, "%ld", &lineNumber);

                   insertLinkage(tree, file, lineNumber, text);

                }
                else
                {
                   insertLinkage(tree, realName, lineNumber, text);
                }
             }
        }

        //
        //  Now we have the tree populated, let us now stick the tree in the symbol viewer
        //

        CXXTLS_FOREACH(DirectoryTree::value_type &dirInfo, tree)
        {
            TreeViewerNode &dirChild = parent.addChild(dirInfo.first);
            dirChild.setHighlighted();

            CXXTLS_FOREACH(FileLinks::value_type &fileInfo, dirInfo.second)
            {

                TreeViewerNode &fileChild = dirChild.addChild(fileInfo.first);  
                fileChild.setHighlighted();


                CXXTLS_FOREACH(LineLinks::value_type &lineInfo, fileInfo.second)
                {
                   string fullFilename = dirInfo.first + fileInfo.first;

                   TreeViewerNode &textChild = fileChild.addChild(lineInfo.second);  

                   if(lineInfo.first == 0)
                     textChild.setHighlighted();
                   else
                       textChild.adoptInfo(  new SymbolViewerNodeInformation(fullFilename,
                                                                             lineInfo.first
                                                                          )
                                        );
                       


                    
                }
            }

            dirChild.close(true); // close the children


        }




}


//-----------------------------------------------------------------------------------


}; // namespace cxxtls

